(in-package 'types)

;;;
;;; Retrieve the value with key x from an association list a.
;;;
(defun assoc-get (x a)
  (cdr (assoc x a :test #'equal)))

;;;
;;;  A, x :: t |- x :: t
;;;
(defun infer-variable (var eqns assump)
  (let ((typ (assoc-get var assump)))
    (if (is-quant typ)
	(quant-generic-instantiate typ eqns)
        typ)))


;;;
;;;     A, x :: t1 |- e :: t2
;;;   -------------------------
;;;     A |- \x.e :: t1 -> t2
;;;
(defun infer-abstraction (abstr eqns assump)
  (let* ((x (second abstr))
	 (term (third abstr))
	 (var (next-var eqns)))
    (type-func var 
	       (infer term eqns (acons x var assump)) 
	       (type-zero) 
	       eqns)))

;;;
;;;  A |- e1 :: t1, e2 :: t2
;;;  -----------------------      t1 <= a -> b  and  t2 <= a
;;;      A |- e1 e2 :: b
;;;
(defun infer-application (app eqns assump)
  (let* ((e1 (first app))
	 (e2 (second app))
	 (f (infer e1 eqns assump))
	 (x (infer e2 eqns assump))
	 (a (next-var eqns))
	 (b (next-var eqns))
	 (c (type-func a b (type-zero) eqns)))
    (add-todo-constraint f c eqns)
    (add-todo-constraint x a eqns)
    b))

;;;
;;; extend application notation to (f x1 ... xn)
;;;
(defun infer-application-list (app eqns assump)
  (infer-application (reduce #'(lambda (z y) (list z y)) app) 
		     eqns 
		     assump))

;;;
;;;  A |- p :: t2
;;;  A |- v :: t1
;;;  -----------------------
;;;  A |- v as p :: t1&t2
;;;
(defun infer-as (pat eqns assump)
  (type-inter 
   (infer (first pat) eqns assump)
   (infer (third pat) eqns assump)
   eqns))

;;;
;;; A |- func :: 1 -> 1 
;;;
(defun infer-patfunc (pat eqns assump)
  (declare (ignore pat assump))
  (type-func (type-one) (type-one) (type-zero) eqns))

;;;
;;;  A |- e :: t,
;;;  A u { v_i1 : a_i1,...,v_in : a_in } |- ei :: ti where t & ^pi <= pi
;;;  --------------------------------------------------------------------------
;;;  A |- case e of ..., pi : ei,... :: U{ ti ? (t & ^pi)}
;;;        where t & ^pi <= pi   
;;;              t <= U_i pi
;;;
;;; where ^pi = pi with all variables replaced by 1
;;;
(defun pattern-assumptions (pat eqns assump)
  (cond
   ((func? pat)
    assump)
   ((var? pat)
    (acons pat (next-var eqns) assump))
   ((constructor? pat)
    assump)
   ((constr-app? pat)
    (pattern-list-assumptions (cdr pat) eqns assump))
   ((aspat? pat)
    (let* ((var (first pat))
	   (npat (third pat))
	   (new-assumptions (pattern-assumptions npat eqns assump))
	   (hat-type (type-hat
		      (infer npat eqns new-assumptions)
		      eqns))
	   (assump (type-inter
		    (next-var eqns)
		    hat-type
		    eqns)))
      (acons var assump new-assumptions)))))
   
(defun pattern-list-assumptions (plist eqns assump)
  (if (null plist)
      assump
    (pattern-list-assumptions (cdr plist) 
			      eqns 
			      (pattern-assumptions (first plist) eqns assump))))

(defun infer-one-case (cas eqns assump)
  (let* ((pattern (first cas))
	 (branch (second cas))
	 (new-assump (pattern-assumptions pattern eqns assump))
	 (cas-typ (infer pattern eqns new-assump)))
    (cons 
     (infer branch eqns new-assump)
     cas-typ)))

(defun infer-case (term eqns assump)
  (let* ((e (second term))
	 (cases (cdddr term))
	 (typ (infer e eqns assump))
	 (typs (mapcar #'(lambda (x) (infer-one-case x eqns assump))
		       cases))
	 ;; typs is of the form (...(ti . pi')...)
	 (bound (reduce #'(lambda (x y) (type-union x y eqns))
			(mapcar #'cdr typs) 
			:initial-value (type-zero)))
	 (ans (reduce #'(lambda (x y) (type-union x y eqns))
		      (mapcar #'(lambda (x)
				  (type-dep (car x)
					    (type-inter typ 
							(type-hat (cdr x) eqns)
							eqns)
					    eqns))
			      typs)
		      :initial-value (type-zero))))
    (add-todo-constraint typ bound eqns)
    ans))
    
 
;;;
;;; A |- f : t1  where t1 <= a -> b
;;; A u { v_i1 : a_i1,...,v_in : a_in } |- p : s, e :: t2 
;;; -------------------------------------------------------------
;;; A |- f extend (p e) : c -> (b ? (c & ~s)) | (t2 ? (c & ^s)) 
;;;        where  c <= (a&~s | s)

(defun infer-extend (term eqns assump)
  (let* ((f (first term))
	 (ext (third term))
	 (t1 (infer f eqns assump))
	 (etyps (infer-one-case ext eqns assump))  ;; etyps is (t2 . s)
	 (t2 (car etyps))
	 (s (cdr etyps))
	 (a (next-var eqns))
	 (b (next-var eqns))
	 (c (next-var eqns)))
    (add-todo-constraint t1 (type-func a b (type-zero) eqns) eqns)
    (add-todo-constraint c
			 (type-union
			  (type-inter a
				      (type-compl s eqns)
				      eqns)
			  s
			  eqns)
			 eqns)
    (type-func
     c
     (type-union
      (type-dep b (type-inter c (type-compl s eqns) eqns) eqns)
      (type-dep t2 (type-inter c (type-hat s eqns) eqns) eqns)
      eqns)
     (type-zero)
     eqns)))    
    
;;;
;;;   A |- e1 : t1    A u { x : t1 } |- e2 : t2
;;; --------------------------------------------
;;;        A |- let x = e in e' : t2
;;;
;;; Generalizations are performed as part of the
;;; inference of the type of a let-bound variable.
;;; Care is taken to ensure that variables free in the
;;; assumptions are not generalized.
;;;
(defun vars-in-assumptions (assump eqns)
  (apply #'append
	 (mapcar #'(lambda (pair) 
		     (free-vars (type-vars-reachable (cdr pair) eqns)))
		 assump)))

(defun vars-not-free-in-assumptions (typ eqns assump)
  (let* ((fvs (free-vars (type-vars-reachable typ eqns :fv-expand t)))
	 (avs (vars-in-assumptions assump eqns)))
    (remove-if #'(lambda (v) (member v avs :test #'equal)) fvs)))

(defun infer-let (term eqns assump)
  (let* ((x (second term))
	 (e1 (fourth term))
	 (e2 (sixth term))
	 (typ1 (infer e1 eqns assump)))
    (if (eq (protect-solve-constraints eqns) :consistent)
	(let* ((typ2 (var-elim typ1 eqns :initPN (vars-in-assumptions assump eqns)))
	       (genvars (vars-not-free-in-assumptions typ2 eqns assump))
	       (typ (type-quant genvars typ2 eqns)))
	  (infer e2 eqns (acons x typ assump)))
        (type-err))))

;;;
;;; A |- ei : ti   1 <= i <= n
;;; --------------------------------
;;; A |- c(e1,...,en) : c(t1,...,tn)
;;;
(defun infer-constr (term eqns assump)
  (let ((typs (mapcar #'(lambda (x) (infer x eqns assump)) (cdr term)))
	(cid (cons-id-get (first term) eqns)))
    (mk-cons cid typs)))

;;;
;;; -------------------------- 
;;; A |- () : Rec(0 -> 0)
;;;
(defun infer-empty-rec (term eqns assump)
  (declare (ignore term assump))
  (type-record (mk-func (type-zero) (type-zero) (type-zero))
	       eqns))

;;;
;;;  A |-  r : R        R <= Rec(A -> B)
;;;  A |-  x : X
;;;------------------------------------------------------------------
;;;  A |- r with L = x : Rec((A & ~L) | (v & L)) -> (B | (X?(v&L)))
;;;
(defun infer-with (term eqns assump)
  (let* ((typr (infer (first term) eqns assump))
	 (typx (infer (fifth term) eqns assump))
	 (a (next-var eqns))
	 (b (next-var eqns))
	 (v (next-var eqns))
	 (r (type-record (type-func a b (type-zero) eqns) eqns))
	 (L (type-cons (cons-id-get (third term) eqns) nil eqns)))
    (add-todo-constraint typr r eqns)
    (type-record
     (type-func
      (type-union
       (type-inter a (type-compl L eqns) eqns)
       (type-inter v L eqns)
       eqns)
      (type-union
       b
       (type-dep typx (type-inter v L eqns) eqns)
       eqns)
      (type-zero)
      eqns)
     eqns)))
    
;;;
;;; A |-   r :: R   R <= Rec(a -> b)
;;; ------------------------------------
;;; A |- r without L :: Rec(a & ~L -> b)
;;;
(defun infer-without (term eqns assump)
  (let* ((typr (infer (first term) eqns assump))
	 (a (next-var eqns))
	 (b (next-var eqns))
	 (r (type-record (type-func a b (type-zero) eqns) eqns))
	 (L (type-cons (cons-id-get (third term) eqns) nil eqns)))
    (add-todo-constraint typr r eqns)
    (type-record
     (type-func
      (type-inter a (type-compl L eqns) eqns)
      b
      (type-zero)
      eqns)
     eqns)))

;;;
;;;   A |- r : R     
;;;----------------   R <= Rec(a -> b),  L <= a
;;;   A |- r.L : b
;;;
(defun infer-select (term eqns assump)
  (let* ((typr (infer (car term) eqns assump))
	 (a (next-var eqns))
	 (b (next-var eqns))
	 (r (type-record (type-func a b (type-zero) eqns) eqns))
	 (L (type-cons (cons-id-get (cdr term) eqns) nil eqns)))
    (add-todo-constraint typr r eqns)
    (add-todo-constraint L a eqns)
    b))

;;;
;;; type-record takes a type X and forms the type REC(X)
;;;	 
(defun type-record (typ eqns)
  (type-cons (cons-id-get 'rec eqns)
	     (list typ)
	     eqns))

;;;
;;; infer is the main type inference routine.  It is recursive
;;; on structure of the term; the order of cases is significant
;;; here.
;;;
(defun infer (term eqns assump)
  (cond
   ((integerp term)
    (mk-cons (cons-id-get :int eqns) nil))

   ((floatp term)
    (mk-cons (cons-id-get :float eqns) nil))

   ((constructor? term)
    (mk-cons (cons-id-get term eqns) nil))

   ((func? term)
    (infer-patfunc term eqns assump))
   
   ((empty-rec? term)
    (infer-empty-rec term eqns assump))

   ((var? term)
    (infer-variable term eqns assump))

   ((extend? term)
    (infer-extend term eqns assump))

   ((with? term)
    (infer-with term eqns assump))

   ((without? term)
    (infer-without term eqns assump))
   
   ((select? term)
    (infer-select term eqns assump))

   ((paren-term? term)
    (infer (first term) eqns assump))

   ((abstraction? term)
    (infer-abstraction term eqns assump))

   ((case? term)
    (infer-case term eqns assump))

   ((let? term)
    (infer-let term eqns assump))

   ((constr-app? term)
    (infer-constr term eqns assump))

   ((aspat? term)
    (infer-as term eqns assump))

   ((application? term)
    (infer-application-list term eqns assump))

   (t (break "error in infer"))))

;;;
;;; The main type inference function.  
;;;
(defun typ (term)
  (let* ((eqns (parse-term term))
	 (typ (infer term eqns nil)))
    (if (eq (protect-solve-constraints eqns) :consistent)
	(let ((simp-typ (var-elim typ eqns)))
	  (deparse-term term)
	  (format t "***~%")
	  (ppt (type-quant-all simp-typ eqns) eqns))
        (type-err))))

(defun type-err () 
  (format t "*** type error ***") 
  (break))
