(in-package 'types)
;;;
;;; This file contains sample Illyria expressions for use with
;;; the type inference system.  
;;;

(defvar t-id          '(lam x x))                               ;; \x.x
(defvar t-k           '(lam x (lam y x)))                       ;; \xy.x
(defvar t-half-bottom '(lam x (x x)))                           ;; \x.x x
(defvar t-bottom      `(,t-half-bottom ,t-half-bottom))         ;; (\x. x x)(\x. x x)

(defvar t-y '(lam u ((lam x (u (x x))) (lam x (u (x x))))))

(defvar t-yid `(,t-y ,t-id))                                  ;; Y id
(defvar t-yk  `(,t-y ,t-k))                                   ;; Y K

(defvar t-twice '(lam f (lam x (f (f x)))))                   ;; \f.\x.f(f x)
(defvar t-twicetwice `(,t-twice ,t-twice))                    ;; twice twice
(defvar t-twice-id `(,t-twice ,t-id))                         ;; twice id
(defvar t-y-twice-id `(,t-y (,t-twice ,t-id)))                ;; Y twice id
(defvar t-notype `((:true) :true))                            ;; true true

;;;
;;;  a version of Y combinator that is delayed
;;;    to make it appropriate for strict functions.
;;;  In particular, Yd ("Y delayed") has the following property:
;;;      (Yd f)    ==>    \x . ( f (Yd f) x )      ;; provided x is a fresh var
;;;  Therefore, the following also hold:
;;;      (Yd f) rewrites to a head normal form
;;;      ((Yd f) x)   ==>   f (Yd f) x
;;;

(defvar t-halfy-delayed '(lam x (lam w (
				      (u (x x)) w
				      )))
      ) ;;  \x. \w. ( u (x x) w )

(defvar t-y-delayed '(lam u ( 
			   (lam x (lam w (
					  (u ( x x )) w
					  )))
			   (lam x (lam w (
					  (u ( x x )) w
					  )))
			   ))
      ) ;; \u. (   (\x.\w. (u (x x) w))   (\x.\w. (u (x x) w))   )
;;;
;;; The following examples are due to Lennart Augustsson.
;;;
;;; The first group of examples is an implementation of lists 
;;; in pure lambda calculus.  The representation for lists uses
;;; a general scheme for encoding pattern matching.
;;; If a data type has k constructors, then each constructor is
;;; encoded by a function of arity n + k, where n is the arity of
;;; the constructor.  The general form of the encoding of a constructor ci 
;;; is
;;;      \a1,...,an,c1,...,cn. ci a1 ... an
;;;
;;; Given an element s of the data type, pattern matching is done by
;;; applying s to k functions, one for each constructor in the data type.
;;;
(defvar t-nil '(lam n (lam c n)))                            ;; \n,c.n
(defvar t-cons '(lam hd (lam tl (lam n (lam c (c hd tl)))))) ;; \hd,tl,n,c. c hd tl
;; Y \map.\f.\s.(s nil \hd.\tl. cons (f hd) (map f tl))
(defvar t-map `(,t-y (lam map (lam f (lam s (s ,t-nil (lam hd (lam tl (,t-cons (f hd) (map f tl))))))))))

;; cons id nil
(defvar t-id-list `(,t-cons ,t-id ,t-nil))

;; \s. s nil \hd.\tl. hd
(defvar t-car `(lam s (s ,t-nil (lam hd (lam tl hd)))))

;; \s. s nil \hd.\tl. tl
(defvar t-cdr `(lam s (s ,t-nil (lam hd (lam tl tl)))))

;; \s. s true \hd.\tl. false
(defvar t-isnil `(lam s (s :true (lam hd (lam tl :false)))))

(defvar t-select-id `(,t-car ,t-id-list))
(defvar t-select-nil `(,t-cdr ,t-id-list))

(defvar t-true `(,t-isnil ,t-nil))
(defvar t-false `(,t-isnil ,t-id-list))

;;; 
;;; An implementation of the natural numbers with addition, multiplication,
;;; and factorial.
;;;
; zero = \z s.z
(defvar t-zero '(lam z (lam s z)))

; succ n = \z s.s n
(defvar t-succ '(lam n (lam z (lam s (s n)))))

; add n m = n m (\k.succ (add k m))
(defvar t-add `(,t-y (lam add (lam n (lam m ((n m) (lam k (,t-succ ((add k) m)))))))))


(defvar t-one `(,t-succ ,t-zero))

;;    t-succ   has type   a -> 1 -> (a -> b) -> b
;;    t-zero   has type   a -> 1 -> a

(defvar t-two `(,t-succ ,t-one))		;; 1 -> ((1 -> ((a -> 1 -> a) -> b) -> b) -> c) -> c
(defvar t-z+z `(,t-add ,t-zero ,t-zero))	;; a -> 1 -> a
(defvar t-o+z `(,t-add ,t-one ,t-zero))	;; v N.a->1->a | 1->(N->b)->b
(defvar t-z+o `(,t-add ,t-zero ,t-one))	;;        1 -> ((a -> 1 -> a) -> b) -> b
(defvar t-o+o `(,t-add ,t-one ,t-one))	;; v N.1 -> ((a -> 1 -> a) -> b) -> b | 1->(N->b)->b

;; mul n m = n zero \k. add m (mul k m)
(defvar t-mul `(,t-y (lam mul (lam n (lam m (n ,t-zero (lam k (,t-add m (mul k m)))))))))

;; fac n = n one \k. mul n (fac k)
(defvar t-fac `(,t-y (lam fac (lam n (n ,t-one (lam k (,t-mul n (fac k))))))))

(defvar t-z*z `(,t-mul ,t-zero ,t-zero))
(defvar t-z*o `(,t-mul ,t-zero ,t-one))
(defvar t-o*z `(,t-mul ,t-one ,t-zero))
(defvar t-o*o `(,t-mul ,t-one ,t-one))

(defvar t-z! `(,t-fac ,t-zero))
(defvar t-o! `(,t-fac ,t-one))


;;;
;;; examples with let and case
;;;

(defvar t-let1 '(let x = 1 in x))
(defvar t-let2 '(let x = (lam z z) in (x x)))

(defvar t-case1 '(lam x (case x of (:int x) (:float x))))

;;;
;;; Note the difference in the quality of type information for
;;; the two definitions of the lisp function "car" below.
;;;

(defvar t-if '(lam x (lam y (lam z (case x of (:true y) (:false z))))))
(defvar t-null? '(lam x (case x of (:nil :true) ((:cons y z) :false))))
(defvar t-hd '(lam x (case x of ((:cons y z) y))))
(defvar t-lisp-car `(lam x (,t-if (,t-null? x) :nil (,t-hd x))))
(defvar t-car-app `(,t-lisp-car (:cons 1 :nil)))


(defvar t-simple-car `(lam x (case x of (:nil :nil) ((:cons y z) y))))
(defvar t-simple-car-app `(,t-simple-car (:cons 1 :nil)))
(defvar t-let-car-app `(let x = ,t-simple-car in (x (:cons 1 :nil))))

;;;
;;; Note how the use of an "as" pattern improves type inference
;;; in the next two examples.
;;;
(defvar t-nlast
  `(,t-y (lam f (lam x (case x of ((:cons y :nil) y) ((:cons u (:cons v w)) (f (:cons v w))))))))
(defvar t-nlast-app `(,t-nlast (:cons 1.0 (:cons 2.0 (:cons 1 :nil)))))

(defvar t-aslast
  `(,t-y-delayed (lam f (lam x (case x of ((:cons y :nil) y) ((:cons u (v as (:cons a b))) (f v)))))))
(defvar t-aslast-app `(,t-aslast (:cons 1.0 (:cons 2.0 (:cons 1 :nil)))))

;;;
;;; examples from "Soft Typing with Conditional Types", POPL '94
;;;
(defvar t-examp1 
      `(let car = ,t-simple-car in (car (:cons 1 :nil))))

(defvar t-examp2
      `(let Y = ,t-y-delayed in 
	    (let last = (Y (lam f (lam x (case x of ((:cons y :nil) y) ((:cons u (v as (:cons a b))) (f v)))))) in
		 (let a = (last (:cons 1 (:cons 2 (:cons 3 :nil)))) in
		      (last (:cons 1.0 (:cons 2.0 (:cons a :nil))))))))

;;;
;;; tautology example from Cartwright and Fagan (SIGPLAN '92)
;;;
(defvar t-taut 
      `(let and = (lam x (lam y (case x of (:true y) (:false :false)))) in
        (let Y = ,t-y in
	     (Y (lam taut (lam b (case b of (:true :true) 
                                            (:false :false) 
                                            (func (and (taut (b :true))
						       (taut (b :false))))
					     )))))))

;;;
;;; deep, slightly modified from Cartwright and Fagan
;;;
(defvar t-deep `(let not = (lam x (case x of (:true :false) (:false :true))) in
		   (let Y = ,t-y-delayed in 
			(Y 
			(lam deep 
                             (lam n (case n of (:true (:cons n :nil)) 
				               (:false (:cons (deep (not n)) :nil)))))))))
				
;;;
;;; Examples using records.
;;;

(defvar t-empty ())
(defvar t-one-field '(() with :a = 1))
(defvar t-two-fields `(,t-one-field with :b = :true))
(defvar t-replace `(,t-two-fields with :b = 2.0))

(defvar t-forget1 `(,t-one-field without :a))
(defvar t-forget2 `(,t-two-fields without :a))
(defvar t-forget3 `(,t-replace without :b))

(defvar t-select1 `(,t-one-field . :a))
(defvar t-select2 `(,t-two-fields . :b))
(defvar t-select3 `(,t-replace . :b))
(defvar t-noselect `(,t-forget3 . :b))

(defvar t-let3 '(let r = (() with :a = 1) in
		   (r with :b = 2.0)))

(defvar t-mat `(,t-y (lam sigma (lam x (case x of (:int x) 
					     ((:cons l r) 
					      (let dummy = (sigma l) in 
						   (sigma r))))))))
;;;
;;;  The following examples are due to Didier Remy.
;;;
(defvar t-remy1 `(,t-if :true ,t-one-field ,t-two-fields))
(defvar t-remy2 `(,t-remy1 . :a))
(defvar t-remy3 `(,t-remy1 . :b))   ;; should not (and does not) type

(defvar t-remy4 `(lam x (,t-if (x . :a) x (() with :b = 1))))

;;;
;;; Complex addition, due to TK Lakshman
;;;

(defvar t-complex1 `((() with :real-part = ,t-zero) with :imag-part = ,t-zero))
(defvar t-complex2 `((() with :real-part = ,t-zero) with :imag-part = ,t-zero))

(defvar t-complex-add 
  `(lam tc1 
	(lam tc2 
	     (( () with :real-part = (,t-add (tc1 . :real-part) 
					     (tc2 . :real-part)))
	      with :imag-part = (,t-add (tc1 . :imag-part) 
					(tc2 . :imag-part))))))

(defvar t-complex-res `(,t-complex-add ,t-complex1 ,t-complex2))


;;;
;;;  Examples of function extension (a feature suggested by Radu Grosu).
;;;
;;; The system does not do a good job of simplifying the types of these
;;; functions, so the types are hard to read.  However, the types are
;;; accurate.
;;;
(defvar t-base '(lam x (case x of ((y as :x) y))))

(defvar t-ext1 `(,t-base extend ((z as :z) z)))

(defvar t-ext2 `(,t-ext1 extend ((y as (:cons a b)) y)))
(defvar t-app1-ext2 `(,t-ext2 :x))
(defvar t-app2-ext2 `(,t-ext2 :z))
(defvar t-app3-ext2 `(,t-ext2 (:cons 1 2)))

(defvar t-ext3 `(,t-ext2 extend (:z :y)))
(defvar t-app1-ext3 `(,t-ext3 :x))
(defvar t-app2-ext3 `(,t-ext3 :z))
(defvar t-app3-ext3 `(,t-ext3 (:cons 1 2)))
