(in-package 'types)
;;;
;;; Illyria includes the lambda calculus, constructors,
;;; pattern matching (using a case expression), polymorphic
;;; let, record operations (adding a field, removing a field,
;;; and field selection), and an operation for extending functions
;;; with new clauses.
;;;
;;; The grammar for Illyria is given below.  Lisp syntax is
;;; used to simplify parsing.
;;;
;;; term ::= x                                  x
;;;          (lam x term)                       \x.term
;;;          (case term of ...(p_i e_i)...)     case e of ...,p_i: e_i,...
;;;          (let x = term in term)             
;;;          (:c term_1 ... term_n)             c(term_1,...,term_n)
;;;          ()                                 {}  (empty record)
;;;          (r with L = x )                        (record extension)
;;;          (r without L)                          (forgetting a field)
;;;          (r . a)                            r.a (selection; blanks required)
;;;          (f extend (p_i e_i))               extend function f (see below)
;;;          (term term)                        
;;;          (term)
;;;          0,1,2,...                          integers
;;;          0.0,1.0,...                        floats
;;;
;;;  pattern ::= :c | (:c pat_1 ... pat_n) | x | var as pat | func
;;;
;;;  Constructors are written :x to distinguish them from variables.
;;;  A record label is considered to be a (nullary) constructor.
;;;
;;;  The semantics of function extension is as follows:
;;;  
;;;  (f extend (p(y_1,...,y_n) e))(x) =
;;;      if p(t_1,...,t_n) = x then
;;;         e[t_1/y_1,...,t_n/y_n]
;;;      else
;;;         f(x)
;;;
;;;  There are several restrictions on terms:
;;;    - Programs must not contain free variables. The parser checks this.
;;;    - Patterns must be linear (variables appear once). This is not checked.
;;;    - Patterns must be disjoint; i.e., no two patterns in the same
;;;        case expression can match the same value.  This is not checked.
;;;    - A constructor must be used with the same arity in all occurrences.
;;;      This is checked.  The constructors :int and :float are reserved.
;;;

;;;
;;; A function to test whether a list is a real list (i.e., nil terminated).
;;;
(defun true-listp (x)
        (or (null x)
            (and (consp x)
                 (true-listp (cdr x)))))


(defun parse-term (term)
  (parse term (new-tbl) nil))

;;;
;;; A simple parser for Illyria expressions.  The word "parser" is a 
;;; bit glorious; parse really just checks for some syntax errors
;;; (not all!) and builds a table of the consructors of the term.
;;;
;;; The order of cases is significant in parse.
;;;
(defun parse (term tbl freevars)
  (cond
   ; integers
   ((integerp term) 
    tbl)

   ; floats
   ((floatp term) 
    tbl)

   ; nullary constructors
   ((constructor? term) 
    (add-constructor term 0 tbl))

   ; empty record
   ((empty-rec? term)
    tbl)

   ; variables
   ((var? term)
    (check-freevar term freevars)
    tbl)

   ; function extension
   ((extend? term)
    (let ((newtbl (parse (first term) tbl freevars)))
      (parse-case-list (cddr term) newtbl freevars)))

   ; record extension
   ((with? term)
    (let ((newtbl (add-constructor (third term) 0 tbl)))
      (parse-term-list (list (first term) (fifth term)) newtbl freevars)))

   ; field deletion
   ((without? term)
    (let ((newtbl (add-constructor (third term) 0 tbl)))
      (parse (first term) newtbl freevars)))
   
   ; field selection
   ((select? term)
    (let ((newtbl (add-constructor (cdr term) 0 tbl)))
      (parse (car term) newtbl freevars)))

   ; parenthesized term
   ((paren-term? term)
    (parse (first term) tbl freevars))

   ; lambda abstraction
   ((abstraction? term)
    (parse (third term) tbl (cons (second term) freevars)))

   ; case expression
   ((case? term)
    (let ((newtbl (parse (second term) tbl freevars)))
      (parse-case-list (cdddr term) newtbl freevars)))

   ; let expression
   ((let? term)
    (let ((newtbl (parse (fourth term) tbl freevars)))
      (parse (sixth term) newtbl (cons (second term) freevars))))

   ; constructor application
   ((constr-app? term)
    (let ((newtbl (add-constructor (first term) (length (cdr term)) tbl)))
      (parse-term-list (cdr term) newtbl freevars)))

   ; function application
   ((application? term)
    (parse-term-list term tbl freevars))

   ; otherwixse, a syntax error
   (t (syntax-error term))))


(defun parse-term-list (tlist tbl freevars)
  (if (null tlist)
      tbl
      (let ((newtbl (parse (first tlist) tbl freevars)))
	(parse-term-list (cdr tlist) newtbl freevars))))

(defun parse-pattern (pat tbl)
  (cond
   ((func? pat)
    tbl)
   ((var? pat)
    tbl)
   ((constructor? pat)
    (add-constructor pat 0 tbl))
   ((aspat? pat)
    (parse-pattern (third pat) tbl))
   ((constr-app? pat)
    (let ((newtbl (add-constructor (first pat) (length (cdr pat)) tbl)))
      (parse-pattern-list (cdr pat) newtbl)))
   (t (format t "Syntax error in pattern: ~a~%" pat))
   ))
	
(defun parse-pattern-list (plist tbl)
  (if (null plist)
      tbl
      (let ((newtbl (parse-pattern (first plist) tbl)))
	(parse-pattern-list (cdr plist) newtbl))))

(defun parse-case-list (plist tbl freevars)
  (if (null plist)
      tbl
    (if (and (true-listp (first plist))
	     (= (length (first plist)) 2))
	(let* ((pair (first plist))
	       (pat (first pair))
	       (term (second pair))
	       (tbl2 (parse-pattern pat tbl))
	       (pvars (pattern-vars pat))
	       (tbl3 (parse term tbl2 (append pvars freevars))))
	  (parse-case-list (cdr plist) tbl3 freevars))
      (pattern-error (first plist)))))

;;;
;;; pattern-vars returns a list of the the variables 
;;; appearing in a pattern.
;;;
(defun pattern-vars (pat &optional (vars nil))
  (cond
   ((func? pat)
    vars)
   ((var? pat)
    (cons pat vars))
   ((constructor? pat)
    vars)
   ((constr-app? pat)
    (pattern-vars-list (cdr pat) vars))
   ((aspat? pat)
    (pattern-vars (third pat) (cons (first pat) vars)))
   (t (break "error in pattern-vars"))))

(defun pattern-vars-list (plist vars)
  (if (null plist)
      vars
    (pattern-vars-list (cdr plist) (pattern-vars (first plist) vars))))


;;;
;;; The following functions deal with building the list of constructors
;;; in a program.  new-tbl generates a fresh table with the constructors
;;; int, float,and rec (for record operations).
;;;
;;; add-constructor adds a constructor appearing in the program and checks 
;;; that all uses have the same arity.
;;;
(defun new-tbl ()
  (let ((a (make-tbl)))
    (cons-name-add 1 :int a)
    (cons-arity-add 1 0 a)
    (cons-id-add :int 1 a)
    (cons-name-add 2 :float a)
    (cons-arity-add 2 0 a)
    (cons-id-add :float 2 a)
    (cons-name-add 3 'rec a)
    (cons-id-add 'rec 3 a)
    (cons-arity-add 3 1 a)
    a))

(defvar *cons-id* 4)

(defun add-constructor (c arity tbl)
  (let ((cid (cons-id-get c tbl)))
    (if cid
	(if (= (cons-arity-get cid tbl) arity)
	    tbl
	    (arity-error c))
        (progn
	  (cons-id-add c *cons-id* tbl)
	  (cons-arity-add *cons-id* arity tbl)
	  (cons-name-add *cons-id* c tbl)
	  (incf *cons-id*)
	  tbl))))
	
;;;
;;; precicates for testing whether a symbol is a variable or
;;; a constructor.  Currently, any symbol of the form :x
;;; is a constructor.  All other symbols are treated as variables.
;;;
(defun var? (x)
  (and
   (symbolp x)
   (not (constructor? x))))

(defun constructor? (x)
  (and
   (symbolp x)
   (eq (find-package 'keyword) (symbol-package x))))

(defun paren-term? (x)
  (and (true-listp x)
       (= 1 (length x))))

(defun abstraction? (x)
  (and (true-listp x)
       (= 3 (length x))
       (eq (first x) 'lam)
       (var? (second x))))

(defun case? (x)
  (and (true-listp x)
       (<= 4 (length x))
       (eq (first x) 'case)
       (eq (third x) 'of)
       (true-listp (fourth x))))

(defun extend? (x)
  (and (true-listp x)
       (= 3 (length x))
       (eq (second x) 'extend)))

(defun let? (x)
  (and (true-listp x)
       (= 6 (length x))
       (eq (first x) 'let)
       (eq (third x) '=)
       (eq (fifth x) 'in)
       (var? (second x))))

(defun constr-app? (x)
  (and (true-listp x)
       (<= 1 (length x))
       (constructor? (first x))))

(defun application? (x)
  (and (true-listp x)
       (<= 2 (length x))
       (not (constructor? (first x)))
       (not (member (first x) '(lam case let)))))

(defun aspat? (pat)
  (and (true-listp pat)
       (= (length pat) 3)
       (var? (first pat))
       (eq 'as (second pat))))

(defun func? (pat)
  (eq pat 'func))

(defun empty-rec? (term)
  (null term))

(defun with? (term)
  (and (true-listp term)
       (= 5 (length term))
       (eq (second term) 'with)
       (constructor? (third term))
       (eq (fourth term) '=)))

(defun without? (term)
  (and (true-listp term)
       (= 3 (length term))
       (eq (second term) 'without)
       (constructor? (third term))))

(defun select? (term)
  (and (consp term)
       (constructor? (cdr term))))

;;;
;;; routines for reporting syntax errors
;;;
(defun syntax-error (x)
  (format t "Syntax error: ~a~%" x)
  (break))

(defun arity-error (c)
  (format t "Syntax error: constructor ~a used with multiple arities." c)
  (break))

(defun check-freevar (x vars)
  (when (not (member x vars))
	(format t "Syntax error: the program has an unbound variable ~a.~%" x)
	(break)))

(defun pattern-error (x)
  (format t "Syntax error: in a case, ~a is not a (pat term) pair" x)
  (break))
