(in-package 'types)

;;;
;;; This file implements type alpha-substitute function
;;; (alpha-subst X h) represents the result of applying the alpha-substitution
;;; alpha : var -> var | :no-change  to the type X.
;;; Functions in this file are called from the top level alpha-subst function
;;; which is defined in a separate file "alpha-subst_tbl.lisp" (this is done so as to allow
;;; the lisp compiler to define the functions (here) in the right order.
;;; 

;;; 0. E |- alpha-subst(0) = 0
;;;

(defun alpha-subst-zero (typ h p eqns) 
  (declare (ignore typ h p eqns))
  :no-change)


;;; 1. alpha-subst(1) = 1
;;;

(defun alpha-subst-one(typ h p eqns) 
  (declare (ignore typ h p eqns))
  :no-change)

;;; 
;;;
;;; 2.  alpha-subst(A | B) = alpha-subst(A) | alpha-subst(B) 
;;;

(defun alpha-subst-union (X h p eqns)
  (let ((lhs (type-alpha-subst-aux (type-t1 X) h p eqns))
	(rhs (type-alpha-subst-aux (type-t2 X) h p eqns)))
    (cond 
     ((and (eql :no-change lhs) (eql :no-change rhs)) 
      :no-change)
     ((eql :no-change lhs) 
      (mk-union (type-t1 X) rhs :disjoint (is-rtype X) :nosubset (is-nsunion X)))
     ((eql :no-change rhs) 
      (mk-union lhs (type-t2 X) :disjoint (is-rtype X) :nosubset (is-nsunion X)))
     (t 
      (mk-union lhs rhs :disjoint (is-rtype X) :nosubset (is-nsunion X) )))))

;;;
;;;  3. alpha-subst(A*B) = alpha-subst(A) * alpha-subst(B)
;;;
(defun alpha-subst-inter (X h p eqns)
  (let ((lhs (type-alpha-subst-aux (type-t1 X) h p eqns))
	(rhs (type-alpha-subst-aux (type-t2 X) h p eqns)))
    (cond 
     ((and (eql :no-change lhs) (eql :no-change rhs)) 
      :no-change)
     ((eql :no-change lhs) 
      (mk-inter (type-t1 X) rhs  ))
     ((eql :no-change rhs) 
      (mk-inter lhs (type-t2 X)  ))
     (t 
      (mk-inter lhs rhs )))))


;;;    
;;; 4. alpha-subst(c(A1,...,An) = c(alpha-subst(A1),...,alpha-subst(An))
;;;

(defun alpha-subst-cons (X h p eqns)
  (let ((arg-lst  (mapcar #'(lambda (x) (type-alpha-subst-aux x h p eqns)) (cons-args X))))
    (if (every #'(lambda(x) (eql :no-change x)) arg-lst) 
	:no-change
      (mk-cons (cons-id X) 
		 (mapcar #'(lambda (x y) (if (eql x :no-change) y x))
			 arg-lst
			 (cons-args X))))))

;;; 
;;;  5. alpha-subst(X->Y+Z) = alpha-subst(X)-> alpha-subst(Y)+ alpha-subst(Z)
;;;
(defun alpha-subst-func (X h p eqns)
  (let* ((dom (type-alpha-subst-aux (type-t1 X) h p eqns))
	 (range (type-alpha-subst-aux (type-t2 X) h p eqns))
	 (err (type-alpha-subst-aux (type-t3 X) h p eqns)))
    (if (and (eql :no-change dom) 
	     (eql :no-change range)
	     (eql :no-change err))
	:no-change
      (mk-func (if (eql :no-change dom) (type-t1 X) dom)
		 (if (eql :no-change range) (type-t2 X) range)
		 (if (eql :no-change err) (type-t3 X) err)
		 ))))
;;;
;;; 6. free variables
;;;
(defun alpha-subst-var (X h p eqns)
  (funcall h X h p eqns))

;;; 
;;; 7. bound variables
;;;
(defun alpha-subst-bvar (X h p eqns) 
  (funcall h X h p eqns))

;;; 8. alpha-subst(not X) = not X
;;; 
(defun alpha-subst-not (X h p eqns) 
  (declare (ignore X h p eqns))
  :no-change)

;;;     E |- alpha-subst(A) = X, alpha-subst(B) = Y
;;; 9. -------------------------------
;;;     E |- alpha-subst(A ? B) = X?Y
;;;          
(defun alpha-subst-dep (X h p eqns)
  (let ((lhs (type-alpha-subst-aux (type-t1 X) h p eqns))
	(rhs (type-alpha-subst-aux (type-t2 X) h p eqns)))
    (cond 
     ((and (eql :no-change lhs) (eql :no-change rhs)) 
      :no-change)
     ((eql :no-change lhs) 
      (mk-dep (type-t1 X) rhs  ))
     ((eql :no-change rhs) 
      (mk-dep lhs (type-t2 X)  ))
     (t (mk-dep lhs rhs )))))

;;;
;;;  10.  alpha-subst(val v X) = val v alpha-subst(X)
;;;
(defun alpha-subst-value (X h p eqns)
  (let ((arg  (type-alpha-subst-aux (value-typ X) h p eqns)))
    (if (eql :no-change arg)
	:no-change
      (mk-value (value-val X) arg ))))
	
;;;     
;;; Q/L types contain NO free variables, so we define alpha-substitutions to be just the
;;; identity for them.
;;;
;;;      
;;; 11. E |- alpha-subst(for-all x. A) = (for-all x. A)
;;;

(defun alpha-subst-quant (X h p eqns)
  (declare (ignore h p eqns))
  X)

;;;      
;;; 12. E |- alpha-subst(lambda x. A) = (lambda x. A)
;;;

(defun alpha-subst-lambda (X h p eqns)
  (declare (ignore h p eqns))
  X)

