(in-package 'ptools)

;;;
;;; Various Lisp utilities, written by Brian Murphy.
;;;
(export
 '(merge-uniquely		; merge two sorted lists, eliminating dup elts
   uniqify			; destructively remove consecutive dup elts
   sorted-adjoin		; set adjoin on sorted lists
   sorted-diff			; set difference on sorted lists
   sorted-disjunction		; set disjunction on sorted lists
   sorted-intersection		; set intersection on sorted lists
   trans-closure		; transitive closure of a sorted list under f
   list-cross-product		; all pairs of elements from 2 lists
   ))

(defun trans-closure (startset func lttest eqtest)
  (let ((newones (uniqify (sort (mapcan func startset) lttest) eqtest)))
    (let ((difference (sorted-diff newones startset lttest eqtest)))
      (if difference
	  (trans-closure (merge-uniquely startset difference lttest eqtest)
			 func lttest eqtest)
	startset))))

(defun merge-uniquely (l1 l2 lttest eqtest)
  (cond ((null l1) l2)
	((null l2) l1)
	((funcall eqtest (car l1) (car l2))
	 (cons (car l1) (merge-uniquely (cdr l1) (cdr l2) lttest eqtest)))
	((funcall lttest (car l1) (car l2))
	 (cons (car l1) (merge-uniquely (cdr l1) l2 lttest eqtest)))
	('t
	 (cons (car l2) (merge-uniquely l1 (cdr l2) lttest eqtest)))))

(defun sorted-adjoin (x l1 lttest eqtest)
  (cond ((null l1) (list x))
	((funcall lttest (car l1) x)
	 (cons (car l1) (sorted-adjoin x (cdr l1) lttest eqtest)))
	((funcall lttest x (car l1))
	 (cons x l1))
	('t l1)))

;; returns elements in l1 eqtest to an element in l2, 
;; in a sorted list both l1 and l2 must be sorted by lttest
(defun sorted-intersection (l1 l2 lttest eqtest)
  (cond ((or (null l1) (null l2)) '())
	((funcall eqtest (car l1) (car l2))
	 (cons (car l1) (sorted-intersection (cdr l1) (cdr l2) lttest eqtest)))
	((funcall lttest (car l1) (car l2))
	 (sorted-intersection (cdr l1) l2 lttest eqtest))
	('t
	 (sorted-intersection l1 (cdr l2) lttest eqtest))))

;; returns elements not in both l1 and l2 (by eqtest comparison);
;; both l1 and l2 must be sorted by lttest
(defun sorted-disjunction (l1 l2 lttest eqtest)
  (cond ((null l1) l2)
	((null l2) l1)
	((funcall eqtest (car l1) (car l2))
	 (sorted-disjunction (cdr l1) (cdr l2) lttest eqtest))
	((funcall lttest (car l1) (car l2))
	 (cons (car l1) (sorted-disjunction (cdr l1) l2 lttest eqtest)))
	('t
	 (cons (car l2) (sorted-disjunction l1 (cdr l2) lttest eqtest)))))

;; returns elements in l1 but not l2 (by eqtest comparison);
;; both l1 and l2 must be sorted by lttest
(defun sorted-diff (l1 l2 lttest eqtest)
  (cond ((null l1) '())
	((null l2) l1)
	((funcall eqtest (car l1) (car l2))
	 (sorted-diff (cdr l1) (cdr l2) lttest eqtest))
	((funcall lttest (car l1) (car l2))
	 (cons (car l1) (sorted-diff (cdr l1) l2 lttest eqtest)))
	('t
	 (sorted-diff l1 (cdr l2) lttest eqtest))))

;; destroys argument list in the process,
;; eliminates adjacent duplicates
(defun uniqify (l1 pred)
  (cond ((null l1) '())
	('t (uniq2 l1 pred)
	   l1)))
(defun uniq2 (l1 pred)
  (cond ((null l1) l1)
	((null (cdr l1)) l1)
	((funcall pred (car l1) (cadr l1))
	 (rplacd l1 (cddr l1))
	 (uniq2 l1 pred))
	('t (uniq2 (cdr l1) pred))))

(defun list-cross-product (l1 l2)
  (mapcan #'(lambda (item1) (mapcar #'(lambda (item2) (list item1 item2))
				    l2))
	  l1))

