(in-package 'types)

;;;
;;; This file implements type complementation. (type-compl X) represents the  
;;; unique type ~X such that Y & T = 0 iff Y <= ~X
;;; Note also that ~X is the largest type such that X & ~X = 0
;;;

;;; 0. E |- ~0 = 1
;;;

(defun compl-zero(typ eqns) 
  (declare (ignore typ eqns))
  (type-one))

;;; 1. E |- ~1 = 0
;;;

(defun compl-one(typ eqns) 
  (declare (ignore typ eqns))
  (type-zero))


;;;      E |- ~A = X,  ~B = Y
;;; 2. -----------------------------
;;;     E |- ~(A | B) = X * Y
;;;

(defun compl-union (X eqns)
  (type-inter (type-compl-aux (type-t1 X) eqns) (type-compl-aux (type-t2 X) eqns) eqns))

;;;     E |- ^A = X, ^B = Y, ~A = W,  ~B = Z
;;; 3. ---------------------------------------------
;;;    E |-  ~(A * B) = X * Z | Y * W |  W * Z  
;;;
;;; special cases : a. compl(fv * X) = compl(X)
;;;                 b. compl(fn * X) = all-constructions * compl(X)

(defun compl-inter (X eqns)
  (cond ((is-var (type-t1 X))  (type-compl (type-t2 X) eqns))
	((is-func (type-t1 X)) (type-inter (type-all-constructions) 
					   (type-compl (type-t2 X) eqns) 
					   eqns))
	(t (let* (
		  (compl1 (type-compl-aux (type-t1 X) eqns))
		  (compl2 (type-compl-aux (type-t2 X) eqns))
		  )
	     (type-union (type-inter (type-hat-aux (type-t1 X) eqns) 
				     compl2
				     eqns)
			 (type-union (type-inter compl1
						 compl2
						 eqns)
				     (type-inter compl1
						 (type-hat-aux (type-t2 X) eqns) 
						 eqns)
				     eqns :disjoint t)
			 eqns :disjoint t)))))


;;;     E |- ~A1 = X1, ..... ~An = Xn
;;; 4. -----------------------------------------------------------------------------------------------
;;;    E |- ~(c(A1,....An)) = not({c})   UNION 0 -> 1 + 1    UNION {c(X1,1,...1),..... c(1,...,1,Xn)}
;;;                      
;;;     where not({c}) = UNION_{d(1,...1) where d in C - {c}}

(defun compl-cons (X eqns)
  (type-union
   (type-union (type-not X)
	       (type-all-funcs)
	       eqns
	       :disjoint t)
   (constr-negate X eqns)
   eqns
   :disjoint t))

;;;
;;; 5.  E |- ~(X -> Y + Z) = UNION_{c(1,1,...1)} where c in C.
;;;   


(defun compl-func (X eqns)
  (declare (ignore X eqns))
  (type-all-constructions))

;;; 6. E |- ~alpha = 0 (if alpha is a free variable)
;;;

(defun compl-var (X eqns) 
  (declare (ignore X eqns))
  (type-zero))

;;; 7. E, {~a = b}  |- ~a = b (if a is a bound variable)
;;;

(defun compl-bvar (X eqns)  ;;; bv-compl : var * eqns -> var
  (bv-compl X eqns))  ;;; bv-compl gets the variable associated with ~a  
     


;;; 
;;; 8. E |- ~(not {c1,...cn}) = c1(1,...1) UNION ..... UNION cn(1,....1) 
;;;                             UNION (0 -> 1 + 1)
;;; 

(defun compl-not (X eqns) 
  (type-union (type-all-funcs)
	      (constr-ones (not-list X) (not-arities X)) eqns :disjoint t))

;;;      E |- ~A = X
;;; 9. ------------------ if B <> 0  NOTE : this assumption should be validated by the fl front end.
;;;     E |-  ~(A ? B) = X  
;;;

(defun compl-dep (X eqns) 
  (type-compl-aux (type-t1 X) eqns)) 

;;;        E |- ~typ = X
;;; 10. ------------------------------
;;;       E |- ~(val v typ) = X
;;;

(defun compl-value (X eqns)
  (type-compl-aux (value-typ X) eqns))


