(in-package 'types)

;;;
;;; This file implements an algorithm for solving constraints
;;; All the functions needed to solve constraints of the form E |- T1 <= T2
;;; are implemented in this file.


;;; is-inductive : type * var * eqns -> bool
;;; is-inductive returns true if 
;;;     v = a_j   and   TLV(typ) subset of { a_0,...,a_{j-1} }
;;;
;;; NOTE: eqns is needed only for stats reporting done by tlv.
;;;
(defun is-inductive (typ v eqns )
  (let ((vid (var-id v)))        ;; v is a free OR bound-variable
    (every #'(lambda (x) (< (var-id x) vid)) 
	   (type-tlv typ eqns)))) 


;;;   E,C |- 0 <= X
;;;   E,C |- X <= 1
;;;
;;; No constraints need to be generated since the containment is always satisfied

(defun no-c (typ1 typ2 eqns constraints) 
  (declare (ignore typ1 typ2 eqns))
  constraints)

;;; 
;;; If an inconsistency is discovered and
;;; we are in a context where it is OK to have
;;; inconsistent constraints, throw out.
;;;
(defun c-err (x1 x2 eqns constraints) 
  (declare (ignore constraints))
  (if *inconsistent-ok*
      (throw 'inconsistent :inconsistent)
    (progn
      ;; a real bug in the system
      (format t "~% CONSTRAINTS ERROR. The constraint is of the form: ~% ")
      (pt x1 eqns)
      (format t " <= ")
      (pt x2 eqns)
      (error " ~%INCONSISTENT Constraints"))))

;;; union on the left : hi priority rule.
;;;   E, C |- X <= Z,   Y <= Z
;;;   ------------------------
;;;      E, C |- X | Y <= Z
;;;
(defun c-un-l (union any eqns constraints)
  (type-constraints-aux (type-t1 union) 
			any 
			eqns
			(type-constraints-aux (type-t2 union) 
					      any 
					      eqns
					      constraints)))

;;; Special subcase : when the rhs of the constraint is a variable (free/bound)
;;; 
(defun c-un-lT (union a eqns constraints)
  (if (is-inductive union a eqns)
      (constraints-add union a constraints)
    (c-un-l union a eqns constraints)))

;;;
;;; Optimization for constructor on the left and discriminative union on the right.
;;; Actually, union on the right doesn't have to be discriminative; it just
;;; has to be a union of constructions with one occurrence of the constructor
;;; on the left.
;;;
;;; E, C |- c(X1,...,Xn) <= di(...)    if c = di
;;; -----------------------------------------------
;;; E, C |- c(X1,...,Xn) <= d1(...) u ... u dn(...)
;;;


;;;
;;; a complex constructor has the form
;;; v1 & ... & vn & c(...)
;;;
(defun is-complex-constructor (x)
  (cond
   ((is-inter x)
    (is-complex-constructor (type-t2 x)))
   ((is-cons x)
    t)
   (t nil)))

;;;
;;; let c,x be complex constructors.
;;; does c have the same constructor as x?
;;;
(defun same-complex-constructor (c x)
  (cond
   ((is-inter x)
    (same-complex-constructor c (type-t2 x)))
   ((is-inter c)
    (same-complex-constructor (type-t2 c) x))
   (t 
    (= (cons-id c) (cons-id x)))))

;;;
;;; make sure that
;;;  xs is a union of complext constructors and
;;;  cons is not one of those disjuncts
;;;
(defun no-more-disjuncts (cns xs)
  (or
   (and (is-union xs)
	(is-complex-constructor (type-t1 xs))
	(not (same-complex-constructor cns (type-t1 xs)))
	(no-more-disjuncts cns (type-t2 xs)))
   (and (is-complex-constructor xs)
	(not (same-complex-constructor cns xs)))))

;;;
;;; check that the complex constructor cns occurrs once in
;;; the union xs.
;;;
(defun constructor-occurs-once (cns xs)
  (if (is-union xs)
      (if (is-complex-constructor (type-t1 xs))
	  (if (same-complex-constructor cns (type-t1 xs))
	      (if (no-more-disjuncts cns (type-t2 xs))
		  (type-t1 xs))
	      (constructor-occurs-once cns (type-t2 xs))))
      (if (and (is-complex-constructor xs)
	       (same-complex-constructor cns xs))
	  xs)))

(defun c-cl-ur (cns union eqns constraints)
 (let ((ans (constructor-occurs-once cns union)))
   (if ans
       (type-constraints-aux cns ans eqns constraints)
     (c-un-r cns union eqns constraints))))

(defun c-il-ur (cns union eqns constraints)
  (if (is-complex-constructor cns)
      (let ((ans (constructor-occurs-once cns union)))
	(if ans
	    (type-constraints-aux cns ans eqns constraints)
	  (c-in-l cns union eqns constraints)))
    (c-in-l cns union eqns constraints)))

;;; Union on the right is transformed by taking components to the left via complement
;;;
;;; E, C |- X * ~Y <= Z , X * ~Z <= Y
;;; ---------------------------------
;;; E, C |- X <= Y | Z

(defun c-un-r (any union eqns constraints)
  (type-constraints-aux (type-inter any 
				    (type-compl (type-t1 union) 
						eqns) 
				    eqns) 
			(type-t2 union) 
			eqns
			(type-constraints-aux (type-inter any 
							  (type-compl (type-t2 union) eqns) 
							  eqns) 
					      (type-t1 union) 
					     eqns
					     constraints)))

;;; Special subcase : when the lhs of the constraint is a variable (free/bound)
;;; 
(defun c-un-rT (a union eqns constraints)
  (if (is-inductive union a eqns)
      (constraints-add a union constraints)
    (c-un-r a union eqns constraints)))


;;; intersections on the left : are always of the form ai * ^X  <= Y
;;;
;;;                E, C |- X <= Z + ~Y    (disjoint union)
;;;                ---------------------
;;; General rule : E, C |- X <= Z*Y + ~Y
;;;                ---------------------
;;;                E, C |- X * ^Y <= Z   (NOTE ^Y is a hat-type and hence a monotype, X is a free var!)
;;; Specific subcases. 

;;; 1. If Y is 0, c(..), ->, not, i.e., Y is CONTRactive, then
;;;    E, C union {aj <= Y*^X + ~X} |- aj * ^X <= Y SINCE
;;;    fv * TLV(Y * ^X + ~X) = {} <= {a_1,..a_{i-1}}
;;; 

(defun c-in-l (inter contr eqns constraints)
  (assert (is-any-var (type-t1 inter)))  ;; ensure that the intersection on the left is of the form fv * ^X
  (constraints-add (type-t1 inter)
		   (type-union (type-inter contr (type-t2 inter) eqns)
			       (type-compl (type-t2 inter) eqns)
			       eqns 
			       :disjoint t) 
		   constraints))

;;; 2. E, C |- aj * ^X <= ai  where aj is FREE and ai is FREE/BOUND variable. Three subcases
;;;     a. i > j       E, C union {aj * X <= ai} |- aj * X <= ai  since TLV(aj*X) * FV <= {a_1,..a_{i-1}}
;;;     b. i = j       E, C |-  ai * ^X <= ai : no-new-constraints 
;;;     c. i < j       E, C union {aj <= expand(ai * ^X + ~X)} |- aj*^X  <= ai

(defun c-in-l-v (inter a eqns constraints)    ;; NOTE a is a bound/free variable
  (assert (is-any-var a))
  (assert (is-var (type-t1 inter)))  ;; ensure that the intersection on the left is of the form fv * ^X
  (cond ( (> (var-id a) (var-id (type-t1 inter)))
	  (constraints-add inter a constraints))
	( (= (var-id a) (var-id (type-t1 inter))) constraints)
	(t   (c-in-l inter a eqns constraints))))

;;; 
;;; E, C |- Z <= X , Z <= Y
;;; -----------------------
;;; E, C |- Z <= X * Y

(defun c-in-r (any inter eqns constraints)
  (type-constraints-aux any 
			(type-t1 inter) 
			eqns
			(type-constraints-aux any 
					      (type-t2 inter) 
					      eqns
					      constraints)))


;;; Special subcase : when the lhs of the constraint is a variable (free/bound)
;;; 
(defun c-in-rT (a inter eqns constraints)
  (assert (is-any-var a))
  (if (is-inductive inter a eqns)
      (constraints-add a inter constraints)
    (c-in-r a inter eqns constraints)))


;;;
;;;       E, C |- X <= 0
;;;  -------------------------
;;;  E, C |- c(...,X,...) <= 0
;;;
;;;  Rule is non-deterministic; choose first component of c
;;;  that can be made zero.
;;;
(defun choose-one-can-be-zero (ts eqns)
  (cond
   ((null ts) 
    (type-one))
   ((not (type-not-zero-for-sure (car ts) eqns))
    (car ts))
   (t (choose-one-can-be-zero (cdr ts) eqns))))

(defun c-zero (c zero eqns constraints)
  (type-constraints-aux 
   (choose-one-can-be-zero (cons-args c) eqns) 
   zero 
   eqns 
   constraints))


;;;
;;;         E, C |- for all i, X_i <= Y_i
;;;     ----------------------------------------
;;;      E, C |- c(X_1,...,X_n) <= c(Y_1,...,Y_n)
;;;
(defun c-c (c1 c2 eqns constraints) ;; tkl 12/2/92 bug fix
  (when (not (= (cons-id c1) (cons-id c2)))
	(c-err c1 c2 eqns constraints))
  (cond ((null (cons-args c1)) constraints)
	(t (my-reduce-cons (cons-args c1) (cons-args c2) eqns constraints))))

;;; aux function composes constraints generated by (type-constraints-aux x y eqns c) 
;;; for x in l1, y in l2

(defun my-reduce-cons (l1 l2 eqns constraints)
  (if (null l1) 
      constraints
    (type-constraints-aux (car l1) 
			  (car l2) 
			  eqns 
			  (my-reduce-cons (cdr l1) 
					  (cdr l2)
					  eqns
					  constraints))))

;;;
;;;  E , C |- c(X1,...Xn) <= not (B)    if c not-in B
;;;           unsatisfiable             otherwise


(defun c-n (c n eqns constraints)
  (if (member (cons-id c) (not-list n))       
      (c-err c n eqns constraints)
    constraints))

;;;
;;;      E, C |- Y_1 <= X_1, X_2 <= Y_2, X_3 <= Y_3
;;;     ---------------------------------------------
;;;      E, C |- X_1 -> X_2 + X_3 <= Y_1 -> Y_2 + Y_3
;;;
;;;      E, C |- T_1 <= D -> R + E   (no constraints if D ~= 0)
;;;
(defun fun-fun (f1 f2 eqns constraints) ;; tkl 1/21/93 bug fix
  (let ((ran2 (type-t2 f2))
	(err2 (type-t3 f2)))
    (if (and (is-one ran2) (is-one err2)) constraints ;; fast optimization
      (type-constraints-aux (type-t1 f2) ;; antimonotonic in domain
			    (type-t1 f1) 
			    eqns     
			    (type-constraints-aux (type-t2 f1) ;; monotonic in range,error
						  (type-t2 f2) 
						  eqns     
						  (type-constraints-aux (type-t3 f1) 
									(type-t3 f2) 
									eqns 
									constraints))))))



;;;
;;; E, C union {a <= X} |- a <= X if X is 1, c(..), X1 -> X2+X3, not(C)     
;;; where a is a bound or a free variable.
;;;

(defun mk-v-l (var any eqns constraints)
  (declare (ignore eqns))
  (assert (is-any-var var))
  (constraints-add var any constraints))

;;; 
;;; E, C union {X <= a} |- X <= a if X is 1, c(..), X1 -> X2+X3, not(C)     
;;; where a is a bound or a free variable.
;;;

(defun mk-v-r (any var eqns constraints)
  (declare (ignore eqns))
  (assert (is-any-var var))
  (constraints-add any var constraints))

;;; E, C |- ai <= aj    if ai = aj (for ai aj FREE/BOUND vars)
;;; E, C union {ai <= aj} |- ai <= aj  (o/w)
;;;

(defun v-v (v1 v2 eqns constraints)
  (declare (ignore eqns))
  (assert (is-any-var v1))
  (assert (is-any-var v2))
  (if (= (var-id v1) (var-id v2)) constraints
    (constraints-add v1 v2 constraints)))

;;; 
;;; E, C |- X <= Z
;;; --------------------- if Z is not a b/f var
;;; E, C |- val(v,X) <= Z

(defun c-val-l (val any eqns constraints)
  (type-constraints-aux (value-typ val) any eqns constraints))

;;;
;;; a. E, C union {val(v,X) <= ai} |- val(v,X) <= ai    if TLV(X) * FV <= {a_1,...a_{i-1}}
;;; 

;;;    E, C |- X <= Z
;;;b. ---------------------      o/w
;;;    E, C |- val(v,X) <= Z

(defun c-val-lT (val a eqns constraints) ;; NOTE a is a f/b variable
  (assert (is-any-var a))
  (if (is-inductive val a eqns)
      (constraints-add val a constraints)
    (type-constraints-aux (value-typ val) a eqns constraints)))



;;; lhs is neither * nor free/bound var.
;;;
;;; E, C |- X <= Y
;;; --------------------   X not b/f var and X not an * type
;;; E, C |- X <= val(v,Y)

(defun c-val-r (any val eqns constraints)
  (type-constraints-aux any (value-typ val) eqns constraints))


;;; lhs is a free/bound variable
;;;
;;; a. E, C union {ai <= val(v,X)} |- ai <= val(v,X)    if TLV(X) * FV <= {a_1,...a_{i-1}}
;;; 

;;;    E, C |- X <= Z
;;;b. ---------------------      o/w
;;;    E, C |- X <= val(v,Z)

(defun c-val-rT (a val eqns constraints)        ;;; NOTE a is a f/b var.
  (assert (is-any-var a))
  (if (is-inductive val a eqns)
      (constraints-add a val constraints)
      (type-constraints-aux a (value-typ val) eqns constraints))) ;; tkl 11/22/92 bug fix change order of args.


;;; lhs is an intersection type 
;;; a. E, C union {ai <= val(v,Z) * ^X + ~X} |- ai * ^X <= val(v,Z)    if TLV(Z) * FV <= {a_1,...a_{i-1}}
;;; 

;;;    E, C |- ai * X <= Z
;;;b. --------------------------      o/w
;;;    E, C |- ai * X <= val(v,Z)

(defun c-val-rTi (any val eqns constraints)   ;;; NOTE any is an intersection ai * ^X
                                              ;;; ^X is a hat-type and thus contains no var.
  (assert (is-any-var (type-t1 any)))
  (if (is-inductive val (type-t1 any) eqns)
      (c-in-l any val eqns constraints)
      (type-constraints-aux any (value-typ val) eqns constraints))) ;; tkl 11/22/92 bug fix change order of args.


;;;
;;; E, C |- X <= 0 or Y <= 0
;;; ------------------------
;;;      X ? Y <= 0
;;;
;;; It must be the case that Y <= 0 is solvable,
;;; because no ? type is built unless the right-hand
;;; type can be instantiated to 0 in some substitution.
;;;
(defun c-dep-z (dep zero eqns constraints)
  (if (not (type-not-zero-for-sure (type-t1 dep) eqns))
      (type-constraints-aux (type-t1 dep) zero eqns constraints)
    (type-constraints-aux (type-t2 dep) zero eqns constraints)))

;;;
;;; E, C |- X <= Z
;;; ---------------- if Z is not a  var
;;; E, C |- X?Y <= Z
;;;
;;; E, C |- X?Y <= 0
;;; ----------------   used in the case where X & Z = 0
;;; E, C |- X?Y <= Z

(defun c-dep-l (dep any eqns constraints)
  (if (type-disjoint (type-t1 dep) any eqns)
      (type-constraints-aux dep (type-zero) eqns constraints)
    (type-constraints-aux (type-t1 dep) any eqns constraints)))

;;; a. E, C union { (X?Y) <= aj} |- (X ? Y) <= aj       if TLV(X) * FV <= {a_1,..,a_{j-1}}
;;;
;;;    E, C |- X <= Z
;;; b. ----------------            o/w
;;;    E, C |- X?Y <= Z

(defun c-dep-lT (dep a eqns constraints)      ;; NOTE : a is a f/b variable
  (assert (is-any-var a))
  (if (is-inductive dep a eqns)
      (constraints-add dep a constraints)
      (type-constraints-aux (type-t1 dep) a eqns constraints)))


;;; E, A |- NOT(B1) <= NOT(B2)  if B2 <= B1
;;;         not satisfiable     otherwise
;;;

(defun n-n (n1 n2 eqns constraints )
  (if (subsetp (not-list n2) (not-list n1) :test #'equal)
      constraints
    (c-err n1 n2 eqns constraints)))

