;;;
;;; Basic definitions for the Illyria type system
;;;
(in-package 'types)

(export '(union-type
	  inter-type
	  cons-type
	  func-type
	  value-type
	  zero-type
	  one-type
	  var-type
	  dep-type
	  not-type
	  quant-type
	  lambda-type
	  tn
	  ))

(proclaim '(inline is-cons cons-id cons-args 
		   is-func 
		   var-id mk-var is-var
		   mk-bvar is-bvar bvar-id))
		   
;;;
;;; The categories of types
;;;

(defconstant zero-type 0)
(defconstant one-type 1)
(defconstant union-type 2)
(defconstant inter-type 3)
(defconstant cons-type 4)
(defconstant func-type 5)
(defconstant var-type 6)
(defconstant boundvar-type 7)
(defconstant not-type 8)
(defconstant dep-type 9)
(defconstant value-type 10)
(defconstant quant-type 11)
(defconstant lambda-type 12)
	  
(defconstant type-name-list
  `((,zero-type "zero")
    (,one-type "one")
    (,union-type "union type")
    (,inter-type "intersection type")
    (,cons-type "constructor type")
    (,func-type "function type")
    (,var-type "(free) variable type")
    (,boundvar-type "(bound) variable type")
    (,not-type "NOT type")
    (,dep-type "dependent type")
    (,value-type "value type")
    (,quant-type "quantified type")
    (,lambda-type "lambda type (a type function)")
    ))

(defun tn (&optional arg)
  ;; tn stands for "Type Name"
  ;; tn returns a string that describes what class of types
  ;; the argument represents.
  ;; No arguments dumps the mapping of type numbers to descriptions.
  (let ((ans (member arg type-name-list :key #'car)))
    (if ans
	(second (first ans))
      (pprint type-name-list))))


;;;
;;; bit patterns for tagging R, L, and hatted types
;;;   - an R type is one that is permitted to appear 
;;;         as an upper bound (on the rhs) of an inclusion constraint
;;;   - an L type is one permitted to appear as a lower bound
;;;         (on the lhs) of an inclusion constraint
;;;   - a "hat" type is an upward-closed, L, and R type with no free
;;;         variables.  The name "hat" refers to the operation ^a which
;;;         is converts type "a" to a hat type.
;;;
;;; nsiu: a bit pattern that if set ensures that subset tests are NOT performed in unions
;;;
(defconstant *r* 1)       ;; bit pattern 0001
(defconstant *l* 2)       ;; bit pattern 0010
(defconstant *hat* 4)     ;; bit pattern 0100
(defconstant *ns* 8)      ;; bit pattern 1000 no-subset-in-union (relevant only for unions)
(defconstant *notr* (lognot *r*))
(defconstant *notl* (lognot *l*))
(defconstant *nothat* (lognot *hat*))
(defconstant *rl* (logior *r* *l*))
(defconstant *notrl* (lognot *rl*))
(defconstant *rlhat* (logior *r* *l* *hat* ))
(defconstant *notrlhat* (lognot *rlhat*))

;;;
;;; A type is a record with five slots.  The usage of the fields t1-t3
;;; depends on the type field.
;;; 
;;; Type            Use of t1-t3
;;; zero              unused
;;; one               unused
;;; union             t1,t2 are types
;;; intersection         "   "   "
;;; constructor       t1 is a constructor id, t2 is a list of types
;;; function          t1-t3 are types
;;; variable          t1 is variable id
;;; bound variable    t1 is variable id
;;; not               t1 is a list of constructor ids, 
;;;                   t2 is a list of constructor arities
;;; dependent         t1,t2 are types
;;; value             t1 is an FL value, t2 is a type
;;; quantified        t1 is a list of FREE variables, t2 is a list of BOUND variables, t3 is a type
;;; lambda            "   "  "      "   "        "  "  "   "  "   "  "  "  "  "  "  "  "  "  "   "
;;;
;;; The fifth slot rl keeps track of whether a type is an 
;;; R type, an L type, and a "hatted" type (and whether subset tests should be performed in unions
;;;

(defstruct (typ (:predicate is-type) 
		(:conc-name type-) 
		(:constructor make-type)
		(:copier copy-type))
  (type 0 :type fixnum)
  t1
  t2
  t3
  (rl *notrlhat* :type fixnum))


(defun mk-type (type t1 &optional (t2 nil) (t3 nil) (rl *notrlhat*))
  (make-type :type type :t1 t1 :t2 t2 :t3 t3 :rl rl))


;;;
;;; Predicates for determining whether a type is an R type, an L type,
;;; or both
;;;
(defun is-rtype (type) (= (logand (type-rl type) *r*) *r*))
(defun is-ltype (type) (= (logand (type-rl type) *l*) *l*))
(defun is-rltype (type) (= (logand (type-rl type) *rl*) *rl*))
(defun is-hattype (type) (= (logand (type-rl type) *hat*) *hat*))
(defun is-nsunion (type) (= (logand (type-rl type) *ns*) *ns*))
(defun is-rlhattype (type) (= (logand (type-rl type) *rlhat*) *rlhat*))

;;;
;;; swap-rl returns an L (R) type bit pattern if its argument is an R (L) type
;;; The hat bit returned is 0.
;;;
(defun swap-rl (type) 
  (cond
   ((is-rltype type) *rl*)
   ((is-rtype type) *l*)
   ((is-ltype type) *r*)
   (t *notrlhat*)))

;;;
;;; rlh-and returns an rlh description with the 
;;; properties of both argument types
;;;

(defun rlh-and (t1 t2)
  (logand (type-rl t1) (type-rl t2)))

;;;=============================================================
;;;
;;; mk- and is- functions for all type categories
;;;
;;;=============================================================


(defun mk-zero () (make-type :type zero-type :rl *rlhat*))
(defun is-zero (x) (and (is-type x) (= (type-type x) zero-type)))

(defun mk-one () (make-type :type one-type :rl *rlhat*))
(defun is-one (x) (and (is-type x) (= (type-type x) one-type)))


;;;
;;; The flag disjoint should be true if X & Y = 0.
;;; X | Y is an R type if X is an R type and Y is an R type and X&Y = 0
;;; X | Y is an L type if X is an L type and Y is an L type
;;; ^(X | Y) = ^X | ^Y
;;;
;;; Note that if x & y = 0, it might be because one of them is 0.
;;;
(defun mk-union (x y &key (disjoint nil) (nosubset nil))
  (cond
   ((is-zero x) y)
   ((is-zero y) x)
   (t
    (let ((rl (cond (disjoint (logior (logand (type-rl x) (type-rl y)) *ns*))
		    (nosubset (logior (logand (type-rl x) (type-rl y) *notr*) *ns*))
		    (t (logand (type-rl x) (type-rl y) *notr*)))))
      (make-type :type union-type :t1 x :t2 y :rl rl)))))

(defun is-union (x) (and (is-type x) (= (type-type x) union-type)))



;;;
;;; X & Y is an R type is both X and Y are R types
;;; X & Y is an L type if X is an L type and Y is an L type and Y = ^Y
;;; ^(X & Y) = ^X & ^Y
;;; X & 0 = 0
;;; 0 & Y = 0
;;;
(defun mk-inter (x y)
  (if (or (is-zero x) (is-zero y))
      (type-zero)
    (let ((rl (if (is-hattype y)
		  (logand (type-rl x) (type-rl y))
	        (logand (type-rl x) (type-rl y) *notl*))))
      (make-type :type inter-type :t1 x :t2 y :rl rl))))

(defun is-inter (x) (and (is-type x) (= (type-type x) inter-type)))


;;;
;;; c(X_1,...,X_n) is an R (L) type if for all i, X_i is an R (L) type
;;; ^c(X_1,...,X_n) = c(^X_1,...,^X_n)
;;;
(defun reduce-rl (l)
  (if (null l)
      -1
    (logand (type-rl (car l))
	    (reduce-rl (cdr l)))))

(defun mk-cons (consid listoftypes)
  (make-type :type cons-type :t1 consid :t2 listoftypes
	     :rl (reduce-rl listoftypes)))
(defun is-cons (x) (and (is-type x) (= (type-type x) cons-type)))
(defun cons-id (x) (type-t1 x))
(defun cons-args (x) (type-t2 x))

  
;;;
;;; A function type is an R type if it has the form L -> R; R
;;; A function type is an L type if it has the form R -> L; L
;;; A function type ^(X -> Y + Z) = 0 -> 1;1
;;; For efficiency, this algorithm is conservative and simply
;;; assumes functions types are not upward-closed.  The function
;;; type-all-funcs (in mktype.lisp) does return a 0(1) -> 1;1 type  1/21/93
;;; that is marked as a hat type. 
;;;
(defun mk-func (x y z) 
  (make-type :type func-type :t1 x :t2 y :t3 z
	     :rl (logand (swap-rl x) (type-rl y) (type-rl z))))

(defun is-func (x) (and (is-type x) (= (type-type x) func-type)))


;;;
;;; Free and bound variables are both R and L types, but not necessarily
;;; hat types.
;;;
(defun mk-var (x) (make-type :type var-type :t1 x :rl *rl*))
(defun is-var (x) (and (is-type x) (= (type-type x) var-type)))
(defun var-id (x) (type-t1 x))

(defun mk-bvar (x &key (rl *rl*)) 
  (make-type :type boundvar-type :t1 x :rl rl))
(defun is-bvar (x) (and (is-type x) (= (type-type x) boundvar-type)))

;; note var-id should be maintained so that it works for 
;; both free and bound variables.
(defun bvar-id (x) (type-t1 x))

(defun is-any-var (x)
  (or (is-var x) (is-bvar x)))

(defun var-=  (v1 v2) (=  (type-t1 v1) (type-t1 v2)))
(defun var-<= (v1 v2) (<= (type-t1 v1) (type-t1 v2)))

;;;
;;; X ? Y is an L type iff X is an L type.
;;; X ? Y is never an R type or a hat type
;;;
(defun mk-dep (x y) 
  (cond
   ((is-zero y) (type-zero)) ;; X?0 = 0
   ((is-hattype y) x)        ;; X?^Y = X if ^Y \not= 0
   (t (make-type :type dep-type :t1 x :t2 y :rl (logand (type-rl x) *l*)))))

(defun is-dep (x) (and (is-type x) (= (type-type x) dep-type)))

;;;
;;; A value (v,X) is an L,R,hat type iff X is an L,R,hat type.
;;;
(defun mk-value (val typ) 
  (make-type :type value-type :t1 val :t2 typ :rl (type-rl typ)))
(defun is-value (x) (and (is-type x) (= (type-type x) value-type)))
(defun value-val (x) (type-t1 x))
(defun value-typ (x) (type-t2 x))

;;;
;;; A not type is always an R,L, and hat type.
;;;
(defun mk-not (list-of-consid list-of-arities) 
  (make-type :type not-type 
	     :t1 list-of-consid 
	     :t2 list-of-arities
	     :rl *rlhat*))
(defun is-not (x) (and (is-type x) (= (type-type x) not-type)))
(defun not-list (x) (type-t1 x))
(defun not-arities (x) (type-t2 x))   ;; for each constructor, keep track of its arity.

;;; not-combine :: not-type * not-type --> not-type
;;;

(defun not-combine (not1 not2)
  (make-type :type not-type
	     :t1 (append (not-list not1) (not-list not2))
	     :t2 (append (not-arities not1) (not-arities not2))
	     :rl *rlhat*))

;;;
;;; Quantified types and type functions are not R,L, or hat types.
;;;
(defun mk-quant (vars bvars typ) 
  (make-type :type quant-type :t1 vars :t2 bvars :t3 typ))
(defun is-quant (x) (and (is-type x) (= (type-type x) quant-type)))
(defun quant-vars (x) (type-t1 x))
(defun quant-bvars (x) (type-t2 x))
(defun quant-typ  (x) (type-t3 x))


(defun mk-lambda (vars bvars typ)
  (make-type :type lambda-type :t1 vars :t2 bvars :t3 typ))
(defun is-lambda (x) (and (is-type x) (= (type-type x) lambda-type)))
(defun lambda-vars (x) (type-t1 x))
(defun lambda-bvars (x) (type-t2 x))
(defun lambda-typ (x) (type-t3 x))

;;; Since type-ql functions are implemented as GENERIC in mktype.lisp
;;; hence we need Generic functions (that work on BOTH quant AND lambda types)
;;;

(defun ql-free-vars (x) (type-t1 x))
(defun ql-bound-vars (x) (type-t2 x))
(defun ql-type (x) (type-t3 x))


(defun install-type-memo (x m)
  (if (is-func x)
      nil
      (setf (type-t3 x) m)))

(defun type-memo (x)
  (if (is-func x) 
      nil
    (type-t3 x)))
