(in-package 'types)

;;;
;;; This file implements a conservative algorithm for testing
;;; whether one type is disjoint from another type
;;;

;;; Functions that just return true and false.
;;;
(defun isdisj (&rest x) (declare (ignore x)) t)
(defun notdisj (&rest x) (declare (ignore x)) nil)

;;;
;;; note : 1. E,A |- disj(0,X) = true (i.e., use isdisj)
;;;        2. E,A |- disj(c(..), A -> B + D) = true (and the dual case)
;;;        3. E,A |- disj(A -> B + C,not(D)) = true (and the dual case)

;;; E,A |- disj(A,X) , disj(B,X)
;;; ----------------------------
;;; E,A |- disj(A|B,X)

(defun udisj (union B eqns as)
  (and (type-disjoint-aux (type-t1 union) B eqns as)
       (type-disjoint-aux (type-t2 union) B eqns as)))

;;; E,A |- disj(X,A), disj(X,B)
;;; --------------------------
;;; E,A |- disj(X,A|B)
;;;

(defun disju (any union eqns as)
  (and (type-disjoint-aux any (type-t1 union) eqns as)
       (type-disjoint-aux any (type-t2 union) eqns as)))


;;; E,A |- disj(A,X) , disj(B,X)
;;; ----------------------------  This rule is conservative
;;; E,A |- disj(A*B,X)

(defun indisj (inter any eqns as)    
  (and (type-disjoint-aux (type-t1 inter) any eqns as)
       (type-disjoint-aux (type-t2 inter) any eqns as)))

;;; E,A |- disj(X,A), disj(X,B)
;;; --------------------------    This rule is conservative
;;; E,A |- disj(X,A*B)
;;;

(defun disjin (any inter eqns as)
  (and (type-disjoint-aux any (type-t1 inter) eqns as)
       (type-disjoint-aux any (type-t2 inter) eqns as)))



;;; a. E,A |- disj(c(...),d(...)) if c <> d
;;;
;;;         E, A |- exists i : 1 <= i <= n, disj(X_i,Y_i)
;;; b.  -------------------------------------------------
;;;      E, A |- disj(c(X_1,...,X_n),c(Y_1,...,Y_n))
;;;
(defun consdisj-aux (l1 l2 eqns as)
  (if (null l1)
      nil
    (or
     (type-disjoint-aux (car l1) (car l2) eqns as)
     (consdisj-aux (cdr l1) (cdr l2) eqns as))))

(defun consdisj (c1 c2 eqns as)
  (or (/= (cons-id c1) (cons-id c2))
      (consdisj-aux (cons-args c1) (cons-args c2) eqns as)))

;;; E,A |- disj(c(..),not(B)) if c \in B
;;; 
(defun cndisj (c B eqns as)
  (declare (ignore eqns as))
  (member (cons-id c) (not-list B)))

;;; E,A |- disj(not(B),c(..)) if c \in B
;;;
(defun ncdisj (B c eqns as)
  (declare (ignore eqns as))  
  (member (cons-id c) (not-list B)))



;;; E, A union {disj(b, X)} |- disj(b,X)   where b is a bound var
;;; E, A union {disj(X, b)} |- disj(X,b)   where b is a bound var
;;;
;;; E union {b = X}, A union {disj(b, Y)} |- disj(X,Y)
;;; --------------------------------------------------
;;; E union {B = X}, A |- disj(b,Y)
;;;
;;;
(defun bvdisj (a any eqns as)
  (let ((bvid (var-id a)))
    (or (pair-assump-has bvid any as)
	(let ((bvd (bv-def a eqns)))
	  (if bvd (type-disjoint-aux bvd any eqns (pair-assump-add bvid any as)))))))
  
(defun disjbv (any a eqns as)
  (bvdisj a any eqns as))

;;;
;;;      E, A |- disj(X,Z)
;;;     ----------------------
;;;      E, A |- disj(X ? Y,Z)
;;;
;;; conservative
;;;
(defun ldepdisj (dep any eqns as)
  (type-disjoint-aux (type-t1 dep) any eqns as))

;;; E, A |- disj(Z,X)
;;; ------------------
;;; E, A |- disj(Z,X?Y)

(defun rdepdisj (any dep eqns as)
  (type-disjoint-aux any (type-t1 dep) eqns as))


;;;   E, A |-  disj(X, Y)
;;; ------------------------------
;;;   E, A |- disj(<value, X >, Y)
;;;
;;; conservative
;;;
(defun lvaldisj (v1 any eqns as)
  (type-disjoint-aux (value-typ v1) any eqns as))

;;;   E, A |-  disj(X, Y)
;;; ------------------------------
;;;   E, A |- disj(X, <value, Y >)
;;;
;;; conservative
;;;
(defun rvaldisj (any v1 eqns as)
  (type-disjoint-aux any (value-typ v1) eqns as))

