(in-package 'types)

;;;
;;; This file implements a conservative type (non) emptiness checking (i.e. checking if X = 0)
;;; Note : ^X = 0 => X = 0 (i.e., emptiness checking is based on checking whether ^X = 0)
;;; 
;;; Functions in this file are called from the top level type-non-empty_once function
;;; which is in a separate file "empty_tbl.lisp" (this is done so as to allow
;;; the lisp compiler to define the functions (here) in the right order.
;;; 

;;; E |- X : true  ==> X is non-empty
;;; E |- X : false  ==> X is empty

;;; 0. A |- non-emp(0) = '()
;;;

(defun non-empty-zero(typ as eqns) 
  (declare (ignore typ as eqns))
  ())

;;; 1. A |- non-emp(1) = true
;;;

(defun non-empty-one(typ as eqns) 
  (declare (ignore typ as eqns))
  t)

;;;    A |- non-emp(X) 
;;; 2a ---------------
;;;    A |- non-emp(X|Y) 

;;;     A |- non-emp(Y) 
;;; 2b -----------------
;;;    A |- non-emp(X|Y) 
;;;

(defun non-empty-union (X as eqns)
  (or (type-non-empty-once (type-t1 X) as eqns) 
      (type-non-empty-once (type-t2 X) as eqns)))

;;;       A |- non-emp(B)
;;; 3. ----------------
;;;    A |- non-emp(A & B) 
;;;  
;;;  This rule is correct because intersections
;;;  can have only one of the following forms
;;;  after normalization:
;;;
;;;      a_1 & ... & a_n & b     for free a_i and bound b
;;;
;;;      a_1 & ... & a_n & (b_1 -> c_1 + d_1) & ... & (b_m -> c_m + d_m) 
;;;                              for free a_i and some function types 1..m
;;;
(defun non-empty-inter (X as eqns)
  (type-non-empty-once (type-t2 X) as eqns))

;;;    A |- non-emp(A1) = 1,..... non-emp(An) 
;;; 4. -------------------------------------
;;;    A |- non-emp(c(A1,,,,An)) 
;;;

(defun non-empty-cons (X as eqns)
  (every #'(lambda (arg) (type-non-empty-once arg as eqns)) (cons-args X)))


;;; 
;;; 5. A |- non-emp(X->Y+Z) 
;;;

(defun non-empty-func (X as eqns)
  (declare (ignore X as eqns))
  t)

;;; 6. A |- non-emp(alpha)  (for all free variables alpha)
;;;

(defun non-empty-var (X as eqns)
  (declare (ignore X as eqns))
  t)

;;; 7. A |- non-emp(alpha) = 1 iff (member alpha A)  (for all alpha bound)
;;;

(defun non-empty-bvar (X as eqns) 
  (declare (ignore eqns))
  (not (member (bvar-id X) as)))  ;; Elements of as are assumed to be empty in this iteration
                                  ;; Hence, all bound-vars that are NOT in as are taken to be non-empty
;;; 8. A |- non-emp(not X) 
;;; 

(defun non-empty-not (X as eqns)  
  (declare (ignore X as eqns))
  t)

;;;     A |- non-emp(A), non-emp(B)                       
;;; 9. --------------------------
;;;     A |- non-emp(A ? B) 
;;;          
;;;			              

(defun non-empty-dep (X as eqns)
  (and (type-non-empty-once (type-t1 X) as eqns) 
       (type-non-empty-once (type-t2 X) as eqns)))



;;;      A |- non-emp(X)
;;; 10. -----------------------------
;;;     A |- non-emp(val v X) 
;;;

(defun non-empty-value (X as eqns)
  (type-non-empty-once (value-typ X) as eqns))
  

