(in-package 'types)

;;; 
;;; This file is the continuation of empty_fns.lisp
;;; This is where the functions defined in empty_fns.lisp get called 
;;;
;;; type-empty :: Eqns ->  Eqns (where all empty types are replaced by 0
;;;
;;; type-empty-once : type * eqns ->       ()fn -> True/False
;;;

;;; 10/27/92 tkl modified: now type-empty needs to take care of the fact that
;;; boundvar eqns may not be inductive (i.e., the rhs may contain top level vars)
;;; These variables are FIRST set to 1 by type-subst-tlv and then the type-empty
;;; algorithm is invoked.


;;;
;;; The lookup table

(defconstant *emptyarray*
  (make-array '(11)
	      :initial-contents
	      (list #'non-empty-zero #'non-empty-one #'non-empty-union #'non-empty-inter #'non-empty-cons #'non-empty-func 
		    #'non-empty-var #'non-empty-bvar #'non-empty-not #'non-empty-dep #'non-empty-value )))
	      
;;; The auxiliary emptying function : applies the appropriate function once
;;; in the iterative fix point computation.

(defun type-non-empty-once (typ as tbl) 
  (funcall (aref *emptyarray* (type-type typ)) typ as tbl))
  

;;;
;;; type-empty-aux takes a list of todo-eqns and returns a list
;;; of variables that are empty.
;;;
;;; 
;;; todo-eqns has the form listof(var.type)   
;;; where var = type is a bound variable equation.

;;; Standard fixed-point computation.
;;;
;;; a. Initially for all b-vars v  on lhs of todo-eqns (v.e) let v=0
;;; b. At each iteration, for each equation (v.e)  if e is non-empty, set v=1 and continue.
;;; c. Stop when the set of bound variables that are zero in an iteration is the same as the set in
;;;    the previous iteration.

;;;
(defun type-empty-aux (todo-eqns tbl)
  ;; iter has signature assumptions -> assumptions
  (flet ((iter (as)
	       (reduce
		#'(lambda (as bv-eqn)
		    (if (type-non-empty-once (cdr bv-eqn) as tbl)
			(remove (var-id (car bv-eqn)) as)
		      as))
		todo-eqns
		:initial-value as)))
	(do* ((asold (mapcar #'(lambda (el) (var-id (car el))) todo-eqns) asnew)
	      (asnew (iter asold) (iter asold)))
	     ((subsetp asold asnew) 
	      (mapcar #'(lambda (vid) (mk-bvar vid :rl *rlhat*)) asnew)))))


;; apply substitution to each member of todo-bv-eqns;
;; note that this should not generate any new
;; bv equations, since we re just substituting in 0.
(defun zero-bveqns (empty-bvs todo-eqns tbl)
  (let ((substn (uniform-subst empty-bvs (type-zero))))
    (mapc #'(lambda (eqn) 
	      (bv-replace-def 
	       (car eqn) 
	       (type-substitute (cdr eqn) substn tbl) 
	       tbl))
	  todo-eqns)))

;;;
;;; remove top-level variables from bv equations
;;;
(defun expand-bveqns (todo-eqns tbl)
  (mapc #'(lambda (eqn)
		    (bv-replace-def
		     (car eqn)
		     (type-expand (bv-def (car eqn) tbl) 0 tbl) 
		     tbl))
		todo-eqns))

;;;
;;; For each bound variable equation a = E, compute whether
;;; it is recursive in a, and then inline non-recursive
;;; equations where they are used.  
;;;
(defun inline-nonrecursive-bvs (todo-eqns tbl)
  (mapc #'(lambda (eqn)
	    (let ((v (car eqn)))
	      (if (member v (type-vars-reachable (bv-def v tbl) tbl) :test #'var-=)
		  (bv-rec v tbl)
		(bv-nonrec v tbl))))
	todo-eqns)
  (mapc #'(lambda (eqn)
	    (bv-replace-def
	     (car eqn)
	     (type-simplify (bv-def (car eqn) tbl) tbl :rec t)
	     tbl))
	todo-eqns))
    
;;;
;;; The top level emptying function. Calls the type-empty-aux function
;;;
(defun type-empty (tbl)  
  (cond ((no-undef-bvars tbl)
	 (let* ((todo-eqns (todo-get tbl))
		(empty-vars (type-empty-aux todo-eqns tbl)))
	   ;; we are processing these equations,
	   ;; so don't let anything else see them.
	   (todo-clear tbl)
	   (zero-bveqns empty-vars todo-eqns tbl)
	   (assert (null (todo-get tbl)))  ;; should be true . . .
	   (inline-nonrecursive-bvs todo-eqns tbl)
	   (expand-bveqns todo-eqns tbl)   ;; 1/8/93
	  tbl))
	(t tbl)))




	  
	
