(in-package 'types)

;;;
;;; This file implements the top-level data structure for type inference.
;;; The structure consists of:
;;;
;;;    bv-eqns        equations a = E 
;;;
;;;    bv-recursive   memos as to whether equation a = E is 
;;;                   recursive in a.
;;;
;;;    todo-bv-eqns   equations a = E  (^E might be 0)
;;;                           where a is a bound variable,
;;;                                 E is a type
;;;          NOTE: todo-bv-eqns is a subset of bv-eqns.  
;;;          If a = E \in bv-eqns - todo-bv-eqns, then ^E \not= 0
;;;
;;;    hat-eqns       equations between bound variables ^a = b
;;;
;;;    compl-eqns     equations between bound variables ~a = b
;;;
;;;    inter-eqns     equations of the form  a & X = b
;;;                           where a,b are bound variables, 
;;;                                 X is an R and L type
;;;    bv-subset      constraints  bv <= X1,...,bv <= Xn  where
;;;                   bv = Y & X1, Y = X2 & Z, Z = ... are inter equations
;;;                   Needed to guarantee the termination of the intersection
;;;                   algorithm.
;;;
;;;    upper-bounds   constraints a <= X  where a is any variable, X an R type
;;;    lower-bounds   constraints X <= a  where a is any variable, X an L type
;;;
;;;    undef-bvars    a list of bound variables that have no definitions in bv-eqns
;;;
;;;    pending        hat equations ^a = b, compl equations ~a = b,
;;;                   and intersection equations a&X = b where there
;;;                   is (as yet) no equation a = E in bv-eqns.
;;;
;;;    todo-constraints constraints X <= Y that have yet to be reduced
;;;                   to inductive form.
;;;
;;;    named-types    Hash Table: for every type-def (name $ type) records name and type
;;;
;;;    constr-name    Hash Table: for every constr, maps its id to the constr-name
;;;
;;;    constr-id      Hash Table: for every constr, maps its name to its (negative) id
;;;                   (inverse of constr-name)
;;;
;;;    constr-arity   Hash Table: for every constr, maps its (negative) id to its arity.
;;;

(defvar *itable-init* 500)

(defstruct (tbl (:predicate is-eqns))
  (var 1)             ;; Keeps track of the next var-id
  (bv-eqns (mk-lookup-itable *itable-init*))
  (bv-recursive (mk-lookup-itable *itable-init*))
  (todo-bv-eqns nil)
  (hat-eqns (mk-lookup-itable *itable-init*))
  (compl-eqns (mk-lookup-itable *itable-init*))
  (inter-eqns (mk-lookup-table))
  (bv-subset (mk-lookup-itable *itable-init*))
  (undef-bvars nil)
  (pending (mk-empty-queue))
  (upper-bounds (mk-lookup-itable *itable-init* (type-one)))
  (lower-bounds (mk-lookup-itable *itable-init* (type-zero)))
  (todo-constraints nil)
  (named-types (mk-lookup-table)) ;; type defs: (name $ type) get recorded here
  (constr-name (mk-lookup-table)) ;; constructor names get recorded here.
  (constr-arity (mk-lookup-table));; constructor arities get recorded here.
  (constr-id    (mk-lookup-table));; constructor ids get recorded here (inverse of constr-name)
)


(defun adjust-tbl (tbl)
  (when (>= (tbl-var tbl) (array-total-size (tbl-bv-eqns tbl)))
	(let ((new (* 2 (tbl-var tbl))))
	  (setf (tbl-bv-eqns tbl)
		(itable-adjust new (tbl-bv-eqns tbl)))
	  (setf (tbl-bv-recursive tbl)
		(itable-adjust new (tbl-bv-recursive tbl)))
	  (setf (tbl-hat-eqns tbl)
		(itable-adjust new (tbl-hat-eqns tbl)))
	  (setf (tbl-compl-eqns tbl)
		(itable-adjust new (tbl-compl-eqns tbl)))
	  (setf (tbl-bv-subset tbl)
		(itable-adjust new (tbl-bv-subset tbl)))
	  (setf (tbl-upper-bounds tbl)
		(itable-adjust new (tbl-upper-bounds tbl) (type-one)))
	  (setf (tbl-lower-bounds tbl)
		(itable-adjust new (tbl-lower-bounds tbl) (type-zero)))
	  )))

;;; tkl 11/30/92 copy-tbl makes a new copy of tbl which does NOT share any fields with tbl.

(defun copy-tbl-deep (tbl)
  (let ((newtbl (make-tbl)))
    (setf (tbl-var newtbl) (tbl-var tbl))
    (setf (tbl-bv-eqns newtbl) (copy-lookup-itable (tbl-bv-eqns tbl)))
    (setf (tbl-bv-recursive newtbl) (copy-lookup-itable (tbl-bv-recursive tbl)))
    (setf (tbl-todo-bv-eqns newtbl) (copy-alist (tbl-todo-bv-eqns tbl)))
    (setf (tbl-hat-eqns newtbl) (copy-lookup-itable (tbl-hat-eqns tbl)))
    (setf (tbl-compl-eqns newtbl) (copy-lookup-itable (tbl-compl-eqns tbl)))
    (setf (tbl-inter-eqns newtbl) (copy-lookup-table (tbl-inter-eqns tbl)))
    (setf (tbl-bv-subset newtbl) (copy-lookup-itable (tbl-bv-subset tbl)))
    (setf (tbl-undef-bvars newtbl) (copy-list (tbl-undef-bvars tbl)))
    (setf (tbl-pending newtbl) (copy-queue-deep (tbl-pending tbl)))
    (setf (tbl-upper-bounds newtbl) (copy-lookup-itable (tbl-upper-bounds tbl)))
    (setf (tbl-lower-bounds newtbl) (copy-lookup-itable (tbl-lower-bounds tbl)))
    (setf (tbl-todo-constraints newtbl) (copy-alist (tbl-todo-constraints tbl)))
    (setf (tbl-named-types newtbl) (copy-lookup-table (tbl-named-types tbl)))
    (setf (tbl-constr-name newtbl) (copy-lookup-table (tbl-constr-name tbl)))
    (setf (tbl-constr-arity newtbl) (copy-lookup-table (tbl-constr-arity tbl)))
    (setf (tbl-constr-id newtbl) (copy-lookup-table (tbl-constr-id tbl)))
     newtbl))

;;;==========================================
;;;
;;;  Allocating Variables
;;;
;;;==========================================

(defvar *var-step* 1)
(setq *var-step* 1)

(defun get-var-step () *var-step*)

(defun next-var-id (tbl)
  (let ((v (incf (tbl-var tbl) *var-step*)))
    (adjust-tbl tbl)
    v))

(defun next-bvar (tbl &key (rl *rl*))
  (let ((v (mk-bvar (next-var-id tbl) :rl rl)))
    (add-undef-bvar v tbl)
    v))

(defun next-var (tbl)
  (mk-var (next-var-id tbl)))


;;;=======================================
;;;
;;; Undefined Bound Variables
;;;
;;;=======================================

(defun add-undef-bvar (var tbl)
  (push (var-id var) (tbl-undef-bvars tbl)))

(defun remove-undef-bvar (var tbl)
  (setf (tbl-undef-bvars tbl)
	(delete (var-id var) (tbl-undef-bvars tbl))))

(defun no-undef-bvars (tbl)
  (null (tbl-undef-bvars tbl)))

;;;=======================================
;;;
;;; Recursiveness of bound variables
;;;
;;;======================================

(defun is-bv-recursive (var tbl)
  (eq :rec (itable-lookup (var-id var) (tbl-bv-recursive tbl))))

(defun is-bv-non-recursive (var tbl)
  (eq :nonrec (itable-lookup (var-id var) (tbl-bv-recursive tbl))))

(defun bv-rec (var tbl)
  (itable-add (var-id var) :rec (tbl-bv-recursive tbl)))

(defun bv-nonrec (var tbl)
  (itable-add (var-id var) :nonrec (tbl-bv-recursive tbl)))


;;;=======================================
;;;
;;; Processing pending equations.
;;;
;;;=======================================

(defun process-pending (tbl)
  (if (empty-queue (tbl-pending tbl))
      t
    (let* ((next (hdqueue (tbl-pending tbl)))
	   (kind (first next))
	   (var (second next))
	   (rhs (bv-def var tbl)))
      (when rhs
	    (dequeue (tbl-pending tbl))
	    (case kind
		  ;;; 
		  ;;; case (:hat v vhat)
		  ;;; tbl has equation v = X
		  ;;; add equation vhat = ^X
		  ;;;
		  (:hat (bv-add-def (third next) (type-hat rhs tbl) tbl))
		  ;;; 
		  ;;; case (:compl v vcompl)
		  ;;; tbl has equation v = X
		  ;;; add equation vcompl = ~X
		  ;;;
		  (:compl (bv-add-def (third next) (type-compl rhs tbl) tbl))
		  ;;;
		  ;;; case (:inter v Y vinter)
		  ;;; tbl has equation v = X
		  ;;; add equation vinter = Y&X    if X is an R-L-hat type
		  ;;; add equation vinter = X&Y    otherwise
		  ;;; This guarantees that the rhs of v is an L type
		  ;;; whenever possible.
		  ;;;
		  (:inter (bv-add-def (fourth next) 
				      (if (is-rlhattype rhs)
					  (type-inter (third next) rhs tbl)
					(type-inter  rhs (third next) tbl))
				      tbl))
		  (t (error "Unrecognized case in process-pending")))
	    (process-pending tbl)))))


(defun hat-pending (v vhat tbl)
  (enqueue (list :hat v vhat) (tbl-pending tbl))
  (process-pending tbl))

(defun compl-pending (v vcomp tbl)
  (enqueue (list :compl v vcomp) (tbl-pending tbl))
  (process-pending tbl))

(defun inter-pending (lv x rv tbl)
  (enqueue (list :inter lv x rv tbl) (tbl-pending tbl))
  (process-pending tbl))


;;;=======================================
;;;
;;; todo-bv-eqns
;;; bound variable equations that have
;;; yet to be checked for emptiness.
;;;
;;;======================================

(defun todo-add-bv (var rhs tbl)
  (push (cons var rhs) (tbl-todo-bv-eqns tbl)))

(defun todo-clear (tbl)
  (setf (tbl-todo-bv-eqns tbl) nil))

(defun todo-empty (tbl)
  (null (tbl-todo-bv-eqns tbl)))

(defun todo-get (tbl)
  (tbl-todo-bv-eqns tbl))

;;;=======================================
;;;
;;; Lookup and add functions for tbl's
;;;
;;;=======================================

;;;
;;; bv-def returns the definition of a bound variable.
;;; If there is a definition, it is in bv-eqns. If there is no 
;;; definition, nil is returned.
;;;
(defun bv-def (var tbl)
  (itable-lookup (var-id var) (tbl-bv-eqns tbl)))

;;;
;;; bv-add-def adds a new bound variable equation to tbl.
;;; Every new equation is also added to the table
;;; of equations that need to be checked for equality to 0.

(defun bv-add-def (var rhs tbl)
;;; remove the comment brackets for debugging
    #| (when (= (var-id var) 55)
	 (format t "~% WITHIN BV-ADD-DEF, var is:")
	 (pt var tbl)
	 (format t "~% Its rhs is:")
	 (pt rhs tbl)
	 (setq a tbl)
	 (break)) 
  |#
  (itable-add (var-id var) rhs (tbl-bv-eqns tbl))  ;; add   var = rhs to table
  (todo-add-bv var rhs tbl)                       ;; note that this equation may be var = 0
  (remove-undef-bvar var tbl)                     ;; note that var has a definition
  (constr-add-upper-bound var rhs tbl)            ;; var <= rhs
  (constr-add-lower-bound var rhs tbl)            ;; rhs <= var are constraints
  (process-pending tbl)                           ;; process any pending equations
  (type-empty tbl)                                ;; set empty equations to 0
  )

;;;
;;; Once there is a definition, changing must only require
;;; replacing it in the table of equations.
;;;

(defun bv-replace-def (var rhs tbl)
  #| (when (= (var-id var) 7875)
	(format t "~% WITHIN BV-REPLACE-DEF, var is:")
	(pt var tbl)
	(format t "~% Its rhs is:")
	(pt rhs tbl)
	(break)) |#
  (itable-add (var-id var) rhs (tbl-bv-eqns tbl))
  (set-constr-upper-bound var rhs tbl)
  (set-constr-lower-bound var rhs tbl)
  tbl)

;;;
;;; bv-hat(a) returns bound variable b such that ^a = b
;;;
(defun bv-hat (var tbl)
  (assert (is-bvar var))
  (let ((v (itable-lookup (var-id var) (tbl-hat-eqns tbl)))) 
    (cond
     (v (type-simplify v tbl))       ;; if ^var = v       then return v
     ((is-hattype var)               ;; if ^var = var     then add ^var = var to table
      (bv-add-hat var var tbl)
      (type-simplify var tbl))
     (t
      (let ((vhat (next-bvar tbl :rl *rlhat*))) ;; allocate a fresh variable
	                                        ;; vhat is an R, L, and hat type
	(bv-add-hat var vhat tbl)     ;; ^var  = vhat
	(bv-add-hat vhat vhat tbl)    ;; ^vhat = vhat
	(hat-pending var vhat tbl)    ;; note that equation vhat = X needed
	vhat)))))

(defun bv-lookup-hat (var tbl)
  (itable-lookup (var-id var) (tbl-hat-eqns tbl)))


(defun bv-add-hat (var hatvar tbl)
  (itable-add (var-id var) hatvar (tbl-hat-eqns tbl)))

;;;
;;; bv-comp(a) returns bound variable b such that ~a = b
;;;
(defun bv-lookup-compl (var tbl)
  (itable-lookup (var-id var) (tbl-compl-eqns tbl)))

(defun bv-compl (var tbl)
  (let ((v (bv-lookup-compl var tbl)))
    (if v (type-simplify v tbl)               ;; if ~var = v then return v
      (let* ((vhat (bv-lookup-hat var tbl))   ;; get hat ^var = vhat
	     (v (cond                          ;; check if ~vhat is defined
		 ((null vhat)
		  nil)
		 ((is-bvar vhat)
		  (bv-lookup-compl vhat tbl))
		 ((type-memo vhat)
		  (type-memo vhat))
		 (t nil))))
	(if v
	    (progn
	      (bv-add-compl var v tbl)        ;; if ~vhat = v, add memo ~var =v
	      (type-simplify v tbl))          ;; complement is v
	  (let ((vcomp (next-bvar tbl :rl *rlhat*))) ;; o/w, allocate fresh var
	                                      ;; vcomp is an R, L, and hat type
	    (bv-add-compl var vcomp tbl)      ;; ~var = vcomp
	    (bv-add-hat vcomp vcomp tbl)      ;; ^vcomp = vcomp
	    (when (is-bvar vhat)
		  (bv-add-compl vhat vcomp tbl)     ;; ~vhat = vcomp
		  (bv-add-compl vcomp vhat tbl))     ;; ~vcomp = vhat
	    (compl-pending var vcomp tbl)     ;; note that equation for vcomp
	                                      ;;    is needed
	    (type-simplify vcomp tbl)))))))

(defun bv-add-compl (var complvar tbl)
  (itable-add (var-id var) complvar (tbl-compl-eqns tbl)))
  

;;;
;;; bv-inter(b,X,tbl) returns a if  b&X = a  is in tbl
;;;
;;; b must be a bound variable
;;; X must be an rl type
;;;
(defun bv-inter (argvar typ tbl)
  ;; (assert (and (is-bvar argvar) (is-rltype typ))) 
  (assert (is-bvar argvar))
  (let ((v (table-lookup (cons argvar typ) (tbl-inter-eqns tbl))))
    (if v (type-simplify v tbl)  ;; if argvar&typ = v, return v
      ;;
      ;; allocate a variable; its R,L, and hat properties are
      ;; the intersection of the properties of intersected types
      ;;
      (let ((vinter (next-bvar tbl :rl (rlh-and argvar typ)))) 
	(bv-add-inter vinter argvar typ tbl)
	(bv-subset vinter (list argvar typ) tbl) ;; x = y&Z ==> x <=y and x <= Z
	(inter-pending argvar typ vinter tbl)
	(type-simplify vinter tbl)  ;;; change 2/18/93 TKL (used to just return vinter)
	))))


(defun bv-add-inter (rv lv typ tbl)
  (table-add (cons lv typ) rv (tbl-inter-eqns tbl)))


;;;
;;; When a new bound variable x is defined as the intersection of a bound variable
;;; y and another type Z, and intersection equation
;;;     x = y & Z
;;; is added to the system of equations, indexed by the pair (y,Z).  To guarantee 
;;; termination of the type intersection algorithm, however, we also need to 
;;; exploit the fact that (by definition) x <= y and x <= Z (indexed by x).
;;; bv-subset memoizes these containment relationships.  
;;;
(defun bv-subset (x supersets tbl)
    (itable-add (var-id x) supersets (tbl-bv-subset tbl)))

(defun bv-less-by-intersection (bv typ tbl as sols)
  #| (format t "~%In bv-less-by-inter:  ~a <= ~a <= ~a~%"
	  (pt bv tbl nil)
	  (mapcar #'(lambda (x) (pt x tbl nil)) (itable-lookup (var-id bv) (tbl-bv-subset tbl)))
	  (pt typ tbl nil)) |#
  (some #'(lambda (x) (type-<=-aux x typ tbl as sols)) 
	(itable-lookup (var-id bv) (tbl-bv-subset tbl))))


;;;===========================================================
;;;
;;; Functions for manipulating the constraints on variables.
;;;
;;;===========================================================


;;;
;;; Functions that add and remove constraints from the
;;; list of constraints to be processed.
;;;
(defun add-todo-constraint (t1 t2 tbl)
  (push (make-constraint t1 t2) (tbl-todo-constraints tbl)))

(defun next-todo-constraint (tbl)
  (if (not (null (tbl-todo-constraints tbl)))
      (pop (tbl-todo-constraints tbl))))


;;;
;;; lookup and add functions for the upper and lower
;;; bounds of a variable.
;;;
(defun constr-upper-bound (var tbl)
  (itable-lookup (var-id var) (tbl-upper-bounds tbl)))

(defun constr-lower-bound (var tbl)
  (itable-lookup (var-id var) (tbl-lower-bounds tbl)))

(defun set-constr-upper-bound (var typ tbl)
  (itable-add (var-id var) typ (tbl-upper-bounds tbl)))

(defun set-constr-lower-bound (var typ tbl)
  (itable-add (var-id var) typ (tbl-lower-bounds tbl)))

;;;
;;; if a <= X and a <= Y, then a <= X & Y
;;;
(defun constr-add-upper-bound (var typ tbl)
  (itable-add (var-id var) 
	     (type-inter typ (constr-upper-bound var tbl) tbl)
	     (tbl-upper-bounds tbl)))

;;;
;;; if X <= a and Y <= a, then X | Y <= a
;;;
(defun constr-add-lower-bound (var typ tbl)
  (itable-add (var-id var) 
	     (type-union typ (constr-lower-bound var tbl) tbl)
	     (tbl-lower-bounds tbl)))


;;;
;;; When a lower bound X <= a is added, two cases are checked.
;;; First, if for current lower bound Y <= a it happens that X <= Y,
;;; then the constraint is redundant.  If it isn't redundant,
;;; then it is added, and for upper bound a <= Z, the constraint
;;; X <= Z is added to the list of constraints to be processed.
;;;
;;; Formally, implements two rules:
;;;    C, X <= a, Y <= a   ==   C, Y <= a    if X <= Y
;;;    C, X <= a, a <= Z   ==   C, X <= a, a <= Z, X <= Z
;;;
(defun lower-bound (low var tbl)
  (when (not (type-<= low (constr-lower-bound var tbl) tbl))
	(constr-add-lower-bound var low tbl)
	(add-todo-constraint low (constr-upper-bound var tbl) tbl)))

;;;
;;; This is the symmetric case for upper bounds.
;;; Formally, implements two rules:
;;;    C, a <= X, a <= Y   ==   C, Y <= a    if X <= Y
;;;    C, Z <= a, a <= X   ==   C, Z <= a, a <= X, Z <= X
;;;     
(defun upper-bound (var up tbl)
  (when (not (type-<= (constr-upper-bound var tbl) up tbl))
	(constr-add-upper-bound var up tbl)
	(add-todo-constraint (constr-lower-bound var tbl) up tbl)))
	
;;;
;;; The top level routine for solving constraints.
;;; Repeatedly selects a constraint, reduces it to 
;;; inductive constraints, and closes the system under
;;; transitivity.
;;;
(defvar *printit* nil)
(defun solve-constraints (tbl)

  (do* ((constraint (next-todo-constraint tbl) (next-todo-constraint tbl)))
       ((null constraint) tbl)
       (when *printit*
	     (format t "~%~%Processing Constraint from TODO LIST: ~a <= ~a~%~%" 
		     (pt (constraint-lhs constraint) tbl nil) 
		     (pt (constraint-rhs constraint) tbl nil)))


       (mapc 
	#'(lambda (c)
	    (let ((low (constraint-lhs c))
		  (up (constraint-rhs c)))
	      (assert (or (is-any-var low) (is-any-var up)))
	      (when *printit* (format t "Adding Constraint: ~a <= ~a~%~%" 
				      (pt low tbl nil) 
				      (pt up tbl nil)))


	      (cond
	       ;;
	       ;; if the constraint has the form a_i <= a_j, where
	       ;; a_i and a_j are both variables (free or bound),
	       ;; then the constraint is an upper bound on a_i if j < i,
	       ;; and it is a lower bound on a_j if i < j.
	       ;; If i = j the constraint is always satisfied;
	       ;; this case should not arise here but is handled anyway.
	       ;;
	       ((and (is-any-var low) (is-any-var up))
		(cond
		 ((< (var-id low) (var-id up))
		  (lower-bound low up tbl))
		 ((< (var-id up) (var-id low))
		  (upper-bound low up tbl))
		 (t nil)))
	       ;;
	       ;; if the constraint is of the form a_i <= X
	       ;; where X is not a variable, then the constraint
	       ;; is an upper bound on a_i.
	       ;;
	       ((is-any-var low) 
		(upper-bound low up tbl))
	       ;;
	       ;; if the constraint is of the form X <= a_i
	       ;; where X is not a variable, then the constraint
	       ;; is a lower bound on a_i.
	       ;;
	       ((is-any-var up)
		(lower-bound low up tbl))
	       (t (error "Non-inductive constraint in solve-constraints: ~a ~a" 
			 (pt low tbl nil) 
			 (pt up tbl nil))))))
	(type-constraints (constraint-lhs constraint) 
			  (constraint-rhs constraint) 
			  tbl)
	)))

		

;;;
;;; Protected constraints solver permits solving of inconsistent systems.
;;; When the inconsistency is discovered, the catch will get a throw
;;; from the constraint solving functions.  The table is cleaned
;;; up and (:inconsistent,tbl) are returned as the values.  If everything
;;; terminates normally then (:consistent,tbl) is returned as the value.
;;;
(defun protect-solve-constraints (tbl)
  (let* ((*inconsistent-ok* t)
	 (res (catch 'inconsistent (solve-constraints tbl) :consistent)))
    (when (not (equal res :consistent)) ;; cleanup is necessary
	  (assert (null (tbl-undef-bvars tbl)))
	  (assert (empty-queue (tbl-pending tbl)))
	  (assert (null (tbl-todo-bv-eqns tbl)))
	  (setf (tbl-todo-constraints tbl) nil))
    (values (if (equal res :consistent) :consistent :inconsistent) tbl)))


;;;
;;;=========================================================
;;;
;;; Functions for accessing/adding named-types
;;;
;;;=========================================================

;;;
;;; 1/16/93 aa  if a named type is added a second time, discard it
;;;
(defun named-type-add (str typ tbl)
  (if (not (named-type-get str tbl))
      (table-add str typ (tbl-named-types tbl))))

(defun named-type-get (str tbl)
  (table-lookup str (tbl-named-types tbl)))



;;;
;;;=========================================================
;;;
;;; Functions for accessing/adding constructor attributes
;;;
;;;=========================================================

(defun cons-name-add (id name tbl)
  (table-add id name (tbl-constr-name tbl)))

(defun cons-name-get (id tbl)
  (table-lookup id (tbl-constr-name tbl)))

(defun cons-id-add (name id tbl)
  (table-add name id (tbl-constr-id tbl)))

(defun cons-id-get (name tbl)
  (table-lookup name (tbl-constr-id tbl)))

(defun cons-arity-add (id arity tbl)
  (table-add id arity (tbl-constr-arity tbl)))

(defun cons-arity-get (id tbl)
  (table-lookup id (tbl-constr-arity tbl)))

