(in-package 'types)

;;;
;;; This file implements type hatting. 
;;; (type-hat T) is the least upward closed monotype that contains T.
;;; 
;;; Functions in this file are called from the top level type-hat function
;;; which is in a separate file "hat_tbl.lisp" (this is done so as to allow
;;; the lisp compiler to define the functions (here) in the right order.
;;; 

;;; 0. E |- ^0 = 0
;;;

(defun hat-zero(typ eqns) 
  (declare (ignore typ eqns))
  (type-zero))

;;; 1. E |- ^1 = 1
;;;

(defun hat-one(typ eqns) 
  (declare (ignore typ eqns))
  (type-one))

;;; 2a E |- ^(A|B) = ^A*^B      if (A|B) is a hattype
;;;
;;;     E |- ^A = X, ^B = Y, ~A = W , ~B = Z
;;; 2b ---------------------------------------------
;;;    E |- ^(A | B) = X * Z | X * Y |  W * Y  
;;;

(defun hat-union (X eqns)
  ;;; if X is an r type then we know that the hat ^(X1|X2) is just ^X1 | ^X2
  ;;; This is because ^X1 * ^X2 = 0 and ^X1 * ~X2 == ^X1 and ~X1*^X2 == ^X2
  (if (is-rtype X) 
      (type-union 
       (type-hat-aux (type-t1 X) eqns) 
       (type-hat-aux (type-t2 X) eqns) 
       eqns 
       :disjoint t)
  (type-union 
   (type-inter 
    (type-hat-aux (type-t1 X) eqns) 
    (type-compl-aux (type-t2 X) eqns) eqns)
   (type-union 
    (type-inter (type-hat-aux (type-t1 X) eqns) 
		(type-hat-aux (type-t2 X) eqns) eqns)
    (type-inter (type-compl-aux (type-t1 X) eqns) 
		(type-hat-aux (type-t2 X) eqns) eqns) 
    eqns
    :disjoint t) 
   eqns
   :disjoint t)))


;;;       E |- ^A = X, ^B = Y
;;; 3. ----------------------------
;;;    E |- ^(A * B) = X * Y
;;;

(defun hat-inter (X eqns)
  (type-inter (type-hat-aux (type-t1 X) eqns) (type-hat-aux (type-t2 X) eqns) eqns))

;;;    E |- ^A1 = X1,..... ^An = Xn
;;; 4. -------------------------------
;;;    E |- ^c(A1,,,,An) = c(X1,...Xn)
;;;

(defun hat-cons (X eqns)
  (type-cons (cons-id X) (mapcar #' (lambda (x) (type-hat-aux x eqns)) (cons-args X)) eqns ))

;;; 
;;; 5. E |- ^(X->Y) = 0 -> 1 + 1 (all-functions)
;;;

(defun hat-func (X eqns)
  (declare (ignore X eqns))
  (type-all-funcs))

;;; 6. E |- ^alpha = 1 (for all free variables alpha)
;;;

(defun hat-var (X eqns)
  (declare (ignore X eqns))
  (type-one))

;;; 7. E, {^a = b} |- ^a = b
;;;

(defun hat-bvar (X eqns) 
  (bv-hat X eqns)) ;;; bv-hat : var * eqns -> var 
                             ;;; bv-hat gets the variable associated with ^a

;;; 8. E |- ^(not X) = (not X)
;;; 

(defun hat-not (X eqns) 
  (declare (ignore eqns))
  X)

;;;     E |- ^A = X                       
;;; 9. -----------------    assumption: ^Y <> 0 
;;;    E |- ^(A ? Y) = X    If the system is correct, this assumption will hold.    
;;;          
;;;			              

(defun hat-dep (X eqns)
  (type-hat-aux (type-t1 X) eqns))


;;;      E |- ^A = X
;;; 10. -----------------------------
;;;     E |- ^(val v A) = (val v X)
;;;

(defun hat-value (X eqns)
  (type-value (value-val X) (type-hat-aux (value-typ X) eqns) eqns))

