(in-package 'types)

;;;
;;; This file implements type homomorphism function
;;; (type-hom X h) represents the result of applying the homomorphism
;;; h : var -> type | :no-change  to the type X.
;;; Functions in this file are called from the top level type-hom function
;;; which is defined in a separate file "hom_tbl.lisp" (this is done so as to allow
;;; the lisp compiler to define the functions (here) in the right order.
;;; 

;;; 0. E |- hom(0) = 0
;;;

(defun hom-zero (typ h p eqns) 
  (declare (ignore typ h p eqns))
  :no-change)


;;; 1. h(1) = 1
;;;

(defun hom-one(typ h p eqns) 
  (declare (ignore typ h p eqns))
  :no-change)

;;; 
;;;
;;; 2.  h(A | B) = h(A) | h(B) 
;;;

(defun hom-union (X h p eqns)
  (let ((lhs (type-hom-aux (type-t1 X) h p eqns))
	(rhs (type-hom-aux (type-t2 X) h p eqns)))
    (cond 
     ((and (eql :no-change lhs) (eql :no-change rhs)) 
      :no-change)
     ((eql :no-change lhs) 
      (type-union (type-t1 X) rhs eqns :disjoint (is-rtype X) :nosubset (is-nsunion X)))
     ((eql :no-change rhs) 
      (type-union lhs (type-t2 X) eqns :disjoint (is-rtype X) :nosubset (is-nsunion X)))
     (t 
      (type-union lhs rhs eqns :disjoint (is-rtype X) :nosubset (is-nsunion X) )))))

;;;
;;;  3. h(A*B) = h(A) * h(B)
;;;
(defun hom-inter (X h p eqns)
  (let ((lhs (type-hom-aux (type-t1 X) h p eqns))
	(rhs (type-hom-aux (type-t2 X) h p eqns)))
    (cond 
     ((and (eql :no-change lhs) (eql :no-change rhs)) 
      :no-change)
     ((eql :no-change lhs) 
      (type-inter (type-t1 X) rhs eqns ))
     ((eql :no-change rhs) 
      (type-inter lhs (type-t2 X) eqns ))
     (t 
      (type-inter lhs rhs eqns)))))


;;;    
;;; 4. h(c(A1,...,An) = c(h(A1),...,h(An))
;;;
(defun hom-list (xs h p eqns)
  (if (null xs)
      :no-change
    (let* ((x (car xs))
	   (newx (type-hom-aux x h p eqns))
	   (rest (hom-list (cdr xs) h p eqns)))
      (cond
       ((and (eql newx :no-change) (eql rest :no-change))
	:no-change)
       ((eql rest :no-change)
	(cons newx (cdr xs)))
       ((eql newx :no-change)
	(cons x rest))
       (t 
	(cons newx rest))))))

(defun hom-cons (X h p eqns)
  (let ((arg-lst (hom-list (cons-args X) h p eqns)))
    (if (eql arg-lst :no-change)
	:no-change
      (type-cons (cons-id X) arg-lst eqns))))


;;; 
;;;  5. h(X->Y+Z) = h(X)->h(Y)+h(Z)
;;;
(defun hom-func (X h p eqns)
  (let* ((dom (type-hom-aux (type-t1 X) h p eqns))
	 (range (type-hom-aux (type-t2 X) h p eqns))
	 (err (type-hom-aux (type-t3 X) h p eqns)))
    (if (and (eql :no-change dom) 
	     (eql :no-change range)
	     (eql :no-change err))
	:no-change
      (type-func (if (eql :no-change dom) (type-t1 X) dom)
		 (if (eql :no-change range) (type-t2 X) range)
		 (if (eql :no-change err) (type-t3 X) err)
		 eqns))))
;;;
;;; 6. free variables
;;;
(defun hom-var (X h p eqns)
  (funcall h X h p eqns))

;;; 
;;; 7. bound variables
;;;
(defun hom-bvar (X h p eqns) 
  (funcall h X h p eqns))

;;; 8. h(not X) = not X
;;; 
(defun hom-not (X h p eqns) 
  (declare (ignore X h p eqns))
  :no-change)

;;;     E |- h(A) = X, h(B) = Y
;;; 9. -------------------------------
;;;     E |- h(A ? B) = X?Y
;;;          
(defun hom-dep (X h p eqns)
  (let* ((lhs (type-hom-aux (type-t1 X) h p eqns))
	 (rhs (type-hom-aux (type-t2 X) h p eqns))
	 (rhs-arg (if (eql :no-change rhs) (type-t2 X) rhs)))
    (cond 
     ((type-not-zero-for-sure rhs-arg eqns) (if (eql :no-change lhs) (type-t1 X) lhs))
     ((and (eql :no-change lhs) (eql :no-change rhs)) 
      :no-change)
     ((eql :no-change lhs) 
      (type-dep (type-t1 X) rhs eqns ))
     ((eql :no-change rhs) 
      (type-dep lhs (type-t2 X) eqns ))
     (t (type-dep lhs rhs eqns)))))

;;;
;;;  10.  h(val v X) = val v h(X)
;;;
(defun hom-value (X h p eqns)
  (let ((arg  (type-hom-aux (value-typ X) h p eqns)))
    (if (eql :no-change arg)
	:no-change
      (type-value (value-val X) arg eqns))))
	
;;;     
;;; Q/L types contain NO free variables, so we define homomorphisms to be just the
;;; identity for them.
;;;
;;;      
;;; 11. E |- h(for-all x. A) = (for-all x. A)
;;;

(defun hom-quant (X h p eqns)
  (declare (ignore h p eqns))
  X)

;;;      
;;; 12. E |- hom(lambda x. A) = (lambda x. A)
;;;

(defun hom-lambda (X h p eqns)
  (declare (ignore h p eqns))
  X)

