(in-package 'types)

;;;
;;; This file implements type intersection : type-inter : Eqns * Type * Type -> Type*
;;; 
;;; Functions in this file are called from the top level type-inter function
;;; which is in a separate file "inter_tbl.lisp" (this is done so as to allow
;;; the lisp compiler to define the functions (here) in the right order.
;;; 
;;; Intersections are normalized by this algorithm (i.e., driven "in" as far
;;; as possible).  The result is of the form: 
;;;        a. X*Y*Z  or (X1->Y1+Z1)*(X2->Y2+Z2)
;;;    or  b. free-var * type

;;; E |- X * 1 = X
;;;

(defun proj-t1 (typ1 typ2 eqns)
  (declare (ignore typ2 eqns))
  typ1)

;;; E |- 1 * X = X
;;;

(defun proj-t2 (typ1 typ2 eqns)
  (declare (ignore typ1 eqns))
  typ2)

;;; E |- X1 * Z = Y1 , X2 * Z = Y2
;;; ------------------------------
;;; E |- (X1 | X2) * Z = (Y1 | Y2)

(defun un-l (union typ2 eqns)
  (type-union (type-inter-aux (type-t1 union) typ2 eqns)
	      (type-inter-aux (type-t2 union) typ2 eqns) 
	      eqns
	      :disjoint (is-rtype union)
	      :nosubset (is-nsunion union)))

;;; E |- X * Z1 = Y1, X * Z2 = Y2
;;; -----------------------------
;;; E |- X * (Z1 | Z2) = (Y1 | Y2)

(defun un-r (typ1 union eqns)
  (type-union (type-inter-aux typ1 (type-t1 union)  eqns)
	      (type-inter-aux typ1 (type-t2 union)  eqns) 
	      eqns
	      :disjoint (is-rtype union)
	      :nosubset (is-nsunion union)))

;;; E |- W * Z = Y
;;; ------------------------
;;; E |- (X * W) * Z = (X * Y) <- recursive call to type-inter-aux

(defun in-l (inter typ2 eqns)
  (type-inter-aux (type-t1 inter) 
		  (type-inter-aux (type-t2 inter) typ2 eqns) 
		  eqns))

;;; E |- X * (Y * Z) = mkinter(Y,X*Z)
;;;

(defun in-r (typ1 inter eqns)
  (mk-inter (type-t1 inter) 
	    (type-inter-aux typ1 (type-t2 inter) eqns)))

;;; a. E |- c(...) * d(...) = 0   if c <> d
;;;
;;; b. E |- for-all i : 1 <= i <= n; Xi * Yi = Zi
;;;    ------------------------------------------
;;;    E |- c(X1,,,,Xn) * c(Y1,,,Yn) = c(Z1,,,,Zn)

(defun cons-cons (typ1 typ2 eqns)
  (if (= (cons-id typ1) (cons-id typ2)) 
      (type-cons (cons-id typ1) (mapcar #'(lambda (x y) (type-inter-aux x y eqns)) (cons-args typ1) (cons-args typ2)) eqns)
      (type-zero)))

;;; E |- c(...) * not({b1,...bn}) = 0        if c is in  {b1...bn}
;;;                               = c(...)   o/w
;;;

(defun cons-not (typ1 typ2 eqns)
  (declare (ignore eqns))
  (if (member (cons-id typ1) (not-list typ2)) 
      (type-zero)
      typ1))

;;; E |- not({b1,...bn}) * c(...)  = 0        if c is in  {b1...bn}
;;;                                = c(...)   o/w
;;;

(defun not-cons (typ1 typ2 eqns)
  (declare (ignore eqns))
  (if (member (cons-id typ2) (not-list typ1)) 
      (type-zero)
      typ2))

;;; E |- a * X = mk-inter(a,X)       if a is a free variable
;;;

(defun var-l (a typ2 eqns)
  (declare (ignore eqns))
  (mk-inter a typ2))

;;; E |- X * a = mk-inter(a,X)       if a is a free variable
;;;

(defun var-r (typ1 a eqns)
  (declare (ignore eqns))
  (mk-inter a typ1))


;;;
;;; To drive intersections as far "IN" as possible,
;;; unfold bound variables intersected with free variables.
;;;
(defun bv&v (bv v eqns)
  (let ((def (bv-def bv eqns)))
    (if def 
	(type-inter-aux def v eqns)
        (var-r bv v eqns))))

(defun v&bv (v bv eqns)
  (let ((def (bv-def bv eqns)))
    (if def
	(type-inter-aux v def eqns)
      (var-l v bv eqns))))


;;; for bound variables
;;; a. E union {a * X = b} |- a * X = b  if X is an rl type (i.e., an R and an L type)
;;; b. E union {a = Y} |- Y * X = Z
;;;    ----------------------------      if X is not an  rl type 
;;;    E union {a = Y} |- a * X = Z

; (defun bv-l (a typ2 eqns)
;  (if (is-rltype typ2) 
;      (bv-inter a typ2 eqns)
;      (type-inter-aux (bv-def a eqns) typ2 eqns))) 

(defun bv-l (a typ2 eqns)
  (bv-inter a typ2 eqns))


;;; a. E union {a * X = b} |- X * a = b  if X is an rl type (i.e., an R and an L type)
;;;    E union {a = Y} |- X * Y = Z
;;; b. ----------------------------      if X is not an  rl type 
;;;    E union {a = Y} |- X * a = Z

(defun bv-r (typ1 a eqns)
  (bv-inter a typ1 eqns))

;;; E |- not(B1) * not(B2) = not(B1|B2)
;;; 

(defun not-not (typ1 typ2 eqns)
  (declare (ignore eqns))
  (not-combine typ1 typ2))

;;; E |- X * Z = W
;;; ----------------
;;; E |- (X?Y)*Z = W?Y

(defun dep-l (typ1 typ2 eqns)
  (type-dep (type-inter-aux (type-t1 typ1) typ2 eqns) (type-t2 typ1) eqns))

;;; E |- Z * X = W
;;; ------------------
;;; E |- Z* (X?Y) = W?Y

(defun dep-r (typ1 typ2 eqns)
  (type-dep (type-inter-aux typ1 (type-t1 typ2) eqns) (type-t2 typ2) eqns))

;;; E |- X * Y = Z
;;; ---------------------
;;; E |- val(v,X) * Y = Z

(defun val-l (typ1 typ2 eqns)
  (type-inter-aux (value-typ typ1) typ2 eqns))

;;; E |- X * Y = Z
;;; ---------------------
;;; E |- X * val(v,Y)  = Z

(defun val-r (typ1 typ2 eqns)
  (type-inter-aux typ1 (value-typ typ2) eqns))



;;; E |- (A1 -> B1 + C1) * (A2 -> B2 + C2) = mk-inter((A1 -> B1 + C1),(A2 -> B2 + C2))
;;;

(defun mk-fn* (typ1 typ2 eqns)
  (declare (ignore eqns))
  (mk-inter typ1 typ2))


;;;
;;; To drive intersections "in" as far as possible,
;;; unfold bound variables intersected with function types.
;;;
(defun bv&fn (bv fn eqns)
  (let ((def (bv-def bv eqns)))
    (if def
	(type-inter-aux def fn eqns)
      (bv-l bv fn eqns))))

(defun fn&bv (fn bv eqns)
  (let ((def (bv-def bv eqns)))
    (if def
	(type-inter-aux fn def eqns)
      (bv-r fn bv eqns))))
