(in-package 'types)

;;;
;;; This file implements functions for computing the set
;;; of all intersections of variables that occur in R
;;; positions of a type.
;;;
;;; For example, in the type
;;;   a & b  ->   c & d  ->  e
;;; the intersections of variables in R positions are (a&b,c&d).
;;;
;;; This function is used to help identify places where
;;; a conjunction of variables can be replaced by a single variable.
;;; For example, the function
;;;
;;;    Forall a,b.  a&b -> c(a,c(b,nil))
;;;
;;; is equal to
;;;
;;;    Forall d. d -> c(d,c(d,nil))
;;;

(defun flip (rl)
  (if (eq rl :r) :l :r))

(defun il-none (typ rl sofar eqns as)
  (declare (ignore typ rl eqns as))
  sofar)

;;;
;;;   E,A |- il(A,*) = X  il(B,*) = Y
;;;   --------------------------------
;;;   E,A |- il(A u B, *) = X u Y
;;;
(defun il-union (typ rl sofar eqns as)
  (il-aux (type-t1 typ) 
	  rl 
	  (il-aux (type-t2 typ) rl sofar eqns as)
	  eqns
	  as))

;;;
;;; get-conjuncts takes a type  x1 & ... & xn and
;;; returns two lists:
;;;    (... xi ...)  such that the xi are all variables
;;;    (... xj ...)  such that the xi are all non-variables
;;;
(defun get-conjuncts (typ)
  (if (is-inter typ)
      (multiple-value-bind
       (vars nonvars)
       (get-conjuncts (type-t2 typ))
       (if (is-var (type-t1 typ))
	   (values (cons (type-t1 typ) vars)
		   nonvars)
	   (values vars
		   (cons (type-t1 typ) nonvars))))
    (if (is-var typ)
	(values (list typ) nil)
        (values nil (list typ)))))


;;;
;;; A,E |- il(Tj,R) = Xj   for 1 <= j <= m
;;; --------------------------------------
;;; A,E |- il(v1 & ... & vn & T1 & ... & Tm,R) = 
;;;        { (v1,...,vm) } u X1 u ... u Xj
;;;
;;; 
;;; A, E |- il(v & M, L) = {}
;;;
(defun il-inter (typ rl sofar eqns as)
  (if (eq rl :l)
      sofar
    (multiple-value-bind
     (vars nonvars)
     (get-conjuncts typ)
     (il-list nonvars
	      rl
	      (cons vars sofar)
	      eqns
	      as))))

(defun il-list (tlist rl sofar eqns as)
  (if (null tlist)
      sofar
    (il-list
     (cdr tlist)
     rl
     (il-aux (car tlist) rl sofar eqns as)
     eqns
     as)))

;;;
;;; A,E |- il(Ti,*) = Xi   1 <= i <= n
;;; -------------------------------------------
;;; A,E |- il(c(T1,...,Tn),*) = X1 u ... u Xn
;;;
(defun il-cons (typ rl sofar eqns as)
  (il-list (cons-args typ) rl sofar eqns as))


;;;
;;; A,E |- il(B,flip(*)) = X
;;; A,E |- il(C,*) = Y
;;; A,E |- il(D,*) = Z
;;; ---------------------------------
;;; A,E |- il(B -> C + D) = X u Y u Z
;;;
(defun il-func (typ rl sofar eqns as)
  (il-aux (type-t1 typ)
	  (flip rl)
	  (il-aux (type-t2 typ)
		  rl 
		  (il-aux (type-t3 typ) rl sofar eqns as)
		  eqns
		  as)
	  eqns
	  as))

;;;
;;; A,E |- il(v,R) = { v }
;;; A,E |- il(v,L) = {}
;;;
(defun il-var (typ rl sofar eqns as)
  (declare (ignore eqns as))
  (if (eq rl :l)
      sofar
    (cons (list typ) sofar)))

;;;
;;; A u { il(v,*) = {} }, E |- il(v,*) = {}
;;;
;;; A u { il(v,*) = {}}, E u { v = B } |- il(B,*) = X
;;; -------------------------------------------------
;;; A, E u { v = B } |- il(v,*) = X
;;;  
(defun il-bvar (bv rl sofar eqns as)
  (cond
   ((pair-assump-has (var-id bv) rl as)
    sofar)
   ((bv-def bv eqns)
    (il-aux (bv-def bv eqns)
	    rl
	    sofar
	    eqns
	    (pair-assump-add (var-id bv) rl as)))
   (t (error "undefined bound variable in il-aux"))))

   
;;;
;;; A,E |- il(B,*) = X
;;; --------------------
;;; A,E |- il(B?C,*) = X
;;;       
(defun il-dep (typ rl sofar eqns as)
  (il-aux (type-t1 typ) rl sofar eqns as))

;;;
;;; A,E |- il(B,*) = X
;;; -------------------------
;;; A,E |- il(value(v,B),*) = X
;;;
(defun il-val (typ rl sofar eqns as)
  (il-aux (value-typ typ) rl sofar eqns as))
