(in-package 'types)

;;;
;;; This file implements compound type building operations.
;;; These functions exist to help guarantee that
;;; if T = 0 then T is represented by 0 and to perform
;;; some optimizations.
;;;
;;; To save consing, the functions type-one and type-zero always
;;; return the same structure.
;;;
(defconstant *one* (mk-one))
(defun type-one (&rest x) (declare (ignore x)) *one*)

(defconstant *zero* (mk-zero))
(defun type-zero (&rest x) (declare (ignore x)) *zero*)

;;;
;;; The set of all functions is 1 -> 1; 1
;;; *all-funcs* is marked as a hat type.
;;;
(defconstant *all-funcs* 
  (let* ((x (mk-func *one* *one* *one*)))
    (setf (type-rl x) *rlhat*)
    x))

(defun type-all-funcs () *all-funcs*)

;;;
;;; The set of all constructions is NOT(emptyset)
;;;
(defconstant *all-constructions*
  (mk-not nil nil))

(defun type-all-constructions ()
  *all-constructions*)


;;;
;;; The type A | B is represented as A if B <= A
;;; and as B if A <= B.
;;;


;;;
;;; Transformations:
;;;
;;;    c(X,Y) u c(X,Z) = c(X,Y u Z)
;;;    c(X,Y) u c(Z,Y) = c(X u Z,Y)
;;;
(defun constr-unions (x y eqns disjoint)
  (if (and (is-cons x)
	   (is-cons y)
	   (= (cons-id x) (cons-id y))
	   (= (length (cons-args x)) 2))
      (let ((fx (first (cons-args x)))
	     (sx (second (cons-args x)))
	     (fy (first (cons-args y)))
	     (sy (second (cons-args y))))
      (cond
       ((or (eq fx fy)
	    (and (not disjoint)
		 (type-= fx fy eqns)))
	(mk-cons (cons-id x)
		 (list fx
		       (type-union sx sy eqns :disjoint disjoint))))
			
       ((or (eq sx sy)
	    (and (not disjoint)
		 (type-= sx sy eqns)))
	(mk-cons (cons-id x)
		 (list (type-union fx fy eqns :disjoint disjoint)
		       sx)))
       (t nil)))))

(defun save-mk-union (old new-t1 new-t2 disjoint)
  (if (and (eq (type-t1 old) new-t1)
	   (eq (type-t2 old) new-t2))
      old
    (mk-union new-t1 new-t2 :disjoint disjoint)))

;;;
;;; type-union-aux takes a type x and a union of types xs = y1 | ... | yn
;;; and returns  x | yi0 | ... |yik
;;; where yij is included in the union iff yij = some yh and 
;;; yij not<= x.
;;;
(defun type-union-aux-base (x y eqns disjoint)
  (if (and (not disjoint)
	   (type-<= y x eqns))
      (values x (type-zero))
    (let ((constr-opt (constr-unions x y eqns disjoint)))
      (if constr-opt
	  (values constr-opt (type-zero))
	(values x y)))))

(defun type-union-aux (x xs eqns disjoint)
  (if (is-union xs)
      (let* ((this-test 
	      (if (not disjoint) 
		  (type-<= (type-t1 xs) x eqns)))
	     (constr-opt 
	      (if (not this-test)
		  (constr-unions x (type-t1 xs) eqns disjoint))))
	(multiple-value-bind
	 (new-x new-xs)
	 (type-union-aux 
	  (if constr-opt constr-opt x) 
	  (type-t2 xs) 
	  eqns
	  disjoint)
	 (values
	  new-x
	  (if (or constr-opt this-test)
	      new-xs
	    (save-mk-union xs (type-t1 xs) new-xs disjoint)))))
    (type-union-aux-base x xs eqns disjoint)))

(defun type-union-rest (x xs eqns disjoint)
  (multiple-value-bind
   (newx newxs)
   (type-union-aux x xs eqns disjoint)
   (mk-union newx newxs :disjoint disjoint)))
    
	  
(defun type-union (t1 t2 eqns &key (disjoint nil) (nosubset nil))
  (cond
     ;;; unions are right linearized
     ;;; (t11 | t12) | t2 ==> (t11 | (t12 | t2))
   (disjoint
    (mk-union t1 t2 :disjoint t :nosubset t))
   (nosubset
    (mk-union t1 t2 :nosubset t))
   ((and (is-union t1)
	 (not (is-union t2)))
    (type-union t2 t1 eqns :disjoint disjoint :nosubset nosubset))
   ((is-union t1) 
    (type-union (type-t1 t1) 
		(type-union (type-t2 t1) 
			    t2 
			    eqns
			    :disjoint disjoint
			    :nosubset nosubset) 
		eqns
		:disjoint (and (is-rtype t1) disjoint)
		:nosubset (and (is-nsunion t1) nosubset)
		))
   ((type-<= t1 t2 eqns) 
    t2)    
  (t (type-union-rest t1 t2 eqns disjoint))))
;;  ((type-<= t2 t1 eqns) t1 )
;;  (t (mk-union t1 t2))))


;;;
;;; type-inter is in the file inter_tbl.lisp and inter_fns.lisp
;;;
;;;

;;;
;;; A construction is zero if any of the components
;;; is zero.
;;;
(defun any-empty (l)
  (if (null l)
      nil
    (or (is-zero (car l))
	(any-empty (cdr l)))))

(defun type-cons (consid listoftypes eqns)
  (declare (ignore eqns))
  (if (any-empty listoftypes)
      (type-zero)
    (mk-cons consid listoftypes)))


;;;
;;; A function type is always non-zero.
;;; If the type is 0 -> X + Y or Z -> 1 + 1,
;;; then it is the set of all functions.
;;;
;;; The above is not true in the model intended for Illyria,
;;; where the type Rec(0 -> 0) (i.e., the empty record)
;;; is not the same as Rec(1 -> 1) (i.e., a record where all
;;; fields are defined.
;;;
(defun type-func (x y z eqns)
  (declare (ignore eqns))
;  (if (or (is-zero x)
;	  (and
;	   (is-one y)
;	   (is-one z)))
;      (type-all-funcs)
    (mk-func x y z))

      
;;;
;;; A dependent type X ? Y is zero if Y is zero
;;; 1/20/93 tkl changed: furthermore, X ? Y  = X if Y is definitely not zero.

(defun type-dep (x y eqns)
  (cond ((is-zero y) (type-zero))
	((type-not-zero-for-sure y eqns) x)
	(t (mk-dep x y))))

;;;
;;; For completeness the other type- functions are defined
;;; here, even though they don't do anything different
;;; from the mk- functions.
;;;

(defun type-var (x eqns)
  (declare (ignore eqns))
  (mk-var x))

(defun type-bvar (x eqns)
  (declare (ignore eqns))
  (mk-bvar x))

(defun type-value (val typ eqns)
  (declare (ignore eqns))
  (mk-value val typ))


;;;
;;; For type-not the function takes only a single constructor
;;; as an argument.  Use mk-not or not-combine to build
;;; not types with more than one constructor.
;;;

(defun type-not (constype &optional eqns)
  (declare (ignore eqns))
  (mk-not (list (cons-id constype)) 
	  (list (length (cons-args constype)))))

;;; TYPE-LAMBDA AND TYPE-QUANT : 
;;;
;;; Auxiliary function
;;; free-vars  : list(vars) -> list(fvars)
;;; bound-vars : list(vars) -> list(bvars)
;;; free-vars (bound-vars) takes as input a list  of variables and
;;; returns all the FREE (BOUND) variables in this list

(defun free-vars (l)
  (remove-if #'is-bvar l))

(defun bound-vars (l)
  (remove-if #'is-var l))


;;; Generic construct function : constructs a quantified type or a lambda type.
;;; The fn argument determines whether the type to be constructed is a quant or a lambda type 


;;; tkl 1/12/93 defined a function filter-dupl-hat which filters out the duplicates and the
;;; bvs that are hattypes (i.e., no free vars reachable from the bv-eqns)
;;; This is done as an optimization, since if we don't quantify over bvs that are hat types
;;; then the generic instantiate would not make copies of those bv-eqns (thus a saving of space) 


(defun filter-no-fvs-reachable (bv-lst eqns)
  (remove-if-not #'(lambda (bv) (type-free-vars-reachable bv eqns)) 
		 bv-lst))

(defun construct-ql (fv-lst typ eqns fn)
   (funcall fn 
	    fv-lst 
	    (filter-no-fvs-reachable (remove-if #'is-hattype (bound-vars (type-vars-reachable typ eqns :fv-expand t))) eqns) 
	    typ))

;;; type-quant converts a type to a quantified type.
;;;

(defun type-quant (fvs typ eqns)
  (assert (no-undef-bvars eqns))
  (construct-ql fvs typ eqns #'mk-quant))


(defun type-quant-all (typ eqns)
  (type-quant (free-vars (type-vars-reachable typ eqns :fv-expand t)) typ eqns))

;;; type-lambda converts a type to a lambda type 
;;;

(defun type-lambda (fvs typ eqns)
  (assert (no-undef-bvars eqns))
  (construct-ql fvs typ eqns #'mk-lambda))

;;;
;;; instantiate : creates a new instance of a quantified/lambda type
;;; Free/Bound variables are renamed,
;;; New instances of bound-variable equations are created.

(defun instantiate-ql (arg-lst typ eqns )
  (let* ( (vflag (every #'is-var arg-lst))    ;; if each arg in the arg-list is a var then 
	                                      ;; hat/compl of "new" bvars = hat/compl of bvars
;;;	  a. Create new bound-vars

	  (new-bv (mapcar #'(lambda(b) (next-bvar eqns :rl (type-rl b))) 
			  (ql-bound-vars typ))) ;; NOTE ql-bound-vars yields the bound-var list for Q/L type

;;;       b. Create a substitution mapping (free-vars -> arg-lst) and (bound-vars -> new-bvs)

	  (sub-fv (substitution-list (ql-free-vars typ) arg-lst))
	  (sub-bv (substitution-list (ql-bound-vars typ) new-bv))
	  (substn (substitution-append sub-fv sub-bv))
	  (func (if vflag #'type-alpha-substitute #'type-substitute)))
    

;;;       c. Add bound-variable equations...  for new bound-variables

       (mapc #'(lambda (var rhs) (bv-add-def var rhs eqns))
	     new-bv
	     (mapcar #'(lambda (b) (funcall func (bv-def b eqns) substn eqns))
		     (ql-bound-vars typ)))

;;;       c' ( 12/15/92 tkl ) add lower/upper bounds for new free vars

       (mapc #'(lambda (var lb ub) (add-todo-constraint (funcall func lb substn eqns)
							 var 
							 eqns)
                                   (add-todo-constraint var
							(funcall func ub substn eqns)
							eqns)
				   )
	     arg-lst
	     (mapcar #'(lambda(x) (constr-lower-bound x eqns)) (ql-free-vars typ))
	     (mapcar #'(lambda(x) (constr-upper-bound x eqns)) (ql-free-vars typ)))
		 

;;;       d. IF vflag then hat/compl of new-bvar-i = hat/compl of bvar-i
       (when vflag
	(mapc #'(lambda (e1 e2) 
		  (when (bv-lookup-hat e2 eqns)
			(bv-add-hat   e1 (type-hat   e2 eqns) eqns))
		  (when (bv-lookup-compl e2 eqns)
			(bv-add-compl e1 (type-compl e2 eqns) eqns)))
	      new-bv 
	      (ql-bound-vars typ)))

;;;       e. Finally return the instance of the Q/L type.
       (funcall func (ql-type typ) substn eqns)))


;;; * quant-instantiate : instantiates a quantified type.
;;; 

(defun quant-instantiate (arg-lst typ eqns)
  (assert (and (is-quant typ) (no-undef-bvars eqns)))
  (instantiate-ql arg-lst typ eqns))

(defun quant-generic-instantiate (typ eqns)
  (assert (is-quant typ))
  (quant-instantiate
   (mapcar #'(lambda (x) (declare (ignore x)) (next-var eqns)) (ql-free-vars typ))
   typ
   eqns))

;;; * lambda-instantiate : instantiates a lambda type.
;;;

(defun lambda-instantiate (arg-lst typ eqns)
  (assert (and (is-lambda typ) (no-undef-bvars eqns)))
  (instantiate-ql arg-lst typ eqns))
