(in-package 'types)

;;;
;;; This file implements computing the variables that appear in negitive positions in a type
;;; 
;;; Functions in this file are called from the top level type-neg function
;;; which is in a separate file "neg_tbl.lisp" (this is done so as to allow
;;; the lisp compiler to define the functions (here) in the right order.
;;; 
;;; eqns are no longer needed.
;;;
;;; Post-conditions: type-neg returns a list of variables which contains
;;;    a.  no duplicates
;;;    b.  no monotypes (hattypes)

;;; 10/6/92 tkl   created


;;; 0.  |- neg(0) = {}
;;;

(defun neg-zero(typ vs eqns)
  (declare (ignore typ eqns))
  vs)

;;; 1.  |- neg(1) = {}
;;;

(defun neg-one(typ vs eqns) 
  (declare (ignore typ eqns))
  vs)

;;;    |- neg(X) = X1,  |- neg(Y) = Y1
;;; 2. -------------------------------
;;;    |- neg(X|Y) = X1 union Y1

(defun neg-union (typ vs eqns)
  (type-neg-aux (type-t1 typ) 
		(type-neg-aux (type-t2 typ) vs eqns) 
		eqns))


;;;    |- neg(X) = X1,  |- neg(Y) = Y1
;;; 3. -------------------------------
;;;    |- neg(X*Y) = X1 union Y1

(defun neg-inter (typ vs eqns)
  (type-neg-aux (type-t1 typ) 
		(type-neg-aux (type-t2 typ) vs eqns) 
		eqns))


;;;     |- NEG(A1) = X1,..... NEG(An) = Xn
;;; 4. -------------------------------------------
;;;     |- NEG(c(A1,,,,An)) = X1 union .. union Xn
;;;

(defun neg-cons-aux (xs vs eqns)
  (if (null xs) 
      vs
    (neg-cons-aux (cdr xs) 
		  (type-neg-aux (car xs) vs eqns) 
		  eqns)))

(defun neg-cons (typ vs eqns)
  (neg-cons-aux (cons-args typ) vs eqns))


;;;    |- pos(X) = X1,  |- neg(Y) = Y1, |- neg(Z) = Z1
;;; 5. ---------------------------------------------------
;;;    |- neg(X -> Y + Z) = X1 union Y1 union Z1

(defun neg-func (typ vs eqns)
   (type-pos-aux (type-t1 typ)
		 (type-neg-aux (type-t2 typ)
			       (type-neg-aux (type-t3 typ) vs eqns)
			       eqns)
		 eqns))

;;;
;;; 6. |- neg(alpha) = {} (for all free variables alpha)
;;;

(defun neg-var (typ vs eqns)
  (declare (ignore typ eqns))
  vs)

;;;
;;; 7. |- neg(beta) = {} (for all bound variables beta)
;;;

(defun neg-bvar (typ vs eqns)
  (declare (ignore typ eqns))
  vs)

;;;
;;; 8. |- neg((not {...})) = {}
;;; 

(defun neg-not (typ vs eqns) 
  (declare (ignore typ eqns))
  vs)

;;;     |- neg(X) = X1, neg(Y) = Y1
;;; 9. ----------------------------
;;;     |- neg(X ? Y) = X1 union Y1     
;;;          
;;;			              

(defun neg-dep (typ vs eqns)
  (type-neg-aux (type-t1 typ)
		(type-neg-aux (type-t2 typ) vs eqns) 
		eqns))



;;;      |- neg(X) = X1
;;; 10. -----------------------
;;;      |- neg((val v X)) = X1
;;;

(defun neg-value (typ vs eqns)
  (type-neg-aux (value-typ typ) vs eqns))

