(in-package 'types)

;;;
;;; This file implements a conservative type (not) zero checking (i.e. checking if X != 0)
;;; 
;;; Functions in this file are called from the top level type-not-zero-for-sure function
;;; which is in a separate file "not_zero_tbl.lisp" (this is done so as to allow
;;; the lisp compiler to define the functions (here) in the right order.
;;; 

;;; E |- X : true  ==> X is definitely not-zero (under all substitutions)
;;; E |- X : false  ==> X could be zero (under some substitution)

;;; 0. A |- not-zero(0) = '()
;;;

(defun not-zero-zero(typ as eqns) 
  (declare (ignore typ as eqns))
  ())

;;; 1. A |- not-zero(1) = true
;;;

(defun not-zero-one(typ as eqns) 
  (declare (ignore typ as eqns))
  t)

;;;    A |- not-zero(X) OR not-zero(Y) 
;;; 2 --------------------------------
;;;    A |- not-zero(X|Y) 


(defun not-zero-union (X as eqns)
  (or (type-not-zero-aux (type-t1 X) as eqns) 
      (type-not-zero-aux (type-t2 X) as eqns)))

;;;    A |- not-zero(X) AND not-zero(Y) 
;;; 3 --------------------------------
;;;    A |- not-zero(X|Y) 

(defun not-zero-inter (X as eqns)
  (and (type-not-zero-aux (type-t1 X) as eqns) 
       (type-not-zero-aux (type-t2 X) as eqns)))



;;;    A |- not-zero(A1) = 1,..... not-zero(An) 
;;; 4. -------------------------------------
;;;    A |- not-zero(c(A1,,,,An)) 
;;;

(defun not-zero-cons (X as eqns)
  (every #'(lambda (arg) (type-not-zero-aux arg as eqns)) (cons-args X)))


;;; 
;;; 5. A |- not-zero(X->Y+Z) 
;;;

(defun not-zero-func (X as eqns)
  (declare (ignore X as eqns))
  t)

;;; 6. A |- not-zero(alpha)  (for all free variables alpha)
;;;

(defun not-zero-var (X as eqns)
  (declare (ignore X as eqns))
  ())

;;; 7. A |- not-zero(alpha) = () iff (member alpha A)  (for all alpha bound)
;;;

(defun not-zero-bvar (X as eqns) 
  (cond ((member (bvar-id X) as) ()) ;; Elements of as are assumed to be empty (i.e., not-zero-field ())
	((null (bv-def X eqns)) ())  ;; bound vars for which defs are not yet present could be zero.
	(t (type-not-zero-aux (bv-def X eqns) 
			      (cons (bvar-id X) as) 
			      eqns))))

;;; 8. A |- not-zero(not X) 
;;; 

(defun not-zero-not (X as eqns)  
  (declare (ignore X as eqns))
  t)

;;;     A |- not-zero(A) AND not-zero(B)                       
;;; 9. --------------------------------
;;;     A |- not-zero(A ? B) 
;;;          
;;;			              

(defun not-zero-dep (X as eqns)
  (and (type-not-zero-aux (type-t1 X) as eqns)
       (type-not-zero-aux (type-t2 X) as eqns)))

;;;      A |- not-zero(X)
;;; 10. -----------------------------
;;;     A |- not-zero(val v X) 
;;;

(defun not-zero-value (X as eqns)
  (type-not-zero-aux (value-typ X) as eqns))
  

