(in-package 'types)

;;;
;;; This file implements computing the variables that appear 
;;; in positive positions in a type
;;;
;;; Eqns are no longer needed, but are still here.
;;;
;;; Functions in this file are called from the top level type-pos function
;;; which is in a separate file "pos_tbl.lisp" (this is done so as to allow
;;; the lisp compiler to define the functions (here) in the right order.
;;;
;;; post-conditions: type-pos returns a list of variables which contains
;;;    a.  no duplicates
;;;    b.  no monotypes (hattypes)
;;;


;;; 0.  |- pos(0) = {}
;;;

(defun pos-zero(typ vs eqns)
  (declare (ignore typ eqns))
  vs)

;;; 1.  |- pos(1) = {}
;;;

(defun pos-one(typ vs eqns) 
  (declare (ignore typ eqns))
  vs)

;;;    |- pos(X) = X1,  |- pos(Y) = Y1
;;; 2. -------------------------------
;;;    |- pos(X|Y) = X1 union Y1

(defun pos-union (typ vs eqns)
  (type-pos-aux (type-t1 typ) 
		(type-pos-aux (type-t2 typ) vs eqns)
		eqns))


;;;    |- pos(X) = X1,  |- pos(Y) = Y1
;;; 3. -------------------------------
;;;    |- pos(X*Y) = X1 union Y1

(defun pos-inter (typ vs eqns)
  (type-pos-aux (type-t1 typ) 
		(type-pos-aux (type-t2 typ) vs eqns) 
		eqns))


;;;     |- POS(A1) = X1,..... POS(An) = Xn
;;; 4. -------------------------------------------
;;;     |- POS(c(A1,,,,An)) = X1 union .. union Xn
;;;

(defun pos-cons-aux (xs vs eqns)
  (if (null xs) 
      vs
    (pos-cons-aux (cdr xs) 
		  (type-pos-aux (car xs) vs eqns) 
		  eqns)))


(defun pos-cons (typ vs eqns)
  (pos-cons-aux (cons-args typ) vs eqns))

;;;    |- neg(X) = X1,  |- pos(Y) = Y1, |- pos(Z) = Z1
;;; 5. ---------------------------------------------------
;;;    |- pos(X -> Y + Z) = X1 union Y1 union Z1

(defun pos-func (typ vs eqns)
  (type-neg-aux (type-t1 typ) 
		(type-pos-aux (type-t2 typ) 
			      (type-pos-aux (type-t3 typ) vs eqns)
			      eqns)
		eqns))
	 
		

;;;
;;; 6. |- pos(alpha) = {alpha} (for all free variables alpha)
;;;

(defun pos-var (typ vs eqns)
  (declare (ignore eqns))
  (if (member typ vs :test #'var-=)
      vs
      (cons typ vs)))

;;;
;;; 7. |- pos(beta) = {beta} (for all bound variables beta)
;;;

(defun pos-bvar (typ vs eqns)
  (declare (ignore eqns))
  (cond
   ((is-hattype typ) vs)
   ((member typ vs :test #'var-=) vs)
   (t (cons typ vs))))

;;;
;;; 8. |- pos((not {...})) = {}
;;; 

(defun pos-not (typ vs eqns) 
  (declare (ignore typ eqns))
  vs)

;;;     |- pos(X) = X1, pos(Y) = Y1
;;; 9. ----------------------------
;;;     |- pos(X ? Y) = X1 union Y1     
;;;          
;;;			              

(defun pos-dep (typ vs eqns)
  (type-pos-aux (type-t1 typ) 
		(type-pos-aux (type-t2 typ) vs eqns) 
		eqns))



;;;      |- pos(X) = X1
;;; 10. -----------------------
;;;      |- pos((val v X)) = X1
;;;

(defun pos-value (typ vs eqns)
  (type-pos-aux (value-typ typ) vs eqns))

