(in-package 'types)
      
;;;
;;;=========================================================
;;;
;;; Functions for printing the contents of a tbl.
;;;
;;;=========================================================
 
(defun print-eqn-table (table lhsfun rhsfun eqns str &key (format "~a = ~a~%")) 
  (map-table #'(lambda (x y) 
		 (print-equation (funcall lhsfun x) 
				 (funcall rhsfun y)
				 eqns
				 str
				 :fmt format))
	     table))

(defun print-constraints (table str fmt)
  (map-table
   #'(lambda (v typ)
       (print-equation (mk-var v) typ str :fmt fmt))
   table))

(defun id (x) x)

;;;
;;; Maps an identifier to a variable.
;;; An identifier is mapped to a bound variable
;;; if that identifier is the id of a bound variable 
;;; in the list of bv-eqns or if the id is in the
;;; list of undefined bound variables.  Otherwise,
;;; it is mapped to a free variable.
;;;
(defun id-to-var (vid tbl)
  (if (or (bv-def (mk-bvar vid) tbl)
	  (member vid (tbl-undef-bvars tbl)))
      (mk-bvar vid)
    (mk-var vid)))

;;;
;;; Maps an association list to a lookup table.
;;;
(defun assoc-to-table (l)
  (let ((x (mk-lookup-table)))
    (mapc #'(lambda (y)
	      (table-add (var-id (car y)) (cdr y) x))
	  l)
    x))

(defun print-pending (tbl)
  (print-pending-aux (queue-to-list (tbl-pending tbl)) tbl))

(defun print-pending-aux (l tbl)
  (mapc #'(lambda (x)
	    (let* ((kind (first x))
		   (var (second x)))
	      (case kind
		    (:hat (format t "~a = ^~a~%" 
				  (pt (third x) tbl nil) 
				  (pt var tbl nil)))
		    (:compl (format t "~a = \~~a~%" 
				    (pt (third x) tbl nil) 
				    (pt var tbl nil)))
		    (:inter (format t "~a = ~a  &  ~a~% " 
				    (pt (fourth x) tbl nil)
				    (pt var tbl nil)
				    (pt (third x) tbl nil)))
		    (t (error "Bad case in print-pending")))))
	l))


(defun print-tbl (tbl &optional (str t) &key (named-types nil) (hat-eqns nil)
		      (compl-eqns nil) (inter-eqns nil) )
  (when named-types
	(format str "Named Types~%")
	(map-table #'(lambda (x y)
		       (format str "~a :: ~a~%" x (pt y tbl nil)))
		   (tbl-named-types tbl)))

  (format str "~%~%Current Variable: ~a" (tbl-var tbl))

  (format str "~%~%Bound Variable Equations: ~%")
  (print-eqn-table (tbl-bv-eqns tbl) #'mk-bvar #'id tbl str)

  (format str "~%~%Unprocessed Bound Variable Equations: ~%")
  (print-eqn-table (assoc-to-table (tbl-todo-bv-eqns tbl))
		   #'mk-bvar
		   #'id
		   tbl 
		   str)

  (format str "~%~%Undefined Bound Variables: ~a" 
	  (mapcar #'(lambda (typ) (pt typ tbl nil)) 
		  (mapcar #'mk-bvar (tbl-undef-bvars tbl))))

  (when hat-eqns
	(format str "~%~%Hat Equations: ~%")
	(print-eqn-table (tbl-hat-eqns tbl) #'mk-bvar #'id tbl str :format "^~a = ~a~%"))

  (when compl-eqns 
	(format str "~%~%Complement Equations: ~%")
	(print-eqn-table (tbl-compl-eqns tbl) #'mk-bvar #'id tbl str :format "~~~a = ~a~%"))

  (when inter-eqns
	(format str "~%~%Intersection Equations: ~%")
	(print-eqn-table (tbl-inter-eqns tbl)
			 #'(lambda (pair) (mk-inter (car pair) (cdr pair)))
			 #'id
			 tbl 
			 str))

  (format str "~%~%Intersection Memos: ~%")
  (map-table #'(lambda (x ys)
		 (format str "~a <= ~a~%"
			 (pt (mk-bvar x) tbl nil)
			 (mapcar #'(lambda (y) (pt y tbl nil)) ys)))
	     (tbl-bv-subset tbl))

  (format str "~%~%Pending Equations: ~%")
  (print-pending tbl)

  (format str "~%~%Upper Bounds: ~%")
  (print-eqn-table (tbl-upper-bounds tbl) 
		   #'(lambda (vid) (id-to-var vid tbl))
		   #'id
		   tbl 
		   str
		   :format "~a <= ~a~%")

  (format str "~%~%Lower Bounds: ~%")
  (print-eqn-table (tbl-lower-bounds tbl)
		     #'(lambda (vid) (id-to-var vid tbl))
		     #'id
		     tbl 
		     str
		     :format "~a >= ~a~%")

  (format str "~%~%Todo Constraints: ~%")
  (let ((x (mk-lookup-table)))
    (mapc #'(lambda (y) (table-add (car y) (cdr y) x))
	  (tbl-todo-constraints tbl))
    (print-eqn-table x
		     #'id
		     #'id
		     tbl 
		     str
		     :format "~a <= ~a~%"))

t)



	
