(in-package 'types)

;;;
;;; This file implements a simple "spit" printer for types.
;;; A more sophisticated pretty printer is in the file
;;; pretty_printer_tbl.lisp.
;;; This printer is basd on a printer for FL written by Paul Tucker.
;;;

(defconstant *one-glyph* "1")
(defconstant *zero-glyph* "0") 
(defconstant *union-glyph* " | ")
(defconstant *inter-glyph* " & ")
(defconstant *func-glyph1* " -> ")
(defconstant *func-glyph2* " + ")
(defconstant *not-glyph* "NOT")
(defconstant *open-not-glyph* "{")
(defconstant *close-not-glyph* "}")
(defconstant *open-value-glyph* "[")
(defconstant *close-value-glyph* "]")
(defconstant *dep-glyph* " ? ")
(defconstant *quant-glyph* "forall ")
(defconstant *lambda-glyph* "\\")
(defconstant *dot-glyph* ".")
(defconstant *var-glyph* "`")
(defconstant *bvar-glyph* "$")
(defconstant *allfunc-glyph* "FUNCS")


(defun open-paren (s) (format s "("))
(defun close-paren (s) (format s ")"))

;;;
;;; paren? takes the type of the parent and the type of the current
;;; expression being printed and decides whether parentheses are required.
;;; Currently only a few distinctions are made, and some print functions
;;; explicitly pass nil to their children to prevent parens from being inserted.
;;;
;;; paren? could eventually be driven by a big table that would decide
;;; more precisely where to put parentheses.
;;;
(defun paren? (parent thistyp) 
  (cond
   ((or
     (null parent)
     (and (equal parent union-type)
	  (equal thistyp union-type))
     (and (equal parent inter-type)
	  (equal thistyp inter-type)))
    nil)
   (t t)))

;;;
;;; print-list prints a list of "things" separated by commas.
;;; The function func should have functionality
;;;    stream x paren-flag x thing -> anything
;;;
(defun print-paren-list (s func list eqns &key substn)
  (open-paren s)
  (print-list s func list eqns :substn substn)
  (close-paren s))

(defun print-list (s func list eqns &key substn)
  (when (not (null list))
	(funcall func s nil (car list) eqns :substn substn)
	(if (not (null (cdr list)))
	    (format s ", "))
	(print-list s func (cdr list) eqns :substn substn)))

(defun print-one (s p typ eqns &key substn)
  (declare (ignore p typ eqns substn))
  (format s "~a" *one-glyph*))


(defun print-zero (s p typ eqns &key substn)
    (declare (ignore p typ eqns substn))
    (format s "~a" *zero-glyph*))

(defun print-union (s p typ eqns &key substn)
  (let ((paren (paren? p union-type)))
    (when paren (open-paren s))
    (print-type s union-type (type-t1 typ) eqns :substn substn)
    (format s "~a"  *union-glyph*)
    (print-type s union-type (type-t2 typ) eqns :substn substn)
    (when paren (close-paren s))))


(defun print-inter (s p typ eqns &key substn)
  (let ((paren (paren? p inter-type)))
    (when paren (open-paren s))
    (print-type s inter-type (type-t1 typ) eqns :substn substn)
    (format s "~a"  *inter-glyph*)
    (print-type s inter-type (type-t2 typ) eqns :substn substn)
    (when paren (close-paren s))))


(defun print-func (s p typ eqns &key substn)
  (if (equalp typ (type-all-funcs))
      (format s "~a" *allfunc-glyph*)
    (let ((paren (paren? p func-type)))
      (when paren (open-paren s))
      (print-type s func-type (type-t1 typ) eqns :substn substn)
      (format s "~a"  *func-glyph1*)
      (print-type s func-type (type-t2 typ) eqns :substn substn)
      (format s "~a"  *func-glyph2*)
      (print-type s func-type (type-t3 typ) eqns :substn substn)
      (when paren (close-paren s)))))


(defun print-cons (s p typ eqns &key substn)
  (declare (ignore p))
  (format s "~a" (cons-name-get (cons-id typ) eqns))
  (when (not (= (cons-arity-get (cons-id typ) eqns) 0))
	(open-paren s)
	(print-list s #'print-type (cons-args typ) eqns :substn substn)
	(close-paren s)))


(defun print-var (s p typ eqns &key substn)
  (declare (ignore p eqns))
  (if (substitution-get typ substn)
      (format s "~a" (substitution-get typ substn))
      (format s "~a~d" *var-glyph* (var-id typ))))


(defun print-bvar (s p typ eqns &key substn)
  (declare (ignore p eqns))
  (if (substitution-get typ substn)
      (format s "~a" (substitution-get typ substn))
      (format s "~a~d" *bvar-glyph* (bvar-id typ))))


(defun print-not (s p typ eqns &key substn)
  (declare (ignore p ))
  (format s "~a~a" *not-glyph* *open-not-glyph*)
  (print-list s 
	      #'(lambda (s p x &rest y)   
		  (declare (ignore p y))
		  (format s "~a" (cons-name-get x eqns)))
	      (not-list typ)
	      eqns
	      :substn substn)
  (format s "~a" *close-not-glyph*))
	      


(defun print-dep (s p typ eqns &key substn)
  (let ((paren (paren? p dep-type)))
    (when paren (open-paren s))
    (print-type s dep-type (type-t1 typ) eqns :substn substn)
    (format s "~a"  *dep-glyph*)
    (print-type s dep-type (type-t2 typ) eqns :substn substn)
    (when paren (close-paren s))))


(defun print-value (s p typ eqns &key substn)
  (declare (ignore p))
  (format s "~a" *open-value-glyph*)
  (format s "~a" (value-val typ))
  (format s ", ")
  (print-type s nil (value-typ typ) eqns :substn substn)
  (format s "~a" *close-value-glyph*))


;;; quantified types
(defun print-quant (s p typ eqns &key substn)
  (declare (ignore p))
;; 1/5 tkl :: added test to see if quant type is (for-all () typ)
;; if so, just print typ instead of (for-all () typ)
  (if (and (null (quant-vars typ)) 
	   (null (quant-bvars typ)))
      (print-type s nil (quant-typ typ) eqns :substn substn)
    (progn
      (format s "~a" *quant-glyph*)
      (print-paren-list s
			#'(lambda (s p v &rest x) 
			    (declare (ignore x))
			    (print-type s p v eqns :substn substn))
			(quant-vars typ)
			eqns
			:substn substn)
      (format s "~a" *dot-glyph*)
      (print-type s nil (quant-typ typ) eqns :substn substn))))


;;; type functions
(defun print-lambda (s p typ eqns &key substn)
  (declare (ignore p))
  (format s "~a" *lambda-glyph*)
  (print-paren-list s
		    #'(lambda (s p v &rest x) 
			(declare (ignore x))
			(print-type s p v eqns :substn substn))
		    (lambda-vars typ)
		    eqns
		    :substn substn)
  (format s "~a" *dot-glyph*)
  (print-type s nil (lambda-typ typ) eqns :substn substn))
