(in-package 'ptools)

;;;
;;; various Lisp utilities, written by Paul Tucker.
;;;
(export 
 '(all-but-keys				; association list filter
   all-keys				; alist filter
   and-func				; test logical list
   case-eval				; case macro with evaulated keylists
   clobber-file				; macro for overwriting a file
   compress-list			; logically compress sequence
   default-output-stream		; do t & nil synonyming
   function-name			; name of a defined function
   group-equiv				; find equivalent subsets
   iota					; APL function--almost
   len1-list?				; true of lists of length one
   lotsof				; make infinite list of arg
   mapand				; fast logical mapper
   mapor				; fast logical mapper
   remove-from-list			; destructively pinch out element
   right-now				; date & time in nice format
   print-in-columns			; format list of syms & strings
   sym-eql				; print-name equivalence
   take					; macro for APL like function
   ))

(defun all-keys (keys alist)
  ;; Return list of all entries from alist whose key is a member
  ;; of keys.
  (let ((result nil))
    (declare (type cons result))
    (dolist (item alist (nreverse result))
	    (if (member (car item) keys)
		(push item result)))))

(defun all-but-keys (keys alist)
  ;; Return list of all enties from alist whose key is not a member
  ;; of keys.
  (let ((result nil))
    (declare (type cons result))
    (dolist (item alist (nreverse result))
	    (if (not (member (car item)
			     keys))
		(push item result)))))

(defun and-func (l)
  (declare (type list l))
  (dolist (i l 't)
	  (if (not i) (return-from and-func nil))))

(defmacro case-eval (key &rest forms)
  ;; Evaluate the keyform begining each case at macro-expand time.
  (do ((newforms nil)
       (oforms forms (cdr oforms)))
      ((null oforms) (setq forms (nreverse newforms)))
      (push (cons (eval (caar oforms)) (cdar oforms)) newforms))
  `(case ,key ,@forms))

(defmacro clobber-file ((stream filename &rest options) &rest forms)
  `(with-open-file (,stream ,filename :direction :output :if-exists :supersede
			    ,@options)
		   ,@forms))

(defun compress-list (l1 l2)
  ;; Eliminate every element in l1 corresponding to nil element in l2.
  ;; Nondestructive.
  (declare (type list l1 l2))
  (if l1
      (if (car l2)
	  (cons (car l1) (compress-list (cdr l1) (cdr l2)))
	(compress-list (cdr l1) (cdr l2)))))

(defun default-output-stream (stream)
  ;; Many Common Lisp output functions recognize NIL and T
  ;; in place of a stream argument as standing for
  ;; string output and standard output, respectively
  (if (null stream)
      (make-string-output-stream)
    (if (sym-eql stream 't) 
	*standard-output*
      stream)))

(defun function-name (x)
  (if (and (functionp x)
	   (not (symbolp x)))
#+:LUCID
      (progn
	(setq x (lucid::procedure-ref x lucid::procedure-symbol))
	(if (atom x)
	    (values x :compiled)
	  (values (if (symbolp (cadr x))
		      (cadr x)
		    nil)
		  :interpreted)))
#+(not :LUCID)
      (values "Unknown" nil)))

(defun group-equiv (l prev equiv)
  ;; L is a list, prev and equiv are predicates used to sort the list,
  ;; then find equivalent subgroups within it.  
  ;; Returns a list of equivalent sub-lists.
  (when l
	(setq l (sort (copy-list l) prev))
	(do ((res nil)
	     (sublist (list (car l)))
	     (key (car l))
	     (rest (cdr l) (cdr rest)))
	    ((null rest) (cons sublist res))
	    (declare (type list rest))
	    (if (not (funcall equiv key (car rest)))
		(progn
		  (setq res (cons sublist res))
		  (setq sublist (list (car rest)))
		  (setq key (car rest)))
	      (push (car rest) sublist)))))
		

(defun iota (maxint &optional (type :list))
  ;; Generates a sequence of integers from 0 up to but not including
  ;; maxint.
  ;; Type is an optional keyword indicating that result should be a
  ;; list or an array.  Default is a list.
  (if (or (plusp maxint) (zerop maxint))
      (ecase type
	     (:list (let ((r nil))
		      (declare (type list r))
		      (do ((i 0 (incf i)))
			  ((>= i maxint) (nreverse r))
			  (setf r (cons i r)))))
	     (:array (let ((r (make-array maxint :element-type 'fixnum)))
		       (declare (type simple-array r))
		       (do ((i 0 (incf i)))
			   ((>= i maxint) r)
			   (setf (aref r i) i)))))
    (error "Argument to iota must be a non-negative integer")))

(defun len1-list? (l)
  (and (consp l) (null (cdr (the cons l)))))

(defun lotsof (item)
  ;; Create a circular (infinite) list of item
  (let ((cell (cons item nil)))
    (rplacd cell cell)))

;;; Note, the CL function EVERY subsumes this function
(defun mapand (func &rest stuff)
  ;; AND mapper---takes function of arity N and N argument lists.
  ;; Applies function to corresponding tuples of arguments in order.
  ;; Terminates with nil at first non-nil application result, 
  ;; or with 't when shortest list is exhausted.
  ;; Note that (mapand #'foo ()) evaluates to 't.
  (do ((items
	(dolist (x stuff stuff)
		(if (null x)
		    (return-from mapand 't)))
	(do ((sublist items (cdr sublist)))
	    ((null sublist) items)
	    (declare (cons sublist))
	    (if (null (car (rplaca sublist (cdar sublist))))
		(return-from mapand 't)))))
      ((not (apply func (mapcar #'car items))))
      (declare (cons items))))

;;; Note, the CL function SOME subsumes this function.
(defun mapor (func &rest stuff)
  ;; OR mapper---takes function of arity N and N argument lists.
  ;; Applies function to corresponding tuples of arguments in order.
  ;; Returns value of first non-nil application, or nil when shortest argument list
  ;; is exhausted.
  (do ((items 
	(dolist (x stuff stuff)
		(if (null x)
		    (return-from mapor nil)))
	(do ((sublist items (cdr sublist)))
	    ((null sublist) items)
	    (declare (cons sublist))
	    (if (null (car (rplaca sublist (cdar sublist))))
		(return-from mapor nil))))
       result)
      (nil)
      (declare (cons items))
      (setq result (apply func (mapcar #'car items)))
      (if result (return-from mapor result))))

(defun remove-from-list (item list)
  ;; Destructively modifies list to pinch out first element eq to item.
  ;; Skips over first element.
  ;; Returns item if removed, nil otherwise.
  (if (null (cdr list))
      nil
    (if (eq item (cadr list))
	(progn
	  (rplacd list (cddr list))
	  item)
      (remove-from-list item (cdr list)))))

(defun right-now ()
  ;; Return pretty version of current time, as a string.
  (let (second minute hour date month year day)
    (multiple-value-setq (second minute hour date month year day)
			 (get-decoded-time))
    (setq month (nth month '("Nonesuch" "January" "February" "March" 
			     "April" "May" "June" "July" "August"
			     "September" "October" "November" "December")))
    (setq day (nth day '("Monday" "Tuesday" "Wednesday" "Thursday" "Friday"
			 "Saturday" "Sunday")))
    (format nil "~d:~2,'0d:~2,'0d ~a ~a ~d, ~d" 
	    hour minute second day month date year))) 

(defun print-in-columns (item-list &optional (stream *standard-output*)
				   &key (llen 80) (space 2)
				   (cols nil) (direction :hor)
				   (p-fun #'princ))
  ;; Argument must be a list of strings and/or symbols.
  ;; They will be printed to the argument stream, or standard output,
  ;; in as many columns as fit nicely within the indicated bounds.
  ;; If direction is :hor, items are printed in row major order.
  ;; Else they are printed in column major.
  (let ((max-item 0) (item-count 0) column-width rows item)
    (dolist (item item-list)
	    (setq max-item (max max-item (length (string item))))
	    (incf item-count))
    (setq stream (default-output-stream stream))
    (if (not cols)
	(setq cols (max 1 (floor (/ llen (+ space max-item))))))
    (setq rows (ceiling (/ item-count cols)))
    (setq column-width (+ max-item space))
    (if (not (eq direction :hor))
	(setq item-list (transpose-list item-list rows cols)))
    (block loop
	   (dotimes (r rows)
		    (fresh-line stream)
		    (dotimes (c cols)
			     (if (null item-list) (return-from loop))
			     (setq item (pop item-list))
			     (funcall p-fun item stream)
			     (if (/= c (1- cols))
				 (dotimes (i (- column-width (length (string item))))
					  (write-char #\space stream))))
		    (terpri stream))))
  (values))

(defun transpose-list (item-list rows cols)
  ;; Aux function for print-in-columns
  (let ((array (make-array rows :initial-element nil))
	(result nil))
    (block loop
	   (dotimes (c cols)
		    (dotimes (r rows)
			     (if (null item-list) (return-from loop))
			     (push (pop item-list) (aref array r)))))
    (dotimes (r rows)
	     (if (= r 0)
		 (setq result (nreverse (aref array 0)))
	       (nconc result (nreverse (aref array r)))))
    result))
	    
(defun sym-eql (s1 s2)
  (and (symbolp s1)
       (symbolp s2)
       (string-equal (symbol-name s1) (symbol-name s2))))

(defmacro take (n s)
  `(subseq ,s 0 ,n))
