(in-package 'types)

;;;
;;; vars-reachable : type * ans * eqns * assumptions * (type * ans * eqns -> ans) -> ans
;;;
;;; A general reachability function, that takes a function as an argument.
;;;
;;; Used to implement a variety of reachability analyses.
;;;
;;; ans   is the current "answer"; it is threaded through the computation
;;; assumptions  are the current set of variables (free or bound) that are being analyzed
;;; fv-expand  is true if free variables upper and lower bounds should be analyzed
;;; bv-expand  is true if bound variables upper and lower bounds should be analyzed
;;;
 
;;;
;;; E,AS,A |- vars-reachable(0 or 1 or not{}) = A
;;;

(defun vars-reachable-id (typ eqns as fn fv-expand bv-expand)
  (declare (ignore typ eqns fn fv-expand bv-expand))
  as)



;;;  E,A |- vars-reachable(X) = A1;  E,A |- vars-reachable(Y) = A2
;;;  ----------------------------------------------------------------
;;;   E,A |- vars-reachable(X | Y) = A2
;;;

(defun vars-reachable-union (typ eqns as fn fv-expand bv-expand)
  (type-vars-reachable-aux 
   (type-t2 typ) 
   eqns
   (type-vars-reachable-aux (type-t1 typ) 
			    eqns 
			    as
			    fn
			    fv-expand
			    bv-expand)
   fn
   fv-expand
   bv-expand))


;;;    E,A |- vars-reachable(X) = A1;  E,A1 |- vars-reachable(Y)= A2
;;;    -------------------------------------------------------------------
;;;    E, A |- vars-reachable(X * Y) = A2
;;;

(defun vars-reachable-inter (typ eqns as fn fv-expand bv-expand)
  (vars-reachable-union typ eqns as fn fv-expand bv-expand))


;;;  for all i 1<=i<=n, E,AS,Ai-1 |- vars-reachable(Xi) = Ai   where A0 = A
;;;  -------------------------------------------------------
;;;  E,AS,A |- vars-reachable(c(X1,....Xn)) = An
;;;

(defun vars-reachable-cons-aux (args eqns as fn fv-expand bv-expand)
  (if (null args)
      as
    (let ((nas (type-vars-reachable-aux (car args) eqns as fn fv-expand bv-expand)))
      (vars-reachable-cons-aux (cdr args) eqns nas fn fv-expand bv-expand))))

(defun vars-reachable-cons (typ eqns as fn fv-expand bv-expand)
  (vars-reachable-cons-aux (cons-args typ) eqns as fn fv-expand bv-expand))

;;;   for all i 1 <= i <= 3, E,Ai-1 |- vars-reachable(Xi) = Ai  
;;;   -----------------------------------------------------------
;;;    E,A0 |- vars-reachable(X1 -> X2 + X3) = A3
;;;

(defun vars-reachable-func (typ eqns as fn fv-expand bv-expand)
  (type-vars-reachable-aux (type-t3 typ)
			   eqns
			   (type-vars-reachable-aux (type-t2 typ)
						    eqns
						    (type-vars-reachable-aux (type-t1 typ)
									     eqns
									     as
									     fn
									     fv-expand
									     bv-expand)
						    fn
						    fv-expand 
						    bv-expand)
			   fn
			   fv-expand 
			   bv-expand))

;;;
;;; E,A |- vars-reachable(a) = fn(a,A,E)    if a is a free variable  and fv-expand = nil
;;;
;;; E,A u { a } |- vars-reachable(a) = A   if fv-expand not= nil
;;;
;;; E, A |- vars-reachable(a) = fn(a,A,E)   if fv-expand = nil
;;;     
;;; E,A       |- fn(a,A,E) = A1,     
;;; E,A u {a} |- vars-reachable(L) = A1,   where L <= a <= U   in E
;;; E,A1      |- vars-reachable(U) = A2    if a is a free variable  and fv-expand not= nil
;;; ---------------------------------------
;;; E,AS,A |- vars-reachable(a) = A2
;;;
(defun vars-reachable-var (typ eqns as fn fv-expand bv-expand)
  (cond ((member typ as :test #'var-=) 
	 as)
	(fv-expand
	 (type-vars-reachable-aux (constr-lower-bound typ eqns) 
				  eqns
				  (type-vars-reachable-aux  (constr-upper-bound typ eqns) 
							    eqns 
							    (funcall fn typ eqns as)
							    fn
							    fv-expand 
							    bv-expand)
				  fn
				  fv-expand 
				  bv-expand))
	(t (funcall fn typ eqns as))))
    
;;;
;;; 
;;; E,A u {a} |- vars-reachable(a) = A u {a}              a is a bound variable  and bv-expand not= nil
;;; 
;;; E,A, A |- vars-reachable(a) = fn(a,A,E)               if bv-expand = nil
;;;
;;; E u { a = X},AS, A |- fn(a,A,E) = A1,                  if bv-expand not= nil
;;; E u { a = X},AS u {a}, A1 |- vars-reachable(X) = A2
;;; ---------------------------------------------------
;;; E u { a = X},AS, A |- vars-reachable(a) = A2
;;;
;;;
;;; NOTE : if (a = rhs) not-in E then return A
;;;

(defun vars-reachable-bvar (typ eqns as fn fv-expand bv-expand)
  (cond ((or (member typ as :test #'var-=)
	     (null (bv-def typ eqns)))
	 as)
	(bv-expand
	 (type-vars-reachable-aux (bv-def typ eqns) 
				  eqns 
				  (funcall fn typ eqns as)
				  fn
				  fv-expand 
				  bv-expand))
	(t
	 (funcall fn typ eqns as))))


;;;    E,AS,A  |- vars-reachable(B) = A1 
;;;    E,AS,A1 |- vars-reachable(C) = A2
;;;    ------------------------------------------------------
;;;    E,AS,A  |- vars-reachable(B?C) = A2
;;;

(defun vars-reachable-dep (typ eqns as fn fv-expand bv-expand)
  (type-vars-reachable-aux (type-t1 typ) 
			   eqns 
			   (type-vars-reachable-aux (type-t2 typ) 
						    eqns 
						    as 
						    fn
						    fv-expand 
						    bv-expand)
			   fn
			   fv-expand 
			   bv-expand))

;;;     E,AS,A |- vars-reachable(X) = A1
;;;     --------------------------------------------
;;;     E,As,A |- vars-reachable(val(v,X)) = A1
;;;

(defun vars-reachable-value (typ eqns as fn fv-expand bv-expand)
  (type-vars-reachable-aux (value-typ typ) 
			   eqns 
			   as 
			   fn
			   fv-expand 
			   bv-expand))


;;;     E, AS, A |- vars-reachable(X) = L
;;;     ----------------------------------------
;;;     E, AS, A |- vars-reachable (for-all v. X) = L

(defun vars-reachable-quant (typ eqns as fn fv-expand bv-expand)
  (type-vars-reachable-aux (quant-typ typ) 
			   eqns 
			   as 
			   fn
			   fv-expand 
			   bv-expand))

;;;     E, A |- vars-reachable(A) = L
;;;     ----------------------------------------
;;;     E, A |- vars-reachable (lambda x. A) = L

(defun vars-reachable-lamb (typ eqns as fn fv-expand bv-expand)
  (type-vars-reachable-aux (lambda-typ typ) 
			   eqns 
			   as 
			   fn
			   fv-expand 
			   bv-expand))
