(in-package 'types)

;;;
;;; This file implements a conservative algorithm for testing
;;; whether one type is contained in another in ALL substitutions.
;;;

(defun pair-assump-empty () nil)

(defun pair-assump-add (l r as)
  (cons (cons l r) as))

(defun pair-assump-has (l r as)
  (if (null as)
      nil
    (let ((a (car as)))
      (or (and (eq l (car a)) (eq r (cdr a)))
	  (pair-assump-has l r (cdr as))))))


;;; Functions that just return true and false.
;;;
(defun isless (&rest x) (declare (ignore x)) t)
(defun notless (&rest x) (declare (ignore x)) nil)

;;;
;;;   E, A |- X <= Z,   Y <= Z
;;;   ------------------------
;;;      E, A |- X | Y <= Z
;;;
(defun uless (union any eqns as sols)
  (and (type-<=-aux (type-t1 union) any eqns as sols)
       (type-<=-aux (type-t2 union) any eqns as sols)))

;;;
;;;   E, A |- X <= Y,   X <= Z
;;;   ------------------------
;;;      E, A |- X <= Y & Z
;;;
(defun lessin (any inter eqns as sols)
  (and (type-<=-aux any (type-t1 inter) eqns as sols)
       (type-<=-aux any (type-t2 inter) eqns as sols)))


;;;
;;;         E, A |- for all i, X_i <= Y_i
;;;     ----------------------------------------
;;;      E, A |- c(X_1,...,X_n) <= c(Y_1,...,Y_n)
;;;
(defun consless-aux (l1 l2 eqns as sols)
  (or (null l1) 
      (and (type-<=-aux (car l1) (car l2) eqns as sols)
	   (consless-aux (cdr l1) (cdr l2) eqns as sols))))

(defun consless (c1 c2 eqns as sols)
  (and (= (cons-id c1) (cons-id c2))
       (consless-aux (cons-args c1) (cons-args c2) eqns as sols)))


;;;
;;;        E, A |- X <= Y
;;;     -------------------
;;;      E, A |- X <= Y | Z
;;;
;;; conservative
;;;
(defun lessun (any union eqns as sols)
  (or (type-<=-aux any (type-t1 union) eqns as sols)
      (type-<=-aux any (type-t2 union) eqns as sols)))

;;;
;;;        E, A |- X <= Y
;;;     -------------------
;;;      E, A |- X & Z <= Y 
;;;
;;; conservative
;;;
(defun inless (inter any eqns as sols)
  (or (type-<=-aux (type-t1 inter) any eqns as sols)
      (type-<=-aux (type-t2 inter) any eqns as sols)))



;;;
;;;      E, A |- Y_1 <= X_1, X_2 <= Y_2, X_3 <= Y_3
;;;     ---------------------------------------------
;;;      E, A |- X_1 -> X_2 + X_3 <= Y_1 -> Y_2 + Y_3
;;;
(defun funcless (f1 f2 eqns as sols)
  (or (equalp f2 (type-all-funcs))
      (and (type-<=-aux (type-t1 f2) (type-t1 f1) eqns as sols)
	   (type-<=-aux (type-t2 f1) (type-t2 f2) eqns as sols)
	   (type-<=-aux (type-t3 f1) (type-t3 f2) eqns as sols))))

;;;
;;;      E, A |- X <= Z
;;;     -------------------
;;;      E, A |- X ? Y <= Z
;;;
;;; conservative
;;;
(defun depless (dep any eqns as sols)
  (type-<=-aux (type-t1 dep) any eqns as sols))


;;;
;;; E, A |- c(X_1,...,X_n) <= NOT(B)   if c not in B 
;;;
(defun conslnot (c n eqns as sols)
  (declare (ignore eqns as sols))
  (not (member (cons-id c) (not-list n))))

;;;
;;; E, A u { a <= X } |- a <= X
;;;
;;; E u { L <= a <= U }, A u { a <= X } |- U <= X
;;; ---------------------------------------------
;;; E u { L <= a <= U }, A |- a <= X  in all solutions of E
;;;
(defun varless (v1 any eqns as sols)
  (and sols
       (or
	(pair-assump-has v1 any as)
	(type-<=-aux (constr-upper-bound v1 eqns) 
		     any 
		     eqns 
		     (pair-assump-add v1 any as) 
		     sols))))


;;;
;;; E, A u { X <= a } |- X <= a
;;;
;;; E u { L <= a <= U }, A u { X <= a } |- X <= L
;;; ---------------------------------------------
;;; E u { L <= a <= U }, A |- X <= a   in all solutions of E
;;;
(defun lessvar (any v1 eqns as sols)
  (and
   sols
   (or
    (pair-assump-has any v1 as)
    (type-<=-aux any 
		 (constr-lower-bound v1 eqns) 
		 eqns 
		 (pair-assump-add any v1 as) 
		 sols))))

;;;
;;; E, A |- a <= a    NOTE : a is a free variable.
;;;
(defun varlvar (v1 v2 eqns as sols)
  (or (= (var-id v1) (var-id v2))
      (and sols
	   (if (var-<= v1 v2)
	       (lessvar v1 v2 eqns as sols)
	       (varless v1 v2 eqns as sols)))))

;;;
;;; E, A |- a <= a   NOTE : a is a bound variable
;;;
;;; E, A union {a <= Z} |- a <= Z
;;;
;;; E union { a = b&c },A |- a <= b
;;;
;;; If a = X is a contractive equation in E, then
;;;
;;; E union { a = X }, A union {a <= Z} |- X <= Z
;;; ---------------------------------------------
;;;   E union { a = X }, A |- a <= Z
;;;
(defun bvless (v1 any eqns as sols)
  (or
   (pair-assump-has v1 any as)
   (bv-less-by-intersection v1 any eqns as sols)
   (let ((bvd (bv-def v1 eqns)))
     (and bvd
	  (type-<=-aux
	   bvd                          ;;; bv-def : var-id * eqns -> type
	   any                          ;;; bv-def gets the rhs of an eqn (v = rhs) which occurs in eqns
	   eqns
	   (pair-assump-add v1 any as)
	   sols)))))

;;;
;;; special case for intersection <= bound variable
;;;
;;;  x is a bound var and x = y
;;;  -------------------------------
;;;  E, A |-  a1 & ... & an & x <= y

(defun get-last-conjunct (x)
  (if (is-inter x)
      (get-last-conjunct (type-t2 x))
      x))

(defun inlessbv (in v eqns as sols)
   (let ((l (get-last-conjunct in)))
     (if (and (is-bvar l)
	      (var-= l v))
	 t
       (lessbv in v eqns as sols))))

;;;
;;; E, A |- a <= a
;;;
;;; E, A union Z <= a |- Z <= a
;;;
;;; If a = X is a contractive equation, then
;;;
;;; E, A union Z <= a |- Z <= X
;;; ---------------------------
;;;   E, A |- Z <= a
;;;
(defun lessbv (any v1 eqns as sols)
  (or
   (pair-assump-has any v1 as)
   (let ((bvd (bv-def v1 eqns)))
     (and bvd
	  (type-<=-aux
	   any
	   bvd      ;;; bv-def : as before : var * eqns -> type
	   eqns
	   (pair-assump-add any v1 as)
	   sols)))))
       

(defun bvlessbv (v1 v2 eqns as sols)
  (let ((v1id (var-id v1))
	(v2id (var-id v2)))  
    (or (= v1id v2id)
	(pair-assump-has v1id v2id as)   
	(bv-less-by-intersection v1 v2 eqns as sols)
	(let ((bvd1 (bv-def v1 eqns))
	      (bvd2 (bv-def v2 eqns)))
	  (and bvd1 
	       bvd2
	       (type-<=-aux
		 bvd1
		 bvd2    ;;; bv-def : as before : var * eqns -> type
		 eqns
		 (pair-assump-add v1id v2id as)
		  sols))))))
;;;
;;;       E, A |-  X <= Y
;;; --------------------------------
;;;   E, A |-  X <= < value2, Y >
;;;
(defun lessv (v1 v2 eqns as sols)
  (type-<=-aux v1  (value-typ v2) eqns as sols))

;;;
;;;   E, A |-  X <= Y 
;;; ---------------------------
;;;   E, A |- < value, X > <= Y
;;;
(defun vless (v1 any eqns as sols)
  (type-<=-aux (value-typ v1) any eqns as sols))

;;;
;;; E, A |- NOT(B1) <= NOT(B2)  if B2 <= B1
;;;
(defun nlessn (n1 n2 eqns as sols)
  (declare (ignore eqns as sols))
  (subsetp (not-list n2) (not-list n1) :test #'equal)) 

;;;
;;;    E, A |- X <= W, Y <= Z
;;;  -------------------------
;;;   E, A |- X ? Y  <= W ? Z 
;;;
(defun dlessd (d1 d2 eqns as sols)
  (and
   (type-<=-aux (type-t1 d1) (type-t1 d2) eqns as sols)
   (type-<=-aux (type-t2 d1) (type-t2 d2) eqns as sols)))

