(in-package 'types)

;;;
;;; Functions for maintaining a substitution.
;;;
;;; A substitution is implemented as an association list
;;; mapping variables to types.  If a variable is not in
;;; the substitution, nil is returned.
;;;

(defun substitution-empty () nil)

(defun substitution-add (var typ s) 
  (acons (var-id var) typ s))

(defun substitution-get (var s) 
  (cdr (assoc (var-id var) s :test #'=)))

;;; uniform-subst :: varlist * Type --> list (Var -> type)
;;;
;;; For all v in varlist  create (v -> typ)
;;;

(defun uniform-subst (varlist typ)
  (mapcar #'(lambda (v) (cons (var-id v) typ)) varlist))

;;; substitution-list :: list(vars) * list(types) --> list(var --> type)
;;;
;;; For all i, [(dom-i in dom-lst) and (ran-i in ran-lst)] create (dom-i -> ran-i)
;;; Note dom-lst and ran-lst should be of equal length

(defun substitution-list (dom-lst ran-lst)
  (assert (= (length dom-lst) (length ran-lst)))
  (mapcar #'(lambda (d r) (cons (var-id d) r)) dom-lst ran-lst))

(defun substitution-append (&rest substs)
  (apply #'append substs)) 
	    

(defun kt (&rest x)
  (declare (ignore x))
  t)

;;;
;;; Applying a subsitution to a type.
;;;
(defun type-substitute (typ subst eqns)
  (type-hom typ 
	    #'(lambda (var hom pred eqns) 
		(declare (ignore hom pred eqns))
		(let ((s (substitution-get var subst))) 
		  (if s s :no-change))) 
	    #'kt
	    eqns))

;;; Applying an ALPHA subsitution to a type.
;;;
(defun type-alpha-substitute (typ subst eqns)
  (type-alpha-subst typ 
	    #'(lambda (var hom pred eqns) 
		(declare (ignore hom pred eqns))
		(let ((s (substitution-get var subst))) 
		  (if s s :no-change))) 
	    #'kt
	    eqns))

;;;
;;; Eliminate non-recursive bound variables in a type.
;;; Proceeds recursively if optional argument is true.
;;;
(defun type-simplify (typ eqns &key (rec nil))
  (type-hom typ
	    #'(lambda (var hom pred eqns)
		(declare (ignore hom pred))
		(if (and (is-bvar var)
			 (is-bv-non-recursive var eqns))
		    (if rec
			(type-simplify (bv-def var eqns) eqns :rec t)
		        (bv-def var eqns))
		  :no-change))

	    #'kt
	    eqns))

;;;
;;; Simplify a substitution by applying type-simplify to the
;;; types in the range.
;;;
(defun subst-simplify (subst eqns)
  (mapcar #'(lambda (pair)
	      (cons
	       (car pair)
	       (type-simplify (cdr pair) eqns)))
	  subst))

;;;
;;; Expand top level bound variables.  Works recursively.
;;; WARNING: will go into an infinite loop if there is an immediate
;;; recursion on a bound variable, either directly (i.e. x = x) or
;;; indirectly (i.e. x = y, y = x).  This would indicate a bug in
;;; the system.

(defun not-cons-or-func (typ &rest x)
  (declare (ignore x))
  (not
   (or
    (is-cons typ)
    (is-func typ))))

(defun type-expand (typ var-id eqns)
  (type-hom typ
	    #'(lambda (typ hom pred eqns)
		  (cond
		   ((is-var typ) :no-change)
		   ((> (bvar-id typ) var-id)
		    (let ((d (bv-def typ eqns)))
		      (assert d) ;; bound variable must be defined here
		      (type-hom d hom pred eqns)))
		   ((< (bvar-id typ) var-id)
		    :no-change)
		   (t (break "Immediate recursion in bound variable definition ~a!~%" var-id))
		   ))
	    #'not-cons-or-func
	    eqns))

(defun type-subst-tlv (typ eqns)
  (type-hom typ
	    #'(lambda (typ hom pred eqns)
		(declare (ignore hom pred eqns))
		(if (is-bvar typ) (type-one) :no-change))
	    #'not-cons-or-func
	    eqns))

;;; computes the domain of a substitution
(defun substitution-domain (sub)
  (mapcar #'car sub))

;;;  for x in dom(sub1) union dom(sub2)
;;; sub1@sub2 (x) == sub1(sub2(x))   if x in dom(sub2)
;;; sub1(x)                          if x in dom(sub1) - dom(sub2)

(defun squash (l1 l2)
  (if (null l1) l2 (squash (cdr l1) (cons (car l1) l2))))

(defun substitution-compose (sub1 sub2 eqns)
  (cond
   ((null sub1) sub2)
   ((null sub2) sub1)
   (t (substitution-compose-aux nil sub1 sub2 eqns))))

(defun substitution-compose-aux (compsub sub1 sub2 eqns)
  (if (null sub2)
      (squash compsub sub1)
    (let* ((next (car sub2))
	   (x (car next))
	   (sub2x (cdr next)))
      (substitution-compose-aux 
       (acons x (type-substitute sub2x sub1 eqns) compsub)
       sub1
       (cdr sub2)
       eqns))))

				
		

		       
	
	
	
			  
			       
	
    
