(in-package 'types)

;;;
;;; The functions in this file implement elimination of variables 
;;; in a quantified type.
;;;


(defun map&squash-aux (f l sofar)
  (if (null l)
      sofar
    (map&squash-aux f
		    (cdr l)
		    (squash (funcall f (car l)) sofar))))
		    
(defun map&squash (f l &key (init nil))
  (map&squash-aux f l init))


;;;
;;; vars-reachable-from-set computes the set of all variables reachable from 
;;; a set of variables.  In this case "reachable" means reachable from upper and
;;; lower bounds of free variables, as well as definitions of bound variables.
;;;
(defun vars-reachable-from-set (typ-lst eqns)
  (remove-duplicates
   (map&squash #'(lambda (v) (type-vars-reachable v eqns :fv-expand t)) typ-lst)
   :test #'var-=))


;;;
;;; sort variables by their id's
;;;
(defun var-sort (vs)
  (sort (copy-list vs) #'(lambda (v1 v2) (< (var-id v1) (var-id v2)))))


;;;===================================
;;;
;;; Variable Elimination
;;;
;;;====================================
;;;
;;; Consider a quantified type forall a.(T where S), where T is a type and S is
;;; a set of constraints.  The meaning of this type is
;;;
;;;   /\_{u \in Solutions(S)}  u(T)
;;;
;;; In other words, the meaning is the intersection over all solutions u of S of the
;;; the possible meanings u(T). For example, the type
;;; 
;;;   forall a. (a -> a where {})
;;; 
;;; contains all the functions that are <= to the identity function.
;;;
;;; Often it is possible to eliminate a variable in a quantified type without changing the
;;; meaning of the infinite intersection.  This may be because the variable
;;; appears only monotonically (in which case it can be set to its lower bound implied 
;;; by the constraints) or anti-monitonically (in which case it can be set to its upper bound).
;;; For example, in the type
;;;
;;;    forall a,b.  a -> b where { 0 <= a <= 1,  0 <= b <= 1}
;;;
;;; a appears only anti-monotonically, so it can be set to its upper bound 1, and b appears
;;; only monotonically, so it can be set to its lower bound 0.  Thus this type is equivalent
;;; to the unquantified type  1 -> 0.  
;;;
;;; The notion of when a variable appears monotonically or anti-monotonically is complicated by
;;; the presence of constraints.  If a type is monotonic in a variable a, then the type is also
;;; monotonic in the monotonic variables in a's lower bound, and it is anti-monotonic in the
;;; anti-monotonic variables in a's lower bound.  Conversely, if a type is anti-monotonic in a,
;;; then the type is also anti-monotonic in the monotonic variables in a's upper bound, and
;;; it is monotonic in the variables in a's upper bound.  This leads to the following algorithm.
;;; 
;;; Definition: a variable is "positive in T" if it appears in a monotonic position in T.  A variable
;;; is "negative in T" if it appears in an anti-monotonic position. 
;;;
;;; For a type t with constraints S = { L_i <= \alpha_i <= U_i }, 
;;; we want to compute the least sets P and N such that:
;;;
;;; (1) Pos(t) <= P  ( <= means "contained in")
;;; (2) Neg(t) <= N
;;; (3)  \alpha_i \in P   implies that   Pos(L_i) <= P
;;;                                and   Neg(L_i) <= N  
;;;
;;; (4)  \alpha_i \in N   implies that   Pos(U_i) <= N
;;;                               and    Neg(U_i) <= P  
;;;
;;; b. NEXT, variables are eliminated as follows.
;;;    1. Vars in P - N are set to their LOWER bounds
;;;    2. Vars in N - P are set to their UPPER bounds
;;;    3. Vars in P * N are replaced by fvs with l/u bounds 
;;;       obtained from prev l/u bounds by substitution
;;;    NOTE: Variables in neither set are unchanged
;;; 
;;; 
;;; Initially
;;;           todoP <- Pos(t)
;;;           todoN <- Neg(t)
;;;
;;; while todoP <> {}, todoN <> {}
;;;
;;;   if todoP <> {}
;;;      {
;;;       choose \alpha_i \in todoP
;;;       todoP = todoP - { \alpha_i }
;;;       if \alpha_i \not\in P then
;;;          {
;;;            P <- P | { \alpha_i }
;;;            todoP <- todoP | Pos(L_i)
;;;            todoN <- todoN | Neg(L_i)
;;;          }
;;;     }
;;;
;;;   if todoN <> {}
;;;      {
;;;       choose \alpha_i \in todoN
;;;       todoN = todoN - { \alpha_i }
;;;       if \alpha_i \not\in N then
;;;          {
;;;            N <- N | { \alpha_i }
;;;            todoN <- todoN | Pos(U_i)
;;;            todoP <- todoP | Neg(U_i)
;;;          }
;;;     }
;;;
;;; The algorithm takes an initial set initPN of variables that are not to be minimized or maximized---
;;; that is, if x \in initPN then x \in P and x \in N.  An important assumption that is built into
;;; the algorithm in this version is that vars-reachable-from(InitPN) <= initPN.  In other words, if
;;; a variable is in initPN, then all variables reachable from it are in initPN.
;;;
;;;

(defun get-bound (f v e)
  (cond
   ((and (is-bvar v)
	 (bv-def v e))
    (bv-def v e))
   ((is-bvar v)
    (error "undefined bound variable in var-elim"))
   (t
    (funcall f v e))))

(defun pos-and-neg (typ-lst eqns &key (initPN nil))
  (let* ((todoP (map&squash #'(lambda (typ) (type-pos typ eqns)) typ-lst))
	 (todoN (map&squash #'(lambda (typ) (type-neg typ eqns)) typ-lst))
	 (P-fix nil)
	 (N-fix nil))
    (do* ()
	 ((and (null todoP) (null todoN))
	  (values P-fix N-fix))
	 
	 (when todoP 
	       (let ((alpha (pop todoP)))
		 (when (and (not (member alpha P-fix :test #'var-=))
			    (not (member alpha initPN :test #'var-=)))
		       (push alpha P-fix)
		       (setq todoP (squash
				    (type-pos (get-bound #'constr-lower-bound alpha eqns) eqns) 
				    todoP))
		       (setq todoN (squash
				    (type-neg (get-bound #'constr-lower-bound alpha eqns) eqns) 
				    todoN)))))
	 (when todoN 
	       (let ((alpha (pop todoN)))
		 (when (and (not (member alpha N-fix :test #'var-=))
			    (not (member alpha initPN :test #'var-=)))
		       (push alpha N-fix)
		       (setq todoN (squash
				    (type-pos (get-bound #'constr-upper-bound alpha eqns) eqns) 
				    todoN))
		       (setq todoP (squash
				    (type-neg (get-bound #'constr-upper-bound alpha eqns) eqns) 
				    todoP)))))
	 )))


;;; 
;;; instantiate variables to their upper or lower bounds.
;;;
(defun minimize-var (v sub-var eqns)
  (bv-add-def (type-substitute v sub-var eqns)
	      (type-substitute (constr-lower-bound v eqns) sub-var eqns) 
	      eqns))


(defun maximize-var (v sub-var eqns)
  (bv-add-def (type-substitute v sub-var eqns)
	      (type-substitute (constr-upper-bound v eqns) sub-var eqns) 
	      eqns))


(defun add-var-constr (v L U eqns)
  (add-todo-constraint L v eqns)
  (add-todo-constraint v U eqns))

;;;
;;; min&max-vars instantiates variables to their new values.
;;; There are several cases depending on whether the variable is free or bound,
;;; and whether the variable is to be minimized, maximized, or neither.
;;;
;;; Every new definition is expanded to get rid of bound variables at the top level.
;;; Also, the variables are processed in order from lowest to highest;
;;; this ensures that bound variables are defined when they are needed for expansion.
;;;
(defun min&max-vars (dom sub-var eqns min-dom max-dom)
  (mapc
   #'(lambda (v) 
       (cond 
	;;
	;; Bound variables are instantiated to new bound variables where
	;; the rhs has the substitution applied.
	;;
	((is-bvar v) 
	 (bv-add-def (type-substitute v sub-var eqns)
		     (type-substitute (bv-def v eqns) sub-var eqns)
		     eqns))
	;;
	;; Free variables in min-dom are minimized.
	;;
	((and (is-var v) (member v min-dom :test #'var-=)) 
	 (minimize-var v sub-var eqns))
	;;
	;; Free variables in max-dom are maximized.
	;;
	((and (is-var v) (member v max-dom :test #'var-=))
	 (maximize-var v sub-var eqns))
	;;
	;; Any other variables must be in P&N.  These variables are just
	;; renamed and the substitution applied to their constraints.
	;;
	((is-var v)
	 (add-var-constr
	  (type-substitute v sub-var eqns)
	  (type-substitute (constr-lower-bound v eqns) sub-var eqns)
	  (type-substitute (constr-upper-bound v eqns) sub-var eqns)
	  eqns))
	( t (break "ERROR in computing P/N; argument not a b/f var" v))))
   dom)
  eqns)


;;;
;;; remove elements of vs which do not reach a variable in reach
;;;
(defun remove-if-not-reaching (vs reach eqns)
  (remove-if-not 
   #'(lambda (v) (type-reaches-vars v reach eqns))
   vs))
       


;;;==============================================================
;;;
;;; Functions for determining when an intersection of variables
;;; can be replaced by a single variable.
;;;
;;;
;;; Thm:  Let forall a,b. T where {Li <= Xi <= Ui } be a type and assume that
;;; in every anti-monotonic position in T in which a (resp. b) occurs,
;;; a (resp. b) is intersected with b (resp. a). Furthermore, assume
;;; that for each constraint Li <= Xi <= Ui,
;;; that in every anti-monotonic position in Li and monotonic position
;;; in Ui in which a (resp. b) occurs, a (resp. b) is intersected with b (resp. a).
;;; Finally, assume that for La <= a <= Ua and Lb <= b <= Ub, La <= a u b and
;;; Lb <= a u b.  Then
;;;
;;;  forall a,b. T where S  = forall c. T[c/a,c/b] where S[c/a,c/b]
;;;
;;; Proof: Let \sigma be any solution of S.  Consider
;;; \sigma' = \sigma[ a <- \sigma(a&b), b <- \sigma(a&b) ].
;;; Show \sigma' is a solution.  Let Li <= xi <= Ui be a constraint
;;; where a \= xi \= b.
;;; Now from the assumptions and because \sigma(a&b) = \sigma'(a&b) it follows that
;;;
;;; \sigma'(L) <= \sigma(L) <= \sigma(xi) = \sigma'(xi) <= \sigma(Ui) <= \sigma'(Ui)
;;;
;;; Furthermore,
;;; \sigma'(La) <= \sigma'(a u b) = \sigma'(a&b) = \sigma'(a) <= \sigma'(Ua)
;;; \sigma'(Lb) <= \sigma'(a u b) = \sigma'(a&b) = \sigma'(b) <= \sigma'(Ub)
;;;
;;; so \sigma' is a solution of S.  From the assumptions, we know
;;; \sigma'(T) <= \sigma(T).
;;;
;;;==============================================================

;;;
;;; compute-one-inter takes an intersection of variables V0= v1&...&vn
;;; (represented as a list of variables) and a list of intersections
;;; of variables (V1,...,Vm) (represented as a list of lists).
;;; It returns two values Z and Y, defined as follows.
;;; Let X = { ... Vi ... } be all the Vi such that Vi and V0 share a variable,
;;;    for 1 <= i <= m.
;;; Let Y = { ... Vj ... } be all the Vj such that Vj and V0 share no variables
;;;    for 1 <= j <= m.
;;; Then compute-one-inter returns:
;;;      Z = V0 & x0 & ... & xk   where { X = {x0,...,xk }
;;; and  Y
;;;
;;; Z represents a set of variables that are guaranteed to always appear together in
;;; negative positions in a type.  These variables can be replaced by a single variable.
;;; 
;;; Y is the set of all intersections of variables that do not interact at all
;;; with V0.
;;;
(defun compute-one-inter (ts listofts)
  (if (null listofts)
      (values ts :no-change)
    (multiple-value-bind
     (min-ts new-listofts)
     (compute-one-inter ts (cdr listofts))
     (let ((s (intersection ts (car listofts) :test #'var-=)))
       (cond
	((and (null s)
	      (eq :no-change new-listofts))
	 (values ts :no-change))
	((and (not (null s))
	      (eq :no-change new-listofts))
	 (values s (cdr listofts)))
	((and (null s)
	      (not (eq :no-change new-listofts)))
	 (values min-ts 
		 (cons (car listofts) new-listofts)))
	(t ;; (and (not (null s)) (not (eq :no-change new-listofts)))
	 (values (intersection s min-ts :test #'var-=)
		 new-listofts)))))))

;;;
;;; This function iterates compute-one-inter to compute sets of variables which
;;; are guaranteed to appear together in intersections in negative positions in 
;;; a type.  If variables a and b always appear as  a&b in every occurrence of
;;; a or b in an anti-monotonic position, then a&b can be replaced by a single 
;;; variable c.  Sets of size 0 or 1 are discarded, since it is not possible
;;; to reduce the total number of variables in these cases.
;;;
(defun compute-minimizable-intersections (todo sofar)
  (if (null todo)
      sofar
    (multiple-value-bind
     (next-inter new-todo)
     (compute-one-inter (car todo) (cdr todo))
     (compute-minimizable-intersections 
      (if (eq :no-change new-todo)
	  (cdr todo)
	new-todo)
      ;;
      ;; throw out lists of length 0 or 1
      ;;
      (if (or (null next-inter)
	      (null (cdr next-inter)))
	  sofar 
	(cons next-inter sofar))))))


(defun min-by-intersection (ts P N eqns initPN)
  (let* (
	 ;;
	 ;; compute anti-monotone intersections in the type
	 ;;
	 (typ-inters 
	  (mapcar #'(lambda (typ) (inter-list typ :l eqns)) ts))
	 ;;
	 ;; compute anti-monotone intersections in lower bounds of relevant vars
	 ;;
	 (p-inters 
	  (mapcar #'(lambda (v) 
		      (inter-list (constr-lower-bound v eqns) :l eqns))
		  P))
	 ;;
	 ;; compute monotone intersections in upper bounds of relevant vars
	 ;;
	 (n-inters 
	  (mapcar #'(lambda (v)
		      (inter-list (constr-upper-bound v eqns) :r eqns))
		  N))
	 ;;
	 ;; combine the three sets
	 ;;
	 (all-inters 
	  (append 
	   (apply #'append typ-inters)
	   (apply #'append p-inters) 
	   (apply #'append n-inters)))
	 ;;
	 ;; discard variables v in sets V where
	 ;;    1. v is reachable from the assumptions
	 ;;    2. L <= v is a constraint and L Not<= {union of x \in V}
	 ;;
	 (all-but 
	  (mapcar #'(lambda (vs)
		      (let ((vtyp (reduce #'(lambda (t1 t2) (type-union t1 t2 eqns :nosubset t))
				       vs :initial-value (type-zero))))
			(remove-if #'(lambda (v) 
				       (or (not (type-<= (constr-lower-bound v eqns) vtyp eqns))
					   (member v initPN :test #'var-=)))
				       vs)))
		  all-inters))
	 ;;
	 ;; compute sets of variables that appear together in intersections
	 ;;
	 (min-groups 
	  (compute-minimizable-intersections all-but  nil)))
    
    (if (null min-groups) ;; An optimization: If there are no variables to instantiate, we are done.
	(values ts nil)
      (let* ((elim-vars (apply #'append min-groups))  ;; variables to be eliminated
	     (rename                                  ;; variables that depend on eliminated variables
	      (remove-if-not-reaching
	       (union
		(set-difference P elim-vars :test #'var-=)
		(set-difference N elim-vars :test #'var-=)
		:test #'var-=)
	       elim-vars
	       eqns))
	     ;;
	     ;; for each group of vars, allocate 1 new var
	     ;;
	     (elim-range                                    
	      (apply #'append
		     (mapcar #'(lambda (vs)
				 (let ((v (next-var eqns)))
				   (mapcar #'(lambda (x) (declare (ignore x)) v) vs)))
			     min-groups)))
	     ;;
	     ;; allocate new var for every var to be renamed
	     ;;
	     (rename-range                                  
	      (mapcar #'(lambda (v)
			  (if (is-var v)
			      (next-var eqns)
			    (next-bvar eqns :rl (type-rl v))))
		      rename))
	     ;;
	     ;; build and apply the substitution
	     ;;
	     (dom (append elim-vars rename))          
	     (range (append elim-range rename-range))
	     (sub-var (substitution-list dom range))
	     (eqns (min&max-vars dom sub-var eqns nil nil))
	     (sub (subst-simplify sub-var eqns)))
	(values
	 (mapcar #'(lambda (x) (type-expand x 0 eqns))
		 (mapcar #'(lambda (x) (type-substitute x sub eqns))
			 ts))
	 sub)))))
	



;;;================================================================
;;;
;;;    Functions for eliminating variables L <= a <= U
;;;    where U <= L in all solutions.
;;;
;;;===============================================================

;;;
;;; In general, completely-constrained-vars takes a set of variables V and computes
;;; the subset Y <= V such that for every x in Y,
;;; if L <= x <= U is a constraint, then U <= L in all substitutions.
;;; This is a cheap way of picking up some more variables that can
;;; be eliminated, in this case by setting the variable to either
;;; its upper or lower bound.  Care must be taken to pick a bound
;;; that is both an R and an L type.
;;;
(defun completely-constrained-vars (V eqns)    
  (let ((new-min nil)
	(new-max nil)
	(other nil))
    (mapc
     #'(lambda (x)
	 (if (type-<= (constr-upper-bound x eqns)
		      (constr-lower-bound x eqns)
		      eqns
		      :sols t)
	     (cond
	      ((is-rltype (constr-lower-bound x eqns))
	       (push x new-min))
	      ((is-rltype (constr-upper-bound x eqns))
	       (push x new-max))
	      (t
	       (push x other)))
	   (push x other)))
     V)
    (values new-min new-max other)))



;;;===============================================================
;;;
;;; Functions for eliminating variables that are either
;;; monotonic or antimonotonic.
;;;
;;;===============================================================

;;;
;;; min&max-substitution takes a list of types and a list of positive and negative variables
;;; and builds a substitution such that:
;;;   - variables in P - N are minimized
;;;   - variables in N - P are maximized
;;; The substitution is applied to the types in tlist, and the new list of types and
;;; the substitution are returned.
;;;
;;; There is a difficulty with variables in P & N.  These variables cannot be minimized or
;;; maximized, but they may depend on variables that are minimized or maximized.  For example,
;;; consider a constraint   L <= v <= U.  The correct solution is to allocate a new variable
;;; v', and add constraints  s(L) <= s(v) <= s(U), where s is the min-max substitution and
;;; s(v) = v'.
;;;
;;; The algorithm below performs several optimizations to avoid allocating new variables
;;; when possible.
;;; 
(defun  min&max-substitution (tlist P N eqns initPN)
  (declare (ignore initPN))
  (let* ((Pfv (remove-if-not #'is-var P))
	 (Nfv (remove-if-not #'is-var N))
	 (min (set-difference Pfv Nfv :test #'var-=)) ;; minimizable fvars
	 (max (set-difference Nfv Pfv :test #'var-=)) ;; maximizable fvars
	 (PN (intersection Pfv Nfv :test #'var-=)))   ;; fvars in both P and N
    (multiple-value-bind
     (min-add max-add rest)
     (completely-constrained-vars PN eqns)            ;; fvars L <= v <= U where U = L
     (let* ((min (append min-add min))                ;; new set to be minimized
	    (max (append max-add max))                ;; new set to be maximized
	    (mod (append max min)))                   ;; all free vars that will be instantiated

    (if (null mod)    ;; An optimization: If there are no variables to instantiate, we are done.
	(values tlist nil)
      (let* (;;
	     ;; P&N is the set of free variables that are both positive and negative, but depend
	     ;; on at least one free var that is either positive or negative; these variables
	     ;; cannot be maximized or minimized, but must be renamed.
	     ;;
	     (P&N (remove-if-not-reaching rest mod eqns))
	     ;;
	     ;; The bvs that must be renamed are those that depend on at least one free
	     ;; free variable that is instantiated.
	     ;;
	     (bvs (remove-if-not-reaching
		   (union (remove-if-not #'is-bvar P)
			  (remove-if-not #'is-bvar N)
			  :test #'var-=)
		   mod eqns))
	     (dom (var-sort (append mod P&N bvs)))
	     (range (mapcar #'(lambda (v)
				(if (and (is-var v) (member v P&N :test #'var-=)) 
				    (next-var eqns)
				    (next-bvar eqns :rl (type-rl v))))
			    dom))
	     (sub-var (substitution-list dom range))         ;; build substitution 
	     (eqns (min&max-vars dom sub-var eqns min max))  ;; perform instantiations
	     (sub (subst-simplify sub-var eqns)))            ;; eliminate zero-variables 
	(values                                              ;; perform subst on types & expand
	 (mapcar #'(lambda (x) (type-expand x 0 eqns))
		 (mapcar #'(lambda (x) (type-substitute x sub eqns))
			 tlist))
	 sub)))))))
  

			      
;;;
;;; var-elim iteratively eliminates variables from a type until no more can be
;;; eliminated.  The result is a new list of types and the composition of all
;;; the substitutions performed.
;;;
(defun var-elim-list (tlist eqns &key (initPN nil))
  (let ((initPN-vars (vars-reachable-from-set initPN eqns)))
    (var-elim-list-aux tlist nil initPN-vars eqns)))

(defun var-elim-list-aux (tlist subst initPN eqns)
  (multiple-value-bind
   (tlist1 subst1 any1)
   (var-elim-aux #'min&max-substitution tlist subst initPN eqns)
   (declare (ignore any1))
   (multiple-value-bind
    (tlist2 subst2 any2)
    (var-elim-aux #'min-by-intersection tlist1 subst1 initPN eqns nil)
    (declare (ignore any2))
    (values tlist2 subst2))))
 
(defun var-elim-aux (fn tlist subst initPN eqns &optional (repeat t) (any nil))
  (multiple-value-bind
   (P N) 
   (pos-and-neg tlist eqns :initPN initPN)
   (multiple-value-bind
    (newtlist newsubst)
    (funcall fn tlist P N eqns initPN)
    (solve-constraints eqns)
    (cond
     ((null newsubst)
      (values tlist subst any))
     ((null repeat)
      (values newtlist
	      (substitution-compose newsubst subst eqns)
	      t))
     (t
      (var-elim-aux fn
		    newtlist 
		    (substitution-compose newsubst subst eqns) 
		    initPN 
		    eqns
		    t t))))))

;;;
;;; Overloaded version of var-elim-list that works on a single type or
;;; a list of types.
;;;
(defun var-elim (arg eqns &key (initPN nil))
  (let ((tlist (if (is-type arg) (list arg) arg)))
    (multiple-value-bind
     (l s)
     (var-elim-list tlist eqns :initPN initPN)
     (values
      (if (is-type arg) (first l) l)
      s))))
